/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2006 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "stdafx.h"
#include "ExtractResource.h"

#define WIDTHBYTES(bits)      ((((bits) + 31)>>5)<<2)

CIconExtractor::CIconExtractor()
{
    m_GKounter     = 0;
	m_GNameKounter = 0;
}

HICON CIconExtractor::ExtractIconFromExeEx(LPCSTR SourceEXE, UINT IconIndex, HBITMAP &destBitmap, int &maxIconSize)
{
	HICON				resultIcon = NULL;
	LPICONRESOURCE      lpIR = NULL;
	HINSTANCE        	hLibrary;
	DWORD               ret;

	ASSERT(IconIndex < MAX_ICONS);

    // Load the EXE - NOTE: must be a 32bit EXE for this to work
    if( (hLibrary = LoadLibraryEx( SourceEXE, NULL, LOAD_LIBRARY_AS_DATAFILE )) == NULL )
		return NULL;
 
    HRSRC        	hRsrc   = NULL;
    HGLOBAL        	hGlobal = NULL;
    LPMEMICONDIR    lpIcon  = NULL;
	UINT i;

	m_GKounter = 0;
    m_GNameKounter = 0;

 	ret = EnumResourceNames( hLibrary, RT_GROUP_ICON, EnumResNameProc, (LONG)this );

	if(!ret)
	{
		ret = GetLastError();
		FreeLibrary(hLibrary);
		return NULL;
	}

	// Find the group icon resource
	//Finding the resource with the ID first
	if(m_GKounter > 0)
		hRsrc = FindResource(hLibrary,MAKEINTRESOURCE(m_GroupIconID[IconIndex]),RT_GROUP_ICON);

	if(hRsrc == NULL)//Searching the icon with names if any
		hRsrc = FindResource(hLibrary,m_GroupIconName[IconIndex],RT_GROUP_ICON);
	
	if(hRsrc == NULL)
	{
		ret = GetLastError();
		FreeLibrary(hLibrary);
		return NULL;
	}

    if( (hGlobal = LoadResource( hLibrary, hRsrc )) == NULL )
    {
        FreeLibrary( hLibrary );
		return NULL;
    }
    if( (lpIcon = (LPMEMICONDIR)LockResource(hGlobal)) == NULL )
    {
        FreeLibrary( hLibrary );
		return NULL;
    }
 
	// Allocate enough memory for the images
    if( (lpIR = (LPICONRESOURCE) malloc( sizeof(ICONRESOURCE) + ((lpIcon->idCount-1) * sizeof(ICONIMAGE)) )) == NULL )
    {
		FreeLibrary( hLibrary );
		return NULL;
	}
 
	// Fill in local struct members
    lpIR->nNumImages = lpIcon->idCount;
   
    // Loop through the images
    for( i = 0; i < lpIR->nNumImages; i++ )
    {
		// Get the individual image
        if( (hRsrc = FindResource( hLibrary, MAKEINTRESOURCE(lpIcon->idEntries[i].nID), RT_ICON )) == NULL )
        {
			FreeLibrary( hLibrary );
	  		free(lpIR);
   			return NULL;
        }
        if( (hGlobal = LoadResource( hLibrary, hRsrc )) == NULL )
        {
			FreeLibrary( hLibrary );
	  		free(lpIR);
    		return NULL;
        }
		// Store a copy of the resource locally
        lpIR->IconImages[i].dwNumBytes = SizeofResource( hLibrary, hRsrc );

	    lpIR->IconImages[i].lpBits =(LPBYTE) malloc( lpIR->IconImages[i].dwNumBytes );
		if(lpIR->IconImages[i].lpBits == NULL)
		{	
			FreeLibrary( hLibrary );
			free(lpIR);
			return NULL;
		}

        memcpy( lpIR->IconImages[i].lpBits, LockResource( hGlobal ), lpIR->IconImages[i].dwNumBytes );

        // Adjust internal pointers
        if( ! AdjustIconImagePointers( &(lpIR->IconImages[i]) ) )
        {
			FreeLibrary( hLibrary );
			free(lpIR);
    		return NULL;
        }
	}

	int maxColor = 0;
	LPICONIMAGE iconbest = NULL;

	// First (Optimal) Loop through the images
	for(UINT j = 0; j < lpIR->nNumImages; j++)
	{
		LPICONIMAGE icon1 = &(lpIR)->IconImages[j];
		
		if ( IsOS( OS_2KXP ) && icon1->Colors == 32 && icon1->Width == maxIconSize )
		{
			HICON hIcon = MakeIconFromResource(icon1);
				
			ICONINFO iconinfo;
			GetIconInfo(hIcon, &iconinfo);

			maxIconSize = icon1->Width;

			destBitmap = (HBITMAP) CopyImage(iconinfo.hbmColor, IMAGE_BITMAP, 0, 0, LR_CREATEDIBSECTION);

			DeleteObject(iconinfo.hbmColor);
			DeleteObject(iconinfo.hbmMask);
			DestroyIcon( hIcon );

			break;
		}
		
		if ( icon1->Colors < 32 )
		{
			if ( icon1->Colors >= maxColor && icon1->Width == maxIconSize )
			{
				iconbest = icon1;
				maxColor = icon1->Colors;
			}
		}
	}

	if ( iconbest == NULL && destBitmap == NULL )
	{
		int bestSize = 0;

		// Second Loop through the images
		for(UINT j = 0; j < lpIR->nNumImages; j++)
		{
			LPICONIMAGE icon1 = &(lpIR)->IconImages[j];
			
			if ( IsOS( OS_2KXP ) && icon1->Colors == 32 && icon1->Width >= bestSize )
			{
				HICON hIcon = MakeIconFromResource(icon1);
					
				ICONINFO iconinfo;
				GetIconInfo(hIcon, &iconinfo);

				bestSize = icon1->Width;

				if (destBitmap != NULL)
				{
					DeleteObject(destBitmap);
					destBitmap = NULL;
				}

				destBitmap = (HBITMAP) CopyImage(iconinfo.hbmColor, IMAGE_BITMAP, 0, 0, LR_CREATEDIBSECTION);

				DeleteObject(iconinfo.hbmColor);
				DeleteObject(iconinfo.hbmMask);
				DestroyIcon( hIcon );
			}
			
			if ( icon1->Colors < 32 )
			{
				if ( icon1->Colors >= maxColor && icon1->Width >= bestSize )
				{
					iconbest = icon1;
					maxColor = icon1->Colors;
					bestSize = icon1->Width;
				}
			}
		}
	}

	if ( destBitmap == NULL && iconbest != NULL )
	{
		resultIcon = MakeIconFromResource(iconbest);
	}

    FreeLibrary( hLibrary );
	free(lpIR);

	if ( destBitmap != NULL )
	{
		if ( resultIcon )
			DestroyIcon( resultIcon );

		return NULL;
	}
	else
	{
		if ( resultIcon )
			return resultIcon;
		else 
			return NULL;
	}
}

HICON CIconExtractor::ExtractIconFromIcoEx( LPCSTR SourceICO, HBITMAP &destBitmap, int &maxIconSize )
{
	HICON				resultIcon = NULL;
	LPICONRESOURCE    	lpIR = NULL;
	LPICONRESOURCE		lpNew = NULL;
	HANDLE            	hFile = NULL;
	UINT                i;
	DWORD            	dwBytesRead;
	LPICONDIRENTRY    	lpIDE = NULL;

	// Open the file
	if( (hFile = CreateFile( SourceICO, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL )) == INVALID_HANDLE_VALUE )
	{
		return NULL;
	}
	// Allocate memory for the resource structure
	if( (lpIR = (LPICONRESOURCE) malloc( sizeof(ICONRESOURCE) )) == NULL )
	{
		CloseHandle( hFile );
		return NULL;
	}
	// Read in the header
	if( (lpIR->nNumImages = ReadICOHeader( hFile )) == (UINT)-1 )
	{
		CloseHandle( hFile );
		free( lpIR );
		return NULL;
	}
	// Adjust the size of the struct to account for the images
	if( (lpNew = (LPICONRESOURCE) realloc( lpIR, sizeof(ICONRESOURCE) + ((lpIR->nNumImages-1) * sizeof(ICONIMAGE)) )) == NULL )
	{
		CloseHandle( hFile );
		free( lpIR );
		return NULL;
	}
	lpIR = lpNew;

	// Allocate enough memory for the icon directory entries
	if( (lpIDE = (LPICONDIRENTRY) malloc( lpIR->nNumImages * sizeof( ICONDIRENTRY ) ) ) == NULL )
	{
		CloseHandle( hFile );
		free( lpIR );
		return NULL;
	}
	// Read in the icon directory entries
	if( ! ReadFile( hFile, lpIDE, lpIR->nNumImages * sizeof( ICONDIRENTRY ), &dwBytesRead, NULL ) )
	{
		CloseHandle( hFile );
		free( lpIR );
		free( lpIDE );
		return NULL;
	}
	if( dwBytesRead != lpIR->nNumImages * sizeof( ICONDIRENTRY ) )
	{
		CloseHandle( hFile );
		free( lpIR );
		free( lpIDE );
		return NULL;
	}
	// Loop through and read in each image
	for( i = 0; i < lpIR->nNumImages; i++ )
	{
		// Allocate memory for the resource
		if( (lpIR->IconImages[i].lpBits = (BYTE*)malloc(lpIDE[i].dwBytesInRes)) == NULL )
		{
			CloseHandle( hFile );
			free( lpIR );
			free( lpIDE );
			return NULL;
		}
		lpIR->IconImages[i].dwNumBytes = lpIDE[i].dwBytesInRes;
		// Seek to beginning of this image
		if( SetFilePointer( hFile, lpIDE[i].dwImageOffset, NULL, FILE_BEGIN ) == 0xFFFFFFFF )
		{
			CloseHandle( hFile );
			free( lpIR );
			free( lpIDE );
			return NULL;
		}
		// Read it in
		if( ! ReadFile( hFile, lpIR->IconImages[i].lpBits, lpIDE[i].dwBytesInRes, &dwBytesRead, NULL ) )
		{
			CloseHandle( hFile );
			free( lpIR );
			free( lpIDE );
			return NULL;
		}
		if( dwBytesRead != lpIDE[i].dwBytesInRes )
		{
			CloseHandle( hFile );
			free( lpIR );
			free( lpIDE );
			return NULL;
		}
		// Set the internal pointers appropriately
		if( ! AdjustIconImagePointers( &(lpIR->IconImages[i]) ) )
		{
			CloseHandle( hFile );
			free( lpIR );
			free( lpIDE );
			return NULL;
		}
	}

	// Clean up	
	CloseHandle( hFile );
	free( lpIDE );

	int maxColor = 0;
	LPICONIMAGE iconbest = NULL;

	// First (Optimal) Loop through the images
	for(UINT j = 0; j < lpIR->nNumImages; j++)
	{
		LPICONIMAGE icon1 = &(lpIR)->IconImages[j];
		
		if ( IsOS( OS_2KXP ) && icon1->Colors == 32 && icon1->Width == maxIconSize )
		{
			HICON hIcon = MakeIconFromResource(icon1);
				
			ICONINFO iconinfo;
			GetIconInfo(hIcon, &iconinfo);

			maxIconSize = icon1->Width;

			destBitmap = (HBITMAP) CopyImage(iconinfo.hbmColor, IMAGE_BITMAP, 0, 0, LR_CREATEDIBSECTION);

			DeleteObject(iconinfo.hbmColor);
			DeleteObject(iconinfo.hbmMask);
			DestroyIcon( hIcon );

			break;
		}
		
		if ( icon1->Colors < 32 )
		{
			if ( icon1->Colors >= maxColor && icon1->Width == maxIconSize )
			{
				iconbest = icon1;
				maxColor = icon1->Colors;
			}
		}
	}

	if ( iconbest == NULL && destBitmap == NULL )
	{
		int bestSize = 0;

		// Second Loop through the images
		for(UINT j = 0; j < lpIR->nNumImages; j++)
		{
			LPICONIMAGE icon1 = &(lpIR)->IconImages[j];
			
			if ( IsOS( OS_2KXP ) && icon1->Colors == 32 && icon1->Width > bestSize )
			{
				HICON hIcon = MakeIconFromResource(icon1);
				
				ICONINFO iconinfo;
				GetIconInfo(hIcon, &iconinfo);

				bestSize = icon1->Width;

				if (destBitmap != NULL)
				{
					DeleteObject(destBitmap);
					destBitmap = NULL;
				}

				destBitmap = (HBITMAP) CopyImage(iconinfo.hbmColor, IMAGE_BITMAP, 0, 0, LR_CREATEDIBSECTION);

				DeleteObject(iconinfo.hbmColor);
				DeleteObject(iconinfo.hbmMask);
				DestroyIcon( hIcon );
			}
			
			if ( icon1->Colors < 32 )
			{
				if ( icon1->Colors >= maxColor && icon1->Width >= bestSize )
				{
					iconbest = icon1;
					maxColor = icon1->Colors;
					bestSize = icon1->Width;
				}
			}
		}
	}

	if ( destBitmap == NULL && iconbest != NULL )
	{
		resultIcon = MakeIconFromResource(iconbest);
	}

	free(lpIR);

	if ( destBitmap != NULL )
	{
		if ( resultIcon )
			DestroyIcon( resultIcon );

		return NULL;
	}
	else
	{
		if ( resultIcon )
			return resultIcon;
		else 
			return NULL;
	}
}

HICON CIconExtractor::MakeIconFromResource( LPICONIMAGE lpIcon )
{
	HICON hIcon = NULL;

	// Sanity Check
	if( lpIcon == NULL )
		return NULL;
	if( lpIcon->lpBits == NULL )
		return NULL;

	// Let the OS do the real work :)
	hIcon = CreateIconFromResourceEx( lpIcon->lpBits, lpIcon->dwNumBytes, TRUE, 0x00030000, 
		(*(LPBITMAPINFOHEADER)(lpIcon->lpBits)).biWidth, (*(LPBITMAPINFOHEADER)(lpIcon->lpBits)).biHeight/2, 0 );

	// It failed, odds are good we're on NT so try the non-Ex way
	if( hIcon == NULL )
	{
		// We would break on NT if we try with a 16bpp image
		if(lpIcon->lpbi->bmiHeader.biBitCount != 16)
		{	
			hIcon = CreateIconFromResource( lpIcon->lpBits, lpIcon->dwNumBytes, TRUE, 0x00030000 );
		}
	}

	return hIcon;
}

UINT CIconExtractor::ReadICOHeader( HANDLE hFile )
{
	WORD    Input;
	DWORD	dwBytesRead;

	// Read the 'reserved' WORD
	if( ! ReadFile( hFile, &Input, sizeof( WORD ), &dwBytesRead, NULL ) )
		return (UINT)-1;
	// Did we get a WORD?
	if( dwBytesRead != sizeof( WORD ) )
		return (UINT)-1;
	// Was it 'reserved' ?   (ie 0)
	if( Input != 0 )
		return (UINT)-1;
	// Read the type WORD
	if( ! ReadFile( hFile, &Input, sizeof( WORD ), &dwBytesRead, NULL ) )
		return (UINT)-1;
	// Did we get a WORD?
	if( dwBytesRead != sizeof( WORD ) )
		return (UINT)-1;
	// Was it type 1?
	if( Input != 1 )
		return (UINT)-1;
	// Get the count of images
	if( ! ReadFile( hFile, &Input, sizeof( WORD ), &dwBytesRead, NULL ) )
		return (UINT)-1;
	// Did we get a WORD?
	if( dwBytesRead != sizeof( WORD ) )
		return (UINT)-1;
	// Return the count
	return Input;
}

BOOL CIconExtractor::AdjustIconImagePointers( LPICONIMAGE lpImage )
{
    // Sanity check
    if( lpImage==NULL )
        return FALSE;

    // BITMAPINFO is at beginning of bits
    lpImage->lpbi = (LPBITMAPINFO)lpImage->lpBits;
    // Width - simple enough
    lpImage->Width = lpImage->lpbi->bmiHeader.biWidth;
    // Icons are stored in funky format where height is doubled - account for it
    lpImage->Height = (lpImage->lpbi->bmiHeader.biHeight)/2;
    // How many colors?
    lpImage->Colors = lpImage->lpbi->bmiHeader.biPlanes * lpImage->lpbi->bmiHeader.biBitCount;
    // XOR bits follow the header and color table
    lpImage->lpXOR = (PBYTE)FindDIBBits((LPSTR)lpImage->lpbi);
    // AND bits follow the XOR bits
    lpImage->lpAND = lpImage->lpXOR + (lpImage->Height*BytesPerLine((LPBITMAPINFOHEADER)(lpImage->lpbi)));

    return TRUE;
}

LPSTR CIconExtractor::FindDIBBits( LPSTR lpbi )
{
   return ( lpbi + *(LPDWORD)lpbi + PaletteSize( lpbi ) );
}


WORD CIconExtractor::PaletteSize( LPSTR lpbi )
{
    return ( DIBNumColors( lpbi ) * sizeof( RGBQUAD ) );
}

DWORD CIconExtractor::BytesPerLine( LPBITMAPINFOHEADER lpBMIH )
{
    return WIDTHBYTES(lpBMIH->biWidth * lpBMIH->biPlanes * lpBMIH->biBitCount);
}

WORD CIconExtractor::DIBNumColors( LPSTR lpbi )
{
    WORD wBitCount;
    DWORD dwClrUsed;

    dwClrUsed = ((LPBITMAPINFOHEADER) lpbi)->biClrUsed;

    if (dwClrUsed)
        return (WORD) dwClrUsed;

    wBitCount = ((LPBITMAPINFOHEADER) lpbi)->biBitCount;

    switch (wBitCount)
    {
        case 1: return 2;
        case 4: return 16;
        case 8:	return 256;
        default:return 0;
    }

    return 0;
}

BOOL CIconExtractor::AddResource( LPCTSTR lpszType, LPTSTR lpszName )
{
	if(m_GKounter >= MAX_ICONS || m_GNameKounter >= MAX_ICONS)
		return TRUE;

	if(lpszType == RT_GROUP_ICON)
	{
		if((ULONG)lpszName < 65536)//ID
		{
			m_GroupIconID[m_GKounter] =  (ULONG)lpszName ;
			m_GKounter += 1;
		}
		else //NAME
		{
			m_GroupIconName[m_GNameKounter] = lpszName ;
			m_GNameKounter++;

		}
	}

	return TRUE;
}



BOOL CALLBACK CIconExtractor::EnumResNameProc( HMODULE hModule,   // module handle
												LPCTSTR lpszType,  // resource type
												LPTSTR lpszName,   // resource name
												LONG_PTR lParam)    // application-defined parameter
{
	CIconExtractor *ptr =(CIconExtractor*)lParam;
    return ptr->AddResource(lpszType,lpszName);	
}










