/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2006 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#if !defined(AFX_ICONTEXTURE_H__F00FC03F_E2C8_422B_A076_4B8994F42945__INCLUDED_)
#define AFX_ICONTEXTURE_H__F00FC03F_E2C8_422B_A076_4B8994F42945__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Texture.h"

class IconTexture : public Texture
{

public:
	IconTexture();
	virtual ~IconTexture();

	void configure(const string &uniqueKey, const string &groupKey, bool alphaMap);
	void apply(HDC hDC, int x, int y, int width, int height, int transparencyMode, int imageType);
	void applyOverlay(HDC hDC, int x, int y, int width, int height, int alphaValue, int imageType);
	void applyAnim(HDC hDC, int x, int y, int width, int height, int transparencyMode, int imageType, int count) { return; };

	bool isTransparencyNeeded() { return transparencyNeeded; };
	int isTextureType() { return 1; };
	bool useHoverState() { return useHover; };
	bool usePressedState() { return usePressed; };

private:

	bool useHover;
	bool usePressed;

	void GetIconFromLnkTarget(CString path, int _iconSize, int imageMode, HICON &resultIcon);

	void CreateDCFromSource(HDC &destDC, HBITMAP &destBitmap, COLORREF _hueColor, int _hueIntensity, int _luminanceIntensity, int _saturationIntensity, COLORREF _mixColor, int _mixIntensity, int _iconSize, HICON noalphaMapIcon);
	
	static void RGB2HSL(BYTE& R, BYTE& G, BYTE& B, BYTE& H, BYTE& S, BYTE& L)
	{
		R = min(max(0,R),255);
		G = min(max(0,G),255);
		B = min(max(0,B),255);
		BYTE minval = min(min(R,G),B);
		BYTE maxval = max(max(R,G),B);
		DWORD msum = maxval+minval;
		L = (BYTE)(0.5+(msum*240/510.));

		if (maxval == minval)
		{
			S = 0;
			H = 160;
		}
		else
		{
			float delta = (float)maxval-minval;

			S = (BYTE)(0.5+(240*(delta/((msum<=255) ? msum : (510-msum)))));

			float h;
        
			if (R == maxval)
				h = 40*((G-B)/delta);
			else if (G == maxval)
				h = 40*(2+(B-R)/delta);
			else if (B == maxval)
				h = 40*(4+(R-G)/delta);

			if (h > 240)
				h -= 240;

			H = (BYTE)(0.5+h);
		}
	}

	static void HSL2RGB(BYTE& R, BYTE& G, BYTE& B, BYTE& _H, BYTE& _S, BYTE& _L)
	{
		_H = min(max(0,_H),239);
		_S = min(max(0,_S),240);
		_L = min(max(0,_L),240);

		float H = _H*360/240.0f;
		float S = _S/240.0f;
		float L = _L/240.0f;

		if (S)
		{
			float rm2;

			if (_L<=120)
				rm2=L+  L*S;  
			else
				rm2=L+S-L*S;
			float rm1 = 2*L-rm2;   
			
			R = ToRGB(rm1, rm2, H+120);   
			G = ToRGB(rm1, rm2, H);
			B = ToRGB(rm1, rm2, H-120);
		}
		else 
		{
			R=G=B=(BYTE)(0.5+L*255);
		}
	}

	static BYTE ToRGB(float rm1, float rm2, float rh)
	{
        if (rh > 360)
			rh -= 360;
		else if (rh < 0) 
			rh += 360;

        if (rh < 60)
			rm1 = rm1+(rm2-rm1)*rh/60;   
		else if (rh < 180)
			rm1 = rm2;
		else if (rh < 240)
			rm1 = rm1+(rm2-rm1)*(240-rh)/60;      

		return (BYTE)(0.5+rm1*255);
	}

	HDC memoryDC;
	HBITMAP memoryBitmap;

	HDC memoryHoverDC;
	HBITMAP memoryHoverBitmap;

	HDC memoryPressedDC;
	HBITMAP memoryPressedBitmap;

	bool useAlphaMap;

	int iconSize;
	int hovericonSize;
	int pressediconSize;

	bool mustUseAlpha;
	bool hovermustUseAlpha;
	bool pressedmustUseAlpha;

	bool transparencyNeeded;

	int borderMethod;
	int hoverborderMethod;
	int pressedborderMethod;

};

#endif // !defined(AFX_ICONTEXTURE_H__F00FC03F_E2C8_422B_A076_4B8994F42945__INCLUDED_)
