/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2006 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "StdAfx.h"
#include "xFont.h"

//Text Alignment
//-----------------------------------------------------
#define ALIGN_LEFT 0
#define ALIGN_TOP 0
#define ALIGN_CENTER 1
#define ALIGN_RIGHT 2
#define ALIGN_BOTTOM 2
//-----------------------------------------------------

NameValuePair CharSetConstants[] =	{
		{ "ANSI_CHARSET", ANSI_CHARSET },
		{ "BALTIC_CHARSET", BALTIC_CHARSET },
		{ "CHINESEBIG5_CHARSET", CHINESEBIG5_CHARSET },
		{ "DEFAULT_CHARSET", DEFAULT_CHARSET },
		{ "EASTEUROPE_CHARSET", EASTEUROPE_CHARSET },
		{ "GB2312_CHARSET", GB2312_CHARSET },
		{ "GREEK_CHARSET", GREEK_CHARSET },
		{ "HANGUL_CHARSET", HANGUL_CHARSET },
		{ "MAC_CHARSET", MAC_CHARSET },
		{ "RUSSIAN_CHARSET", RUSSIAN_CHARSET },
		{ "SHIFTJIS_CHARSET", SHIFTJIS_CHARSET },
		{ "SYMBOL_CHARSET", SYMBOL_CHARSET },
		{ "TURKISH_CHARSET", TURKISH_CHARSET },
		{ "VIETNAMESE_CHARSET", VIETNAMESE_CHARSET },
		{ "JOHAB_CHARSET", JOHAB_CHARSET },
		{ "ARABIC_CHARSET", ARABIC_CHARSET },
		{ "HEBREW_CHARSET", HEBREW_CHARSET },
		{ "THAI_CHARSET", THAI_CHARSET },
		{ "OEM_CHARSET", OEM_CHARSET },
		{ 0, 0 }
	};

NameValuePair alignValues[] = {
		{ "left", DT_LEFT },
		{ "center", DT_CENTER },
		{ "right", DT_RIGHT },
		{ 0, 0 }
	};

NameValuePair vertAlignValues[] = {
		{ "top", ALIGN_TOP },
		{ "center", ALIGN_CENTER },
		{ "bottom", ALIGN_BOTTOM },
		{ 0, 0 }
	};

xFont::xFont() :
	leftBorder(0),
	rightBorder(0),
	topBorder(0),
	bottomBorder(0),

	useHoverText(false),
	usePressedText(false)
{
}

xFont::~xFont()
{
}

void xFont::configure(const string &uniqueKey, const string &groupKey)
{
	//Initialize Basic Helper Strings
	int numTokens = 0;
	string one = "";
	string two = "";
	string three = "";
	string four = "";
	string five = "";
	string six = "";
	string seven = "";

	normal.name = Getx2RCString(uniqueKey, groupKey, "Font", "Arial");
	normal.charset = Getx2RCNamedValue(uniqueKey, groupKey, "FontCharSet", CharSetConstants, DEFAULT_CHARSET);
	normal.height = Getx2RCInt(uniqueKey, groupKey, "FontHeight", 15, 6);
	normal.color = Getx2RCColor(uniqueKey, groupKey, "FontColor", RGB(0, 0, 0));

	normal.bold = Getx2RCBool(uniqueKey, groupKey, "FontBold");
	normal.italic = Getx2RCBool(uniqueKey, groupKey, "FontItalic");
	normal.underline = Getx2RCBool(uniqueKey, groupKey, "FontUnderline");

	normal.shadow = Getx2RCBool(uniqueKey, groupKey, "FontShadow");
	normal.shadowx = Getx2RCInt(uniqueKey, groupKey, "FontShadowX", 1, 0);
	normal.shadowy = Getx2RCInt(uniqueKey, groupKey, "FontShadowY", 1, 0);
	normal.shadowcolor = Getx2RCColor(uniqueKey, groupKey, "FontShadowColor", RGB(128, 128, 128));

	normal.textoutline = Getx2RCBool(uniqueKey, groupKey, "FontOutline");
	normal.textoutlineextra = Getx2RCBool(uniqueKey, groupKey, "FontOutlineExtra");
	normal.textoutlinecolor = Getx2RCColor(uniqueKey, groupKey, "FontOutlineColor", RGB(128, 128, 128));

	//Alignment
	normal.align = Getx2RCNamedValue(uniqueKey, groupKey, "FontAlign", alignValues, DT_CENTER);
	normal.vertalign = Getx2RCNamedValue(uniqueKey, groupKey, "FontVertAlign", vertAlignValues, ALIGN_CENTER);

	normal.smooth = Getx2RCBool(uniqueKey, groupKey, "FontSmoothing", true);
	if ( IsOS( OS_XP ) && normal.smooth )
		normal.cleartype = Getx2RCBool(uniqueKey, groupKey, "FontClearType", true);
	else
		normal.cleartype = false;

	normal.noellipsis = Getx2RCBool(uniqueKey, groupKey, "FontNoEllipsis", true);

	//-------------------------------------------------------------------------------------------------
	fntNormal.SetName(normal.name.c_str());
	fntNormal.SetCharSet(normal.charset);
	fntNormal.SetSize(normal.height);
	fntNormal.SetBold(normal.bold);
	fntNormal.SetItalic(normal.italic);
	fntNormal.SetUnderline(normal.underline);
	fntNormal.SetSmooth(normal.smooth);
	fntNormal.SetClearType(normal.cleartype);
	//-------------------------------------------------------------------------------------------------

		useHoverText = subconfigure(uniqueKey, groupKey, "Hover", hover, normal);

		fntHover.SetName(hover.name.c_str());
		fntHover.SetCharSet(hover.charset);
		fntHover.SetSize(hover.height);
		fntHover.SetBold(hover.bold);
		fntHover.SetItalic(hover.italic);
		fntHover.SetUnderline(hover.underline);
		fntHover.SetSmooth(hover.smooth);
		fntHover.SetClearType(hover.cleartype);

		usePressedText = subconfigure(uniqueKey, groupKey, "Pressed", pressed, hover);
		
		fntPressed.SetName(pressed.name.c_str());
		fntPressed.SetCharSet(pressed.charset);
		fntPressed.SetSize(pressed.height);
		fntPressed.SetBold(pressed.bold);
		fntPressed.SetItalic(pressed.italic);
		fntPressed.SetUnderline(pressed.underline);
		fntPressed.SetSmooth(pressed.smooth);
		fntPressed.SetClearType(pressed.cleartype);

	//-------------------------------------------------------------------------------------------------

	lineBreak = Getx2RCBool(uniqueKey, groupKey, "AutoLineBreak");
	newlineSpace = Getx2RCInt(uniqueKey, groupKey, "NewLineSpace", 0);

	//Borders
	string tmpBorders = Getx2RCLine(uniqueKey, groupKey, "Borders", "");
	if ( !tmpBorders.empty() )
	{
		numTokens = SplitString(tmpBorders, one, two, three, four, five, six, seven);

		if (numTokens == 1)
		{
			leftBorder = rightBorder = topBorder = bottomBorder = atoi(one.c_str());
		}
		else if (numTokens == 2)
		{
			leftBorder = rightBorder = atoi(one.c_str());
			topBorder = bottomBorder = atoi(two.c_str());
		}
		else if (numTokens == 4)
		{
			leftBorder = atoi(one.c_str());
			rightBorder = atoi(two.c_str());
			topBorder = atoi(three.c_str());
			bottomBorder = atoi(four.c_str());	
		}
		
		if ( leftBorder < 0 )
			leftBorder = 0;
		if ( rightBorder < 0 )
			rightBorder = 0;
		if ( topBorder < 0 )
			topBorder = 0;
		if ( bottomBorder < 0 )
			bottomBorder = 0;
	}

	if ( leftBorder == 0 )
		leftBorder = Getx2RCInt(uniqueKey, groupKey, "LeftBorder", 0, 0);
	if ( topBorder == 0 )
		topBorder = Getx2RCInt(uniqueKey, groupKey, "TopBorder", 0, 0);
	if ( rightBorder == 0 )
		rightBorder = Getx2RCInt(uniqueKey, groupKey, "RightBorder", 0, 0);
	if ( bottomBorder == 0 )
		bottomBorder = Getx2RCInt(uniqueKey, groupKey, "BottomBorder", 0, 0);
}

bool xFont::subconfigure(const string &uniqueKey, const string &groupKey, const string &prefixKey, FONTSETTINGS &workSettings, FONTSETTINGS &defaultSettings)
{
	bool difference = false;
	
	workSettings.name = Getx2RCString(uniqueKey, groupKey, (prefixKey + "Font"), defaultSettings.name);
	if ( stricmp(workSettings.name.c_str(), defaultSettings.name.c_str()) != NULL )
		difference = true;

	workSettings.charset = Getx2RCNamedValue(uniqueKey, groupKey, (prefixKey + "FontCharSet"), CharSetConstants, defaultSettings.charset );
	if ( workSettings.charset != defaultSettings.charset )
		difference = true;
	
	workSettings.height = Getx2RCInt(uniqueKey, groupKey, (prefixKey + "FontHeight"), defaultSettings.height, 0);
	if ( workSettings.height != defaultSettings.height )
		difference = true;
	
	workSettings.color = Getx2RCColor(uniqueKey, groupKey, (prefixKey + "FontColor"), defaultSettings.color);
	if ( workSettings.color != defaultSettings.color )
		difference = true;

	workSettings.bold = Getx2RCBool(uniqueKey, groupKey, (prefixKey + "FontBold"), defaultSettings.bold);
	if ( workSettings.bold != defaultSettings.bold )
		difference = true;

	workSettings.italic = Getx2RCBool(uniqueKey, groupKey, (prefixKey + "FontItalic"), defaultSettings.italic);
	if ( workSettings.italic != defaultSettings.italic )
		difference = true;
		
	workSettings.underline = Getx2RCBool(uniqueKey, groupKey, (prefixKey + "FontUnderline"), defaultSettings.underline);
	if ( workSettings.underline != defaultSettings.underline )
		difference = true;

	workSettings.shadow = Getx2RCBool(uniqueKey, groupKey, (prefixKey + "FontShadow"), defaultSettings.shadow);
	if ( workSettings.shadow != defaultSettings.shadow )
		difference = true;

	workSettings.shadowx = Getx2RCInt(uniqueKey, groupKey, (prefixKey + "FontShadowX"), defaultSettings.shadowx);
	if ( workSettings.shadowx != defaultSettings.shadowx )
		difference = true;
		
	workSettings.shadowy = Getx2RCInt(uniqueKey, groupKey, (prefixKey + "FontShadowY"), defaultSettings.shadowy);
	if ( workSettings.shadowy != defaultSettings.shadowy )
		difference = true;
		
	workSettings.shadowcolor = Getx2RCColor(uniqueKey, groupKey, (prefixKey + "FontShadowColor"), defaultSettings.shadowcolor);
	if ( workSettings.shadowcolor != defaultSettings.shadowcolor )
		difference = true;

	workSettings.textoutline = Getx2RCBool(uniqueKey, groupKey, (prefixKey + "FontOutline"), defaultSettings.textoutline);
	if ( workSettings.textoutline != defaultSettings.textoutline )
		difference = true;
	
	workSettings.textoutlineextra = Getx2RCBool(uniqueKey, groupKey, (prefixKey + "FontOutlineExtra"), defaultSettings.textoutlineextra);
	if ( workSettings.textoutlineextra != defaultSettings.textoutlineextra )
		difference = true;
	
	workSettings.textoutlinecolor = Getx2RCColor(uniqueKey, groupKey, (prefixKey + "FontOutlineColor"), defaultSettings.textoutlinecolor);
	if ( workSettings.textoutlinecolor != defaultSettings.textoutlinecolor )
		difference = true;

	workSettings.align = Getx2RCNamedValue(uniqueKey, groupKey, (prefixKey + "FontAlign"), alignValues, defaultSettings.align);
	if ( workSettings.align != defaultSettings.align )
		difference = true;
		
	workSettings.vertalign = Getx2RCNamedValue(uniqueKey, groupKey, (prefixKey + "FontVertAlign"), vertAlignValues, defaultSettings.vertalign);
	if ( workSettings.vertalign != defaultSettings.vertalign )
		difference = true;

	workSettings.smooth = Getx2RCBool(uniqueKey, groupKey, (prefixKey + "FontSmoothing"), defaultSettings.smooth);
	if ( workSettings.smooth != defaultSettings.smooth )
		difference = true;
		
	if ( IsOS( OS_XP ) && workSettings.smooth )
		workSettings.cleartype = Getx2RCBool(uniqueKey, groupKey, (prefixKey + "FontClearType"), defaultSettings.cleartype);
	else
		workSettings.cleartype = false;
	if ( workSettings.cleartype != defaultSettings.cleartype )
		difference = true;

	workSettings.noellipsis = Getx2RCBool(uniqueKey, groupKey, (prefixKey + "FontNoEllipsis"), defaultSettings.noellipsis);
	if ( workSettings.noellipsis != defaultSettings.noellipsis )
		difference = true;

	return difference;
}

void xFont::apply(HDC hDC, int left, int right, int top, int bottom, const string &text, StringList &labelHREFClickRegions, int imageType, bool smoothoff, int setalign, int setvertAlign, bool scrolling)
{
	//Convert into MFC variables
	CDC tmpCDC;
	tmpCDC.Attach(hDC);

	CString strText = text.c_str();

	if ( strText.Find("<br") == 0 || strText.Find("< br") == 0 )
		strText.Insert(0, " ");
	else if ( strText.Find("<p") == 0 || strText.Find("< p") == 0 )
		strText.Insert(0, " ");
	strText.Replace("><br>", "> <br>");
	strText.Replace(">< br>", "> < br>");
	strText.Replace("><p", "> <p");
	strText.Replace(">< p", "> < p");

	COLORREF usedcolor;

	bool usedshadow;
	int usedshadowx;
	int usedshadowy; 
	int usedshadowcolor;

	bool usedtextoutline;
	bool usedtextoutlineextra;
	int usedtextoutlinecolor;

	short usedalign;
	short usedvertalign;

	bool usedsmooth;
	bool usedcleartype;

	bool usednoellipsis;

	CHTMLFont fntUsed;

	if (imageType == 0)
	{
		usedcolor = normal.color;

		usedshadow = normal.shadow;
		usedshadowx = normal.shadowx;
		usedshadowy = normal.shadowy; 
		usedshadowcolor = normal.shadowcolor;

		usedtextoutline = normal.textoutline;
		usedtextoutlineextra = normal.textoutlineextra;
		usedtextoutlinecolor = normal.textoutlinecolor;

		usedalign = normal.align;
		usedvertalign = normal.vertalign;

		usedsmooth = normal.smooth;
		usedcleartype = normal.cleartype;

		usednoellipsis = normal.noellipsis;

		fntUsed = fntNormal;
	}
	else if (imageType == 1)
	{
		usedcolor = hover.color;

		usedshadow = hover.shadow;
		usedshadowx = hover.shadowx;
		usedshadowy = hover.shadowy; 
		usedshadowcolor = hover.shadowcolor;

		usedtextoutline = hover.textoutline;
		usedtextoutlineextra = hover.textoutlineextra;
		usedtextoutlinecolor = hover.textoutlinecolor;

		usedalign = hover.align;
		usedvertalign = hover.vertalign;

		usedsmooth = hover.smooth;
		usedcleartype = hover.cleartype;

		usednoellipsis = hover.noellipsis;

		fntUsed = fntHover;
	}
	else if (imageType == 2)
	{
		usedcolor = pressed.color;

		usedshadow = pressed.shadow;
		usedshadowx = pressed.shadowx;
		usedshadowy = pressed.shadowy; 
		usedshadowcolor = pressed.shadowcolor;

		usedtextoutline = pressed.textoutline;
		usedtextoutlineextra = pressed.textoutlineextra;
		usedtextoutlinecolor = pressed.textoutlinecolor;

		usedalign = pressed.align;
		usedvertalign = pressed.vertalign;

		usedsmooth = pressed.smooth;
		usedcleartype = pressed.cleartype;

		usednoellipsis = pressed.noellipsis;

		fntUsed = fntPressed;
	}

	if ( setalign != -1 )
		usedalign = setalign;

	if ( setvertAlign != -1 )
		usedvertalign = setvertAlign;

	if ( smoothoff )
		fntUsed.SetSmooth(false);

	if ( usednoellipsis )
		scrolling = true;

	CRect rctPosition; 
	rctPosition.left = left;
	rctPosition.right = right;
	rctPosition.top = top;
	rctPosition.bottom = bottom;

	// Call to CHTMLDrawer
	if (usedshadow && !usedtextoutline)
	{
		fntUsed.SetColor(usedshadowcolor);

		CRect r;
		CopyRect(&r, &rctPosition);
		//Right
		r.OffsetRect(usedshadowx, usedshadowy);

		CHTMLDrawer::DrawText(&tmpCDC,
								strText,
								fntUsed,
								r,
								usedalign,
								usedvertalign,
								lineBreak,
								newlineSpace,
								scrolling,
								labelHREFClickRegions,
								true);
	}
	else if (usedtextoutline)
	{
		fntUsed.SetColor(usedtextoutlinecolor);

		if ( !usedtextoutlineextra )
		{
			CRect r;
			CopyRect(&r, &rctPosition);
			//Right
			r.OffsetRect(1, 0);
			CHTMLDrawer::DrawText(&tmpCDC,
									strText,
									fntUsed,
									r,
									usedalign,
									usedvertalign,
									lineBreak,
									newlineSpace,
									scrolling,
									labelHREFClickRegions,
									true);
		
			//Left
			r.OffsetRect(-2, 0);
			CHTMLDrawer::DrawText(&tmpCDC,
									strText,
									fntUsed,
									r,
									usedalign,
									usedvertalign,
									lineBreak,
									newlineSpace,
									scrolling,
									labelHREFClickRegions,
									true);

			//(Center/)Bottom
			r.OffsetRect(1, -1);
			CHTMLDrawer::DrawText(&tmpCDC,
									strText,
									fntUsed,
									r,
									usedalign,
									usedvertalign,
									lineBreak,
									newlineSpace,
									scrolling,
									labelHREFClickRegions,
									true);

			//Top
			r.OffsetRect(0, 2);
			CHTMLDrawer::DrawText(&tmpCDC,
									strText,
									fntUsed,
									r,
									usedalign,
									usedvertalign,
									lineBreak,
									newlineSpace,
									scrolling,
									labelHREFClickRegions,
									true);
		}
		else
		{
			CRect r;
			CopyRect(&r, &rctPosition);
			
			r.OffsetRect(1, 1);
			CHTMLDrawer::DrawText(&tmpCDC,
									strText,
									fntUsed,
									r,
									usedalign,
									usedvertalign,
									lineBreak,
									newlineSpace,
									scrolling,
									labelHREFClickRegions,
									true);

			r.OffsetRect(-1, 0);
			CHTMLDrawer::DrawText(&tmpCDC,
									strText,
									fntUsed,
									r,
									usedalign,
									usedvertalign,
									lineBreak,
									newlineSpace,
									scrolling,
									labelHREFClickRegions,
									true);

			r.OffsetRect(-1, 0);
			CHTMLDrawer::DrawText(&tmpCDC,
									strText,
									fntUsed,
									r,
									usedalign,
									usedvertalign,
									lineBreak,
									newlineSpace,
									scrolling,
									labelHREFClickRegions,
									true);

			r.OffsetRect(0, -1);
			CHTMLDrawer::DrawText(&tmpCDC,
									strText,
									fntUsed,
									r,
									usedalign,
									usedvertalign,
									lineBreak,
									newlineSpace,
									scrolling,
									labelHREFClickRegions,
									true);

			r.OffsetRect(0, -1);
			CHTMLDrawer::DrawText(&tmpCDC,
									strText,
									fntUsed,
									r,
									usedalign,
									usedvertalign,
									lineBreak,
									newlineSpace,
									scrolling,
									labelHREFClickRegions,
									true);

			r.OffsetRect(1, 0);
			CHTMLDrawer::DrawText(&tmpCDC,
									strText,
									fntUsed,
									r,
									usedalign,
									usedvertalign,
									lineBreak,
									newlineSpace,
									scrolling,
									labelHREFClickRegions,
									true);

			r.OffsetRect(1, 0);
			CHTMLDrawer::DrawText(&tmpCDC,
									strText,
									fntUsed,
									r,
									usedalign,
									usedvertalign,
									lineBreak,
									newlineSpace,
									scrolling,
									labelHREFClickRegions,
									true);

			r.OffsetRect(0, 1);
			CHTMLDrawer::DrawText(&tmpCDC,
									strText,
									fntUsed,
									r,
									usedalign,
									usedvertalign,
									lineBreak,
									newlineSpace,
									scrolling,
									labelHREFClickRegions,
									true);
		}
	}

	rctPosition.left = left;
	rctPosition.right = right;
	rctPosition.top = top;
	rctPosition.bottom = bottom;

	fntUsed.SetColor(usedcolor);

	CHTMLDrawer::DrawText(&tmpCDC,
							strText,
							fntUsed,
							rctPosition,
							usedalign,
							usedvertalign,
							lineBreak,
							newlineSpace,
							scrolling,
							labelHREFClickRegions,
							false);

	if ( smoothoff )
		fntUsed.SetSmooth(usedsmooth);

	tmpCDC.Detach();
}

POINT xFont::measure(HDC hDC, int left, int right, const string &text, int imageType)
{
	//Convert into MFC variables
	CDC tmpCDC;
	tmpCDC.Attach(hDC);

	CString strText = text.c_str();

	if ( strText.Find("<br") == 0 || strText.Find("< br") == 0 )
		strText.Insert(0, " ");
	else if ( strText.Find("<p") == 0 || strText.Find("< p") == 0 )
		strText.Insert(0, " ");
	strText.Replace("><br>", "> <br>");
	strText.Replace(">< br>", "> < br>");
	strText.Replace("><p", "> <p");
	strText.Replace(">< p", "> < p");

	short usedalign;
	short usedvertalign;

	CHTMLFont fntUsed;

	if (imageType == 0)
	{
		usedalign = normal.align;
		usedvertalign = normal.vertalign;

		fntUsed = fntNormal;
	}
	else if (imageType == 1)
	{
		usedalign = hover.align;
		usedvertalign = hover.vertalign;

		fntUsed = fntHover;
	}
	else if (imageType == 2)
	{
		usedalign = pressed.align;
		usedvertalign = pressed.vertalign;

		fntUsed = fntPressed;
	}

	CRect rctPosition; 
	if (lineBreak)
	{
		rctPosition.left = left;
		rctPosition.right = right;
	}
	else
	{
		rctPosition.left = 0;
		rctPosition.right = 0;
	}
	rctPosition.top = 0;
	rctPosition.bottom = 0;

	// Call to CHTMLDrawer
	POINT returnValues = CHTMLDrawer::MeasureText(&tmpCDC,
													strText,
													fntUsed,
													rctPosition,
													usedalign,
													usedvertalign,
													lineBreak,
													newlineSpace);

	tmpCDC.Detach();
	
	return returnValues;
}

HFONT xFont::gethFont()
{
	LOGFONT lf;
	memset(&lf, 0, sizeof(LOGFONT));

	StringCchCopy(lf.lfFaceName, LF_FACESIZE, normal.name.c_str());
	lf.lfHeight = normal.height;
	lf.lfItalic = normal.italic ? TRUE : FALSE;
	lf.lfUnderline = normal.underline ? TRUE : FALSE;
	lf.lfWeight = normal.bold ? FW_BOLD : FW_NORMAL;
    lf.lfCharSet = normal.charset;

	return CreateFontIndirect(&lf);
}