/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2006 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "StdAfx.h"
#include "Label.h"
#include "xFont.h"
#include "Texture.h"
#include "SystemInfo.h"
#include "GhostLabel.h"

#include "ImageTexture.h"
#include "SolidTexture.h"

#include "verinfo.h"

#include "shlobj.h"
#include "shlwapi.h"

#define GET_X_LPARAM(lp) ((int)(short)LOWORD(lp))
#define GET_Y_LPARAM(lp) ((int)(short)HIWORD(lp))

//MouseWheel
//-----------------------------------------------------
#ifndef WM_MOUSEWHEEL
	#define WM_MOUSEWHEEL 0x020A
#endif
//-----------------------------------------------------

//XBUTTONS
//-----------------------------------------------------
#ifndef WM_XBUTTONDOWN
  #define WM_XBUTTONDOWN 0x020B
#endif
#ifndef WM_XBUTTONUP
  #define WM_XBUTTONUP 0x020C
#endif

#ifndef WM_XBUTTONDBLCLK
  #define WM_XBUTTONDBLCLK 0x020D
#endif

#ifndef MK_XBUTTON1  
   #define MK_XBUTTON1 0x0020
#endif
#ifndef MK_XBUTTON2  
   #define MK_XBUTTON2 0x0040
#endif
#ifndef XBUTTON1  
   #define XBUTTON1 0x0001
#endif
#ifndef XBUTTON2
  #define XBUTTON2 0x0002
#endif
//-----------------------------------------------------

//Ballon Tooltip
//-----------------------------------------------------
#ifndef TTS_BALLOON
	#define TTS_BALLOON 0x40
#endif
//-----------------------------------------------------

//Timers
//-----------------------------------------------------
#define TIMER_MOUSETRACK 1
#define TIMER_UPDATE 2
#define TIMER_SCROLL 3
#define TIMER_AUTOHIDE 4
#define TIMER_ALPHAFADEIN 5
#define TIMER_ALPHAFADEOUT 6
#define TIMER_FRAMEANIMATION 7
#define TIMER_SCROLLPAUSE 8
#define TIMER_ALPHAFADEUPBANG 15
#define TIMER_ALPHAFADEDOWNBANG 16
//-----------------------------------------------------

//SendMessage
//-----------------------------------------------------
#define LM_SETLABELTEXT 9600
#define LM_SETCURSOR    9610
#define LM_TEXTEDITBOX  9650
//-----------------------------------------------------

//Alpha Transparency
//-----------------------------------------------------
#ifndef WS_EX_LAYERED
	#define WS_EX_LAYERED           0x00080000L
	#define LWA_COLORKEY            0x00000001L
	#define ULW_COLORKEY            0x00000001L
	#define LWA_ALPHA               0x00000002L
	#define ULW_ALPHA               0x00000002L
	#define ULW_OPAQUE              0x00000003L
#endif // WS_EX_LAYERED

#ifndef AC_SRC_ALPHA
	#define AC_SRC_ALPHA            0x00000001L
#endif

typedef BOOL (WINAPI *PULW)(HWND hwnd, HDC hdcDst, POINT *pptDst, SIZE *psize, HDC hdcSrc, POINT *pptSrc, COLORREF crKey, BLENDFUNCTION *pblend, DWORD dwFlags);
typedef DWORD (WINAPI *PSLWA)(HWND, DWORD, BYTE, DWORD);

static PULW pUpdateLayeredWindow = NULL;
static PSLWA pSetLayeredWindowAttributes = NULL;

static BOOL alphaInitialized = FALSE;
//-----------------------------------------------------

extern LRESULT WINAPI TextEditProc(HWND hText, UINT msg, WPARAM wParam, LPARAM lParam);

extern int bitsperpixel;

extern int realMonitors;
extern int minscreenX;
extern int minscreenY;

extern int screenWidth;
extern int screenHeight;

extern LabelList labelList;

Label::Label(const string &name) : 
    actualx(0),actualy(0),actualwidth(0),actualheight(0),
	ParentDestroyNotificationReceived(false), alwaysOnTop(false), focusActive(false), 
	hWnd(0), hInstance(0), box(0), hWndTextBox(0), hTextBGBrush(0), 
	oldbackupBMP(0), backupDC(0), backupBitmap(0), oldscrollBMP(0), scrollDC(0), scrollBitmap(0),
	sender(0),
	resizeBorderResize(true),
	TooltipHints(0),
	cursor(0), textcursor(0), handcursor(0),
	labelInitialized(false),
	disablezOrderChange(true),
	background(0), font(0),	texteditboxbackground(0),
	embossText(0),	
	hasHookedModules(false),
	hoverActive(false),	pressedActive(false), useHover(false), usePressed(false),
	mousePressed(false), mouseInside(false), mouseInside2(false), mouseInsideHREF(false),
	movemodifierkeyPressed(false), moveButtonPressed(0), useCustomMoveArea(false), moveRestriction(0),
	skipButton(false),
	autoHideIsRunning(false),
	hasMoved(false),
	dockedtoSide(-1),
	visible(false),
	dynamicText(false), dynamicTooltip(false),
	text(""), originalTextBackup(""),
	current(0),	currentBackup(0),
	afterClass(""),	afterTitle(""),
	oldWidth(-1), oldHeight(-1), oldscrolltextwidth(0), oldscrolltextheight(0),
	scroll(0),	scrollBackup(1), scrollLimit(0), scrollPosition(0),	frozen(false), skipScrollRepaint(false), restartScrollRepaint(false), scrollPause(0), scrollSteps(0), scrollActualStep(1), scrollStepsInitial(0),
	
	resizeCommand(""),
	moveCommand(""),

	//AlphaFade
	alphaFadeBangStep(2),

	//Animations
	frameCounter(0), loopCounter(-1), normalAnim(false), hoverAnim(false),

	layeredFlag(false),	UpdateAlpha(0),	fakealphaMap(false), alphaFadeStep(2), alphaFadeDelay(10),
	leftBorder(0), rightBorder(0), topBorder(0), bottomBorder(0),

	useresizeBorder(false),	resizeBorderLeft(-1), resizeBorderRight(-1), resizeBorderTop(-1), resizeBorderBottom(-1),

	useTextFade(false), textFadeLeft(-1), textFadeRight(-1), textFadeTop(-1), textFadeBottom(-1),

	//TextEdit Box
	textboxhFont(0), textboxXOrg(""), textboxYOrg(""), textboxWidthOrg(""),	textboxHeight(0),
	textboxVisible(false), tebleftBorder(0), tebrightBorder(0),	tebtopBorder(0), tebbottomBorder(0),
	textboxPaintingMode(0)
{
	this->name = name;

	resizeStepSize.x = 0;
	resizeStepSize.y = 0;
}

Label::~Label()
{
	if (cursor && !cursorPath.empty())
		DestroyCursor(cursor);
	if (handcursor)
		DestroyCursor(handcursor);

	if (hWnd && TooltipHints)
	{
		removeHint(hWnd);
		DestroyWindow(TooltipHints);
		TooltipHints = NULL;
	}

	CleanupPointers();

	if (hWndTextBox)
	{
		if ( textcursor )
		{
			DestroyCursor(textcursor);
			textcursor = NULL;
		}

        // Remove the subclass from the edit control. 
        SetWindowLong(hWndTextBox, GWL_WNDPROC, (LONG) wpOrigEditProc); 

		if ( hTextBGBrush )
		{
			DeleteObject(hTextBGBrush);
			hTextBGBrush = NULL;
		}

		if ( textboxhFont )
		{
			DeleteObject( (HGDIOBJ)(HFONT)textboxhFont );
			textboxhFont = NULL;
		}
		
		DestroyWindow(hWndTextBox);
		hWndTextBox = NULL;
	}

	if(IsWindow(hWnd))
	{
		KillTimer(hWnd, TIMER_MOUSETRACK);
		KillTimer(hWnd, TIMER_UPDATE);
		KillTimer(hWnd, TIMER_SCROLL);
		KillTimer(hWnd, TIMER_AUTOHIDE);
		KillTimer(hWnd, TIMER_ALPHAFADEIN);
		KillTimer(hWnd, TIMER_ALPHAFADEOUT);
		KillTimer(hWnd, TIMER_FRAMEANIMATION);
		KillTimer(hWnd, TIMER_SCROLLPAUSE);
		KillTimer(hWnd, TIMER_ALPHAFADEUPBANG);
		KillTimer(hWnd, TIMER_ALPHAFADEDOWNBANG);

		DestroyWindow(hWnd);
	}

	if ( backupBitmap )
	{
		backupBitmap = (HBITMAP) SelectObject(backupDC, oldbackupBMP);
		DeleteObject(backupBitmap);
	}
	if ( backupDC )
		DeleteObject(backupDC);

	if ( scrollBitmap )
	{
		scrollBitmap = (HBITMAP) SelectObject(scrollDC, oldscrollBMP);
		DeleteObject(scrollBitmap);
	}
	if ( scrollDC )
		DeleteObject(scrollDC);

	if( background )
		delete background;
	
	if( font )
		delete font;

	if ( texteditboxbackground )
		delete texteditboxbackground;
}

void Label::CleanupPointers()
{
	//Delete All GhostLabel Pointers
	for(GhostLabelListIterator glit = ghostlabelList.begin(); glit != ghostlabelList.end(); glit++)
	{
		delete *glit;
		*glit = NULL;
	}
	ghostlabelList.clear();

	//Delete All Label_Region Pointers
	Label_RegionListIterator llit;
	for(llit = labelGhostRegions.begin(); llit != labelGhostRegions.end(); llit++)
	{
		delete *llit;
		*llit = NULL;
	}
	labelGhostRegions.clear();
	
	//Left Click Regions
	for(llit = labelLeftClickRegions.begin(); llit != labelLeftClickRegions.end(); llit++)
	{
		delete *llit;
		*llit = NULL;
	}
	labelLeftClickRegions.clear();
	
	for(llit = labelLeftDoubleClickRegions.begin(); llit != labelLeftDoubleClickRegions.end(); llit++)
	{
		delete *llit;
		*llit = NULL;
	}
	labelLeftDoubleClickRegions.clear();
	
	//Middle Click Regions
	for(llit = labelMiddleClickRegions.begin(); llit != labelMiddleClickRegions.end(); llit++)
	{
		delete *llit;
		*llit = NULL;
	}
	labelMiddleClickRegions.clear();

	for(llit = labelMiddleDoubleClickRegions.begin(); llit != labelMiddleDoubleClickRegions.end(); llit++)
	{
		delete *llit;
		*llit = NULL;
	}
	labelMiddleDoubleClickRegions.clear();

	//Right Click Regions
	for(llit = labelRightClickRegions.begin(); llit != labelRightClickRegions.end(); llit++)
	{
		delete *llit;
		*llit = NULL;
	}
	labelRightClickRegions.clear();

	for(llit = labelRightDoubleClickRegions.begin(); llit != labelRightDoubleClickRegions.end(); llit++)
	{
		delete *llit;
		*llit = NULL;
	}
	labelRightDoubleClickRegions.clear();

	//X1 Click Regions
	for(llit = labelX1ClickRegions.begin(); llit != labelX1ClickRegions.end(); llit++)
	{
		delete *llit;
		*llit = NULL;
	}
	labelX1ClickRegions.clear();

	for(llit = labelX1DoubleClickRegions.begin(); llit != labelX1DoubleClickRegions.end(); llit++)
	{
		delete *llit;
		*llit = NULL;
	}
	labelX1DoubleClickRegions.clear();

	//X2 Click Regions
	for(llit = labelX2ClickRegions.begin(); llit != labelX2ClickRegions.end(); llit++)
	{
		delete *llit;
		*llit = NULL;
	}
	labelX2ClickRegions.clear();

	for(llit = labelX2DoubleClickRegions.begin(); llit != labelX2DoubleClickRegions.end(); llit++)
	{
		delete *llit;
		*llit = NULL;
	}
	labelX2DoubleClickRegions.clear();

	//Tooltip Regions 
	for(llit = labelTooltipRegions.begin(); llit != labelTooltipRegions.end(); llit++)
	{
		delete *llit;
		*llit = NULL;
	}
	labelTooltipRegions.clear();

	//Delete All ModClick Pointers
	//------------------------------------------------------------------------------------------------
	ModClickListIterator mlit;
	// Left
	for(mlit = leftModClickList.begin(); mlit != leftModClickList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	leftModClickList.clear();

	for(mlit = leftModClickUpList.begin(); mlit != leftModClickUpList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	leftModClickUpList.clear();

	for(mlit = leftModClickDownList.begin(); mlit != leftModClickDownList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	leftModClickDownList.clear();

	for(mlit = leftModDoubleClickList.begin(); mlit != leftModDoubleClickList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	leftModDoubleClickList.clear();

	// Middle
	for(mlit = middleModClickList.begin(); mlit != middleModClickList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	middleModClickList.clear();

	for(mlit = middleModClickUpList.begin(); mlit != middleModClickUpList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	middleModClickUpList.clear();

	for(mlit = middleModClickDownList.begin(); mlit != middleModClickDownList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	middleModClickDownList.clear();

	for(mlit = middleModDoubleClickList.begin(); mlit != middleModDoubleClickList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	middleModDoubleClickList.clear();

	// Right
	for(mlit = rightModClickList.begin(); mlit != rightModClickList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	rightModClickList.clear();

	for(mlit = rightModClickUpList.begin(); mlit != rightModClickUpList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	rightModClickUpList.clear();

	for(mlit = rightModClickDownList.begin(); mlit != rightModClickDownList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	rightModClickDownList.clear();

	for(mlit = rightModDoubleClickList.begin(); mlit != rightModDoubleClickList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	rightModDoubleClickList.clear();

	// X1
	for(mlit = X1ModClickList.begin(); mlit != X1ModClickList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	X1ModClickList.clear();

	for(mlit = X1ModClickUpList.begin(); mlit != X1ModClickUpList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	X1ModClickUpList.clear();

	for(mlit = X1ModClickDownList.begin(); mlit != X1ModClickDownList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	X1ModClickDownList.clear();

	for(mlit = X1ModDoubleClickList.begin(); mlit != X1ModDoubleClickList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	X1ModDoubleClickList.clear();

	// X2
	for(mlit = X2ModClickList.begin(); mlit != X2ModClickList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	X2ModClickList.clear();

	for(mlit = X2ModClickUpList.begin(); mlit != X2ModClickUpList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	X2ModClickUpList.clear();

	for(mlit = X2ModClickDownList.begin(); mlit != X2ModClickDownList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	X2ModClickDownList.clear();

	for(mlit = X2ModDoubleClickList.begin(); mlit != X2ModDoubleClickList.end(); mlit++)
	{
		delete *mlit;
		*mlit = NULL;
	}
	X2ModDoubleClickList.clear();
}

void Label::load(HINSTANCE hInstance, HWND box, bool dropShadow)
{
	this->hInstance = hInstance;
	
	if (IsWindow(box))
	{
		this->box = box;
	}
	else
	{
		this->box = NULL;
	}
	
	if (IsWindow(hWnd))	// window already exists...
		return;

	if (!IsOS(OS_XP))
		dropShadow = false;

	hWnd = CreateWindowEx( (box ? 0 : WS_EX_TOOLWINDOW),
							(dropShadow ? "xLabelXP" : "xLabel"),
							name.c_str(),
							(box ? WS_CHILD : WS_POPUP) | WS_CLIPCHILDREN,
							actualx, actualy,		
							actualwidth, actualheight,
							(box ? box : GetLitestepWnd()), 
							0,
							hInstance,
							this);
		
	if (hWnd)
	{
		PrefixedLSSetVariable(name, "CurrentHWND", IntToString( (int) hWnd ));

		SetWindowLong(hWnd, GWL_USERDATA, MAGIC_DWORD);

		reconfigure();
	}
	else
	{
		delete this;
	}
}

void Label::reconfigure(bool refresh)
{
	NameValuePair keyValues[] = {
		{ ".none", -1 },
		{ "ctrl", 1 },
		{ "shift", 2 },
		{ "ctrl+shift", 3 },
		{ "shift+ctrl", 3 },
		{ 0, 0 }
	};

	NameValuePair buttonValues[] = {
		{ "left", 1 },
		{ "right", 2 },
		{ "middle", 3 },
		{ 0, 0 }
	};

	NameValuePair moverestrictionValues[] = {
		{ "horizontal", 1},
		{ "vertical", 2},
		{ 0, 0 }
	};

	NameValuePair autoWidthValues[] = {
		{ "left", 1 },
		{ "right", 2 },
		{ "center", 3 },
		{ 0, 0 }
	};

	NameValuePair autoHeightValues[] = {
		{ "top", 1 },
		{ "bottom", 2 },
		{ "center", 3 },
		{ 0, 0 }
	};

	NameValuePair transparencyValues[] = {
		{ "true", 1 },
		{ "real", 1 },
		{ "fake", 2 },
		{ "false", 3 },
		{ "off", 3 },
		{ "none", 3 },
		{ ".none", 3 },
		{ 0, 0 }
	};

	NameValuePair scrollValues[] = {
		{ "false", -1},
		{ "true", 1},
		{ "horizontal", 1},
		{ "vertical", 2},
		{ "horizontal-left", 1},
		{ "vertical-up", 2},
		{ "horizontal-right", 3},
		{ "vertical-down", 4},
		{ 0, 0 }
	};

	NameValuePair dockValues[] = {
		{ "nothing", 0},
		{ "none", 0},
		{ "moving", 1},
		{ "snappedto", 2},
		{ "moving+snappedto", 3},
		{ "snappedto+moving", 3},
		{ 0, 0 }
	};

	//Initialize Basic Helper Strings
	int numTokens = 0;
	string one = "";
	string two = "";
	string three = "";
	string four = "";
	string five = "";
	string six = "";
	string seven = "";

	SetWindowRgn( hWnd, NULL, true );

	if ( refresh ) 
	{
		//For AutoWidth
		oldWidth = -1;
		oldHeight = -1;

		//Erase all Pointers on Refresh (they are created later new)
		CleanupPointers();
	}

	labelInitialized = true;
	
	//LabelGroup Settings
	labelGroup = GetxRCString(name, "AddToGroup", "");

	//Module Hook (only on first Start!)
	if ( !refresh )
	{
		moduleHook = GetxRCLineList( ("*" + name + "ModuleHook") );

		if ( !moduleHook.empty() )
			hasHookedModules = true;
	}

	//--------------------------------------------------------------------------------
	//Internal Sizing Settings
	//--------------------------------------------------------------------------------
	int maxX = 0;
	int maxY = 0;
	if (box != NULL)
	{
		RECT rc;
		GetClientRect(box, &rc);
		maxX = rc.right;
		maxY = rc.bottom;
	}
	else
	{
		maxX = screenWidth;
		maxY = screenHeight;
	}

	//AutoSize
	autoWidthMode = Getx2RCNamedValue(name, labelGroup, "AutoWidthMode", autoWidthValues, 0);
	autoHeightMode = Getx2RCNamedValue(name, labelGroup, "AutoHeightMode", autoHeightValues, 0);
	autoMinWidth = Getx2RCInt(name, labelGroup, "AutoMinWidth", 0, 0, maxX);
	autoMinHeight = Getx2RCInt(name, labelGroup, "AutoMinHeight", 0, 0, maxY);
	autoMaxWidth = Getx2RCDimen(name, labelGroup, "AutoMaxWidth", maxX, maxX, box, false);
	autoMaxHeight = Getx2RCDimen(name, labelGroup, "AutoMaxHeight", maxY, maxY, box, true);

	numTokens = SplitString(Getx2RCLine(name, labelGroup, "AutoAnimation", "0 0"), one, two, three, four, five, six, seven);
	
	if (numTokens >= 2)
	{
		autoAnimSteps = atoi( one.c_str() );
		autoAnimDelay = atoi( two.c_str() );

		if ( autoAnimSteps <= 0 || autoAnimDelay <= 0 )
		{
			autoAnimSteps = 0;
			autoAnimDelay = 0;
		}
		
		autoAnimDelay = min(autoAnimDelay,50);
	}
	else
	{
		autoAnimSteps = 0;
		autoAnimDelay = 0;
	}

	if ( autoMinWidth > autoMaxWidth )
		autoMinWidth = autoMaxWidth;
	if ( autoMinHeight > autoMaxHeight )
		autoMinHeight = autoMaxHeight;

	//Resizable Border
	//--> Affected by Resizable Border (must be set before!)
	fireOnMoveDuringMove = Getx2RCBool(name, labelGroup, "FireOnMoveDuringMove");
	fireOnResizeDuringResize = Getx2RCBool(name, labelGroup, "FireOnResizeDuringResize");
	//-->

	//Use Split Helper Strings
	numTokens = SplitString(Getx2RCLine(name, labelGroup, "ResizeBorder", "-1"), one, two, three, four, five, six, seven);
	
	if (numTokens == 1)
	{
		resizeBorderLeft = resizeBorderRight = resizeBorderTop = resizeBorderBottom = atoi(one.c_str());
	}
	else if (numTokens == 2)
	{
		resizeBorderLeft = resizeBorderRight = atoi(one.c_str());
		resizeBorderTop = resizeBorderBottom = atoi(two.c_str());
	}
	else if (numTokens == 4)
	{
		resizeBorderLeft = atoi(one.c_str());
		resizeBorderRight = atoi(two.c_str());
		resizeBorderTop = atoi(three.c_str());
		resizeBorderBottom = atoi(four.c_str());	
	}

	if ( resizeBorderLeft <= 0 )
		resizeBorderLeft = -1;
	if ( resizeBorderRight <= 0 )
		resizeBorderRight = -1;
	if ( resizeBorderTop <= 0 )
		resizeBorderTop = -1;
	if ( resizeBorderBottom <= 0 )
		resizeBorderBottom = -1;

	if ( resizeBorderLeft > -1 || resizeBorderRight > -1 || resizeBorderTop > -1 || resizeBorderBottom > -1 )
	{
		useresizeBorder = true;
		fireOnResizeDuringResize = true;

		//Use Split Helper Strings
		numTokens = SplitString(Getx2RCLine(name, labelGroup, "ResizeBorderSteps", "0"), one, two, three, four, five, six, seven);
	
		if (numTokens == 1)
		{
			resizeStepSize.x = resizeStepSize.y = atoi(one.c_str());
		}
		else if (numTokens == 2)
		{
			resizeStepSize.x = atoi(one.c_str());
			resizeStepSize.y = atoi(two.c_str());
		}
	}

	//--------------------------------------------------------------------------------
	//Image Setup
	//--------------------------------------------------------------------------------
	
	//Background Painting Class
	if ( background )
	{
		delete background;
		background = NULL;
	}
	background = GetRCTexture(name, labelGroup);	
	//Internal Helper
	transparencyNeeded = background->isTransparencyNeeded();

	//Hover/Pressed Images Activation
	if ( background->useHoverState() )
		useHover = true;
	
	if ( background->usePressedState() )
		usePressed = true;

	//Normal Animation
	imageLoop = GetxRCInt(name, "FrameLoops", -1);
	imageFrameDelay = GetxRCInt(name, "FrameDelay", 100, 1);
	imageFrameCount = GetxRCInt(name, "FrameCount", 0, 0);
	if (imageFrameCount != 0)
		normalAnim = true;
	
	//Hover Animation
	hoverimageLoop = GetxRCInt(name, "HoverFrameLoops", -1);
	hoverimageFrameDelay = GetxRCInt(name, "HoverFrameDelay", 100, 1);
	hoverimageFrameCount = GetxRCInt(name, "HoverFrameCount", 0, 0);
	if (hoverimageFrameCount != 0)
		hoverAnim = true;

	StringList extraPainting;
	//Create Extra Painting Background IMAGE
	extraPainting = GetxRCLineList( ("*" + name + "OverlayLabel") );
	extraPainting.merge( GetxRCLineList( ("*" + labelGroup + "OverlayLabel") ) );
	one = "*";
	one.append( DEFAULT_PREFIX );
	extraPainting.merge( GetxRCLineList( (one + "OverlayLabel") ) );
	
	StringListIterator sit;
	for(sit = extraPainting.begin(); sit != extraPainting.end(); sit++)
	{
		//Use Split Helper Strings
		numTokens = SplitString((*sit), one, two, three, four, five, six, seven);
		
		if ( numTokens >= 5 )
		{
			GhostLabel* newimage = new GhostLabel();
		
			newimage->orgx = one;
			newimage->orgy = two;
			
			newimage->x = xParseCoord(one, 0, actualwidth, hWnd);
			newimage->y = xParseCoord(two, 0, actualheight, hWnd);

			newimage->orgwidth = three;
			newimage->orgheight = four;
			
			newimage->width = xParseDimen(three, 0, actualwidth, hWnd);
			newimage->height = xParseDimen(four, 0, actualheight, hWnd);

			newimage->name = five;
			newimage->labelGroup = GetxRCString(five, "AddToGroup", "");
		
			if ( numTokens >= 6 )
			{
				if ( IsOS( OS_2KXP ) )
					newimage->alphamap = (strchr(six.c_str(), 'a') != 0 || strchr(six.c_str(), 'A') != 0);

				newimage->normal = (strchr(six.c_str(), 'n') != 0 || strchr(six.c_str(), 'N') != 0);
				newimage->hover = (strchr(six.c_str(), 'h') != 0 || strchr(six.c_str(), 'H') != 0);
				newimage->pressed = (strchr(six.c_str(), 'p') != 0 || strchr(six.c_str(), 'P') != 0);
				newimage->standalone = (strchr(six.c_str(), 's') != 0 || strchr(six.c_str(), 'S') != 0);
			
				if ( numTokens == 7 && IsOS( OS_2KXP ) )
				{
					if ( strstr(seven.c_str(), ":") != NULL )
					{
						int tmpalpha = 255;
						CString splitter = seven.c_str();
						tmpalpha = atoi( splitter.Left(splitter.Find(":")) );
						newimage->alpha = tmpalpha;
						splitter = splitter.Right(splitter.GetLength()-(splitter.Find(":")+1));
						if ( strstr(splitter, ":") != NULL )
						{
							tmpalpha = atoi( splitter.Left(splitter.Find(":")) );
							newimage->hoveralpha = tmpalpha;
						
							tmpalpha = atoi( splitter.Right(splitter.GetLength()-(splitter.Find(":")+1)) );
							newimage->pressedalpha = tmpalpha;
						}
						else
						{
							tmpalpha = atoi( splitter );
							newimage->hoveralpha = tmpalpha;
							newimage->pressedalpha = tmpalpha;
						}
					}
					else
					{
						int tmpalpha = atoi( seven.c_str() );
						if (tmpalpha < 0 || tmpalpha > 255 || !IsOS( OS_2KXP ) )
						{
							newimage->alpha = 255;
							newimage->hoveralpha = 255;
							newimage->pressedalpha = 255;
						}
						else
						{
							newimage->alpha = tmpalpha;
							newimage->hoveralpha = tmpalpha;
							newimage->pressedalpha = tmpalpha;
						}
					}
				}
			}

			if ( newimage->standalone )
			{
				Label_Region *insert = new Label_Region();

				insert->orgx = one;
				insert->orgy = two;
				insert->orgwidth = three;
				insert->orgheight = four;

				insert->ghostlabel = newimage;

				labelGhostRegions.insert(labelGhostRegions.end(), insert);

				string tmpleftClickCommand = Getx2RCLine(five, newimage->labelGroup, "OnLeftClick", "");
				string tmpleftDoubleClickCommand = Getx2RCLine(five, newimage->labelGroup, "OnLeftDoubleClick", "");
				
				string tmpmiddleClickCommand = Getx2RCLine(five, newimage->labelGroup, "OnMiddleClick", "");
				string tmpmiddleDoubleClickCommand = Getx2RCLine(five, newimage->labelGroup, "OnMiddleDoubleClick", "");
				
				string tmprightClickCommand = Getx2RCLine(five, newimage->labelGroup, "OnRightClick", "");
				string tmprightDoubleClickCommand = Getx2RCLine(five, newimage->labelGroup, "OnRightDoubleClick", "");

				string tmpX1ClickCommand = Getx2RCLine(five, newimage->labelGroup, "OnX1Click", "");
				string tmpX1DoubleClickCommand = Getx2RCLine(five, newimage->labelGroup, "OnX1DoubleClick", "");

				string tmpX2ClickCommand = Getx2RCLine(five, newimage->labelGroup, "OnX2Click", "");
				string tmpX2DoubleClickCommand = Getx2RCLine(five, newimage->labelGroup, "OnX2DoubleClick", "");

				if ( !tmpleftClickCommand.empty() )
				{
					Label_Region *insertl = new Label_Region();

					insertl->orgx = one;
					insertl->orgy = two;
					insertl->orgwidth = three;
					insertl->orgheight = four;

					insertl->ghostlabel = newimage;

					insertl->action = tmpleftClickCommand;
				
					labelLeftClickRegions.insert(labelLeftClickRegions.end(), insertl);
				}
				if ( !tmpleftDoubleClickCommand.empty() )
				{
					Label_Region *insertld = new Label_Region();

					insertld->orgx = one;
					insertld->orgy = two;
					insertld->orgwidth = three;
					insertld->orgheight = four;

					insertld->ghostlabel = newimage;

					insertld->action = tmpleftDoubleClickCommand;
				
					labelLeftDoubleClickRegions.insert(labelLeftDoubleClickRegions.end(), insertld);
				}
				if ( !tmpmiddleClickCommand.empty() )
				{
					Label_Region *insertm = new Label_Region();

					insertm->orgx = one;
					insertm->orgy = two;
					insertm->orgwidth = three;
					insertm->orgheight = four;

					insertm->ghostlabel = newimage;

					insertm->action = tmpmiddleClickCommand;
			
					labelMiddleClickRegions.insert(labelMiddleClickRegions.end(), insertm);
				}
				if ( !tmpmiddleDoubleClickCommand.empty() )
				{
					Label_Region *insertmd = new Label_Region();

					insertmd->orgx = one;
					insertmd->orgy = two;
					insertmd->orgwidth = three;
					insertmd->orgheight = four;

					insertmd->ghostlabel = newimage;

					insertmd->action = tmpmiddleDoubleClickCommand;
			
					labelMiddleDoubleClickRegions.insert(labelMiddleDoubleClickRegions.end(), insertmd);
				}
				if ( !tmprightClickCommand.empty() )
				{
					Label_Region *insertr = new Label_Region();

					insertr->orgx = one;
					insertr->orgy = two;
					insertr->orgwidth = three;
					insertr->orgheight = four;

					insertr->ghostlabel = newimage;

					insertr->action = tmprightClickCommand;
			
					labelRightClickRegions.insert(labelRightClickRegions.end(), insertr);
				}
				if ( !tmprightDoubleClickCommand.empty() )
				{
					Label_Region *insertrd = new Label_Region();

					insertrd->orgx = one;
					insertrd->orgy = two;
					insertrd->orgwidth = three;
					insertrd->orgheight = four;

					insertrd->ghostlabel = newimage;

					insertrd->action = tmprightDoubleClickCommand;
		
					labelRightDoubleClickRegions.insert(labelRightDoubleClickRegions.end(), insertrd);
				}
				if ( !tmpX1ClickCommand.empty() )
				{
					Label_Region *insertX1 = new Label_Region();

					insertX1->orgx = one;
					insertX1->orgy = two;
					insertX1->orgwidth = three;
					insertX1->orgheight = four;

					insertX1->ghostlabel = newimage;

					insertX1->action = tmpX1ClickCommand;
			
					labelX1ClickRegions.insert(labelX1ClickRegions.end(), insertX1);
				}
				if ( !tmpX1DoubleClickCommand.empty() )
				{
					Label_Region *insertX1d = new Label_Region();

					insertX1d->orgx = one;
					insertX1d->orgy = two;
					insertX1d->orgwidth = three;
					insertX1d->orgheight = four;

					insertX1d->ghostlabel = newimage;

					insertX1d->action = tmpX1DoubleClickCommand;
		
					labelX1DoubleClickRegions.insert(labelX1DoubleClickRegions.end(), insertX1d);
				}
				if ( !tmpX2ClickCommand.empty() )
				{
					Label_Region *insertX2 = new Label_Region();

					insertX2->orgx = one;
					insertX2->orgy = two;
					insertX2->orgwidth = three;
					insertX2->orgheight = four;

					insertX2->ghostlabel = newimage;

					insertX2->action = tmpX2ClickCommand;
			
					labelX2ClickRegions.insert(labelX2ClickRegions.end(), insertX2);
				}
				if ( !tmpX2DoubleClickCommand.empty() )
				{
					Label_Region *insertX2d = new Label_Region();

					insertX2d->orgx = one;
					insertX2d->orgy = two;
					insertX2d->orgwidth = three;
					insertX2d->orgheight = four;

					insertX2d->ghostlabel = newimage;

					insertX2d->action = tmpX2DoubleClickCommand;
			
					labelX2DoubleClickRegions.insert(labelX2DoubleClickRegions.end(), insertX2d);
				}
			}

			//Secure Functions
			//------------------------------------------------------------------------
			if ( !newimage->normal && !newimage->hover && !newimage->pressed )
			{
				newimage->normal = true;
				newimage->hover = true;
				newimage->pressed = true;
			}

			newimage->background = GetRCTexture(five, newimage->labelGroup, true/*External Configure!*/);
			if ( !newimage->alphamap && newimage->background->isTextureType() == 1 && IsOS( OS_2KXP) )
				newimage->alphamap = true;

			newimage->background->configure(five, newimage->labelGroup, newimage->alphamap);

			//Define useHover and usePressed
			if ( !newimage->standalone && newimage->background->useHoverState() )
				useHover = true;
			if ( !newimage->standalone && newimage->background->usePressedState() )
				usePressed = true;

			newimage->originaltext = Getx2RCString(five, newimage->labelGroup, "Text", "");

			if ( !newimage->originaltext.empty() )
				newimage->font = GetRCFont(five, newimage->labelGroup);

			newimage->enterCommand = Getx2RCLine(five, newimage->labelGroup, "OnEnter", "");
			newimage->leaveCommand = Getx2RCLine(five, newimage->labelGroup, "OnLeave", "");

			newimage->visible = !Getx2RCBool(five, newimage->labelGroup, "StartHidden");

			if ( newimage->background != 0 )
				ghostlabelList.insert(ghostlabelList.end(), newimage);
			else
				delete newimage;
		}
	}

	//--------------------------------------------------------------------------------
	//Text/Font Setup
	//--------------------------------------------------------------------------------
	
	//Font Painting Class
	if ( font )
	{
		delete font;
		font = NULL;
	}
	font = GetRCFont(name, labelGroup);

	leftBorder = font->leftBorder;
	rightBorder = font->rightBorder;
	topBorder = font->topBorder;
	bottomBorder = font->bottomBorder;

	//Use Split Helper Strings
	string tmpemboss = Getx2RCLine(name, labelGroup, "FontEmboss", "false");
	if ( _stricmp(tmpemboss.c_str(), "false") != NULL )
	{
		while ( strstr( tmpemboss.c_str(), "#" ) != NULL )
			tmpemboss.replace(tmpemboss.find("#"), 1, "", 0);

		numTokens = SplitString(tmpemboss, one, two, three, four, five, six, seven);

		int color;
		if ( numTokens >= 1 )
		{
			if ( _stricmp( one.c_str(), "sunken") == NULL )
				embossText = -1;
			else
				embossText = 1;

			if ( numTokens >= 2 )
			{
				color = strtol(two.c_str(), NULL, 16);
				lightembossColor = darkembossColor = RGB(GetBValue(color), GetGValue(color), GetRValue(color));

				if ( numTokens >= 3 )
				{
					color = strtol(three.c_str(), NULL, 16);
					darkembossColor = RGB(GetBValue(color), GetGValue(color), GetRValue(color));
				}
			}
			else
			{
				lightembossColor = GetSysColor( COLOR_BTNHIGHLIGHT );
				darkembossColor = GetSysColor( COLOR_BTNSHADOW );
			}
		}
	}

	//Scrolling
	scrollPadLength = Getx2RCInt(name, labelGroup, "ScrollPad", 10);
	scrollInterval = Getx2RCInt(name, labelGroup, "ScrollInterval", 100, 50);
	scrollSpeed = Getx2RCInt(name, labelGroup, "ScrollSpeed", 1, 1);
	
	if (!refresh)
	{
		int tmpScroll = Getx2RCNamedValue(name, labelGroup, "Scroll", scrollValues, 0);
		
		//Compatibility Mode
		if ( tmpScroll == 0 && Getx2RCBool(name, labelGroup, "Scroll") )
			tmpScroll = 1;
		if ( tmpScroll == -1 )
			tmpScroll = 0;

		setScrolling(tmpScroll);
	}

	if ( scroll == 2 || scroll == 4 )
	{
		//Use Split Helper Strings
		numTokens = SplitString(Getx2RCLine(name, labelGroup, "ScrollPerLine", ""), one, two, three, four, five, six, seven);

		if (numTokens == 3)
		{
			scrollStepsInitial = atoi(one.c_str());
			scrollSteps = atoi(two.c_str());
			scrollPause = atoi(three.c_str());
		}
	}

	//Text Updates
	updateInterval = Getx2RCInt(name, labelGroup, "UpdateInterval", 1000, 250);
	alwaysUpdateContent = Getx2RCBool(name, labelGroup, "UpdateAlways");

	//Extra Settings
	noEscape = Getx2RCBool(name, labelGroup, "NoEscapeSequence");
	noTextSplit = Getx2RCBool(name, labelGroup, "NoTextSplit");
	hrefconfig = split(Getx2RCLine(name, labelGroup, "HrefHoverConfig", ""), ";");

	//TextFade Settings
	//Use Split Helper Strings
	if ( autoWidthMode == 0 && autoHeightMode == 0 && IsOS(OS_2KXP) )
	{
		numTokens = SplitString(Getx2RCLine(name, labelGroup, "TextFade", "0"), one, two, three, four, five, six, seven);
		
		if (numTokens == 1)
		{
			textFadeLeft = textFadeRight = textFadeTop = textFadeBottom = atoi(one.c_str());
		}
		else if (numTokens == 2)
		{
			textFadeLeft = textFadeRight = atoi(one.c_str());
			textFadeTop = textFadeBottom = atoi(two.c_str());
		}
		else if (numTokens == 4)
		{
			textFadeLeft = atoi(one.c_str());
			textFadeRight = atoi(two.c_str());
			textFadeTop = atoi(three.c_str());
			textFadeBottom = atoi(four.c_str());
		}

		if ( textFadeLeft <= 0 )
			textFadeLeft = 0;

		if ( textFadeRight <= 0 )
			textFadeRight = 0;

		if ( textFadeTop <= 0 )
			textFadeTop = 0;

		if ( textFadeBottom <= 0 )
			textFadeBottom = 0;

		if ( textFadeLeft > 0 || textFadeRight > 0 || textFadeTop > 0 || textFadeBottom > 0 )
			useTextFade = true;
	}

	//--------------------------------------------------------------------------------
	//Event/Action Setup
	//--------------------------------------------------------------------------------

	leftClickCommand = Getx2RCLine(name, labelGroup, "OnLeftClick", "");
	leftClickDownCommand = Getx2RCLine(name, labelGroup, "OnLeftClickDown", "");
	leftClickUpCommand = Getx2RCLine(name, labelGroup, "OnLeftClickUp", "");
	leftDoubleClickCommand = Getx2RCLine(name, labelGroup, "OnLeftDoubleClick", "");

	middleClickCommand = Getx2RCLine(name, labelGroup, "OnMiddleClick", "");
	middleClickDownCommand = Getx2RCLine(name, labelGroup, "OnMiddleClickDown", "");
	middleClickUpCommand = Getx2RCLine(name, labelGroup, "OnMiddleClickUp", "");
	middleDoubleClickCommand = Getx2RCLine(name, labelGroup, "OnMiddleDoubleClick", "");

	rightClickCommand = Getx2RCLine(name, labelGroup, "OnRightClick", "");
	rightClickDownCommand = Getx2RCLine(name, labelGroup, "OnRightClickDown", "");
	rightClickUpCommand = Getx2RCLine(name, labelGroup, "OnRightClickUp", "");
	rightDoubleClickCommand = Getx2RCLine(name, labelGroup, "OnRightDoubleClick", "");

	X1ClickCommand = Getx2RCLine(name, labelGroup, "OnX1Click", "");
	X1ClickDownCommand = Getx2RCLine(name, labelGroup, "OnX1ClickDown", "");
	X1ClickUpCommand = Getx2RCLine(name, labelGroup, "OnX1ClickUp", "");
	X1DoubleClickCommand = Getx2RCLine(name, labelGroup, "OnX1DoubleClick", "");

	X2ClickCommand = Getx2RCLine(name, labelGroup, "OnX2Click", "");
	X2ClickDownCommand = Getx2RCLine(name, labelGroup, "OnX2ClickDown", "");
	X2ClickUpCommand = Getx2RCLine(name, labelGroup, "OnX2ClickUp", "");
	X2DoubleClickCommand = Getx2RCLine(name, labelGroup, "OnX2DoubleClick", "");

	StringList ExtraModClicks;
	one = "*";
	one.append( DEFAULT_PREFIX );

	//Left
	ExtraModClicks.clear();
	ExtraModClicks = GetxRCLineList( ("*" + name + "OnLeftClick") );
	ExtraModClicks.merge( GetxRCLineList( ("*" + labelGroup + "OnLeftClick") ) );
	ExtraModClicks.merge( GetxRCLineList( (one + "OnLeftClick") ) );
	FillModClickList(ExtraModClicks, leftModClickList);

	ExtraModClicks.clear();
	ExtraModClicks = GetxRCLineList( ("*" + name + "OnLeftClickUp") );
	ExtraModClicks.merge( GetxRCLineList( ("*" + labelGroup + "OnLeftClickUp") ) );
	ExtraModClicks.merge( GetxRCLineList( (one + "OnLeftClickUp") ) );
	FillModClickList(ExtraModClicks, leftModClickUpList);

	ExtraModClicks.clear();
	ExtraModClicks = GetxRCLineList( ("*" + name + "OnLeftClickDown") );
	ExtraModClicks.merge( GetxRCLineList( ("*" + labelGroup + "OnLeftClickDown") ) );
	ExtraModClicks.merge( GetxRCLineList( (one + "OnLeftClickDown") ) );
	FillModClickList(ExtraModClicks, leftModClickDownList);

	ExtraModClicks.clear();
	ExtraModClicks = GetxRCLineList( ("*" + name + "OnLeftDoubleClick") );
	ExtraModClicks.merge( GetxRCLineList( ("*" + labelGroup + "OnLeftDoubleClick") ) );
	ExtraModClicks.merge( GetxRCLineList( (one + "OnLeftDoubleClick") ) );
	FillModClickList(ExtraModClicks, leftModDoubleClickList);

	//Middle
	ExtraModClicks.clear();
	ExtraModClicks = GetxRCLineList( ("*" + name + "OnMiddleClick") );
	ExtraModClicks.merge( GetxRCLineList( ("*" + labelGroup + "OnMiddleClick") ) );
	ExtraModClicks.merge( GetxRCLineList( (one + "OnMiddleClick") ) );
	FillModClickList(ExtraModClicks, middleModClickList);

	ExtraModClicks.clear();
	ExtraModClicks = GetxRCLineList( ("*" + name + "OnMiddleClickUp") );
	ExtraModClicks.merge( GetxRCLineList( ("*" + labelGroup + "OnMiddleClickUp") ) );
	ExtraModClicks.merge( GetxRCLineList( (one + "OnMiddleClickUp") ) );
	FillModClickList(ExtraModClicks, middleModClickUpList);

	ExtraModClicks.clear();
	ExtraModClicks = GetxRCLineList( ("*" + name + "OnMiddleClickDown") );
	ExtraModClicks.merge( GetxRCLineList( ("*" + labelGroup + "OnMiddleClickDown") ) );
	ExtraModClicks.merge( GetxRCLineList( (one + "OnMiddleClickDown") ) );
	FillModClickList(ExtraModClicks, middleModClickDownList);

	ExtraModClicks.clear();
	ExtraModClicks = GetxRCLineList( ("*" + name + "OnMiddleDoubleClick") );
	ExtraModClicks.merge( GetxRCLineList( ("*" + labelGroup + "OnMiddleDoubleClick") ) );
	ExtraModClicks.merge( GetxRCLineList( (one + "OnMiddleDoubleClick") ) );
	FillModClickList(ExtraModClicks, middleModDoubleClickList);

	//Right
	ExtraModClicks.clear();
	ExtraModClicks = GetxRCLineList( ("*" + name + "OnRightClick") );
	ExtraModClicks.merge( GetxRCLineList( ("*" + labelGroup + "OnRightClick") ) );
	ExtraModClicks.merge( GetxRCLineList( (one + "OnRightClick") ) );
	FillModClickList(ExtraModClicks, rightModClickList);

	ExtraModClicks.clear();
	ExtraModClicks = GetxRCLineList( ("*" + name + "OnRightClickUp") );
	ExtraModClicks.merge( GetxRCLineList( ("*" + labelGroup + "OnRightClickUp") ) );
	ExtraModClicks.merge( GetxRCLineList( (one + "OnRightClickUp") ) );
	FillModClickList(ExtraModClicks, rightModClickUpList);

	ExtraModClicks.clear();
	ExtraModClicks = GetxRCLineList( ("*" + name + "OnRightClickDown") );
	ExtraModClicks.merge( GetxRCLineList( ("*" + labelGroup + "OnRightClickDown") ) );
	ExtraModClicks.merge( GetxRCLineList( (one + "OnRightClickDown") ) );
	FillModClickList(ExtraModClicks, rightModClickDownList);

	ExtraModClicks.clear();
	ExtraModClicks = GetxRCLineList( ("*" + name + "OnRightDoubleClick") );
	ExtraModClicks.merge( GetxRCLineList( ("*" + labelGroup + "OnRightDoubleClick") ) );
	ExtraModClicks.merge( GetxRCLineList( (one + "OnRightDoubleClick") ) );
	FillModClickList(ExtraModClicks, rightModDoubleClickList);

	// X1
	ExtraModClicks.clear();
	ExtraModClicks = GetxRCLineList( ("*" + name + "OnX1Click") );
	ExtraModClicks.merge( GetxRCLineList( ("*" + labelGroup + "OnX1Click") ) );
	ExtraModClicks.merge( GetxRCLineList( (one + "OnX1Click") ) );
	FillModClickList(ExtraModClicks, X1ModClickList);

	ExtraModClicks.clear();
	ExtraModClicks = GetxRCLineList( ("*" + name + "OnX1ClickUp") );
	ExtraModClicks.merge( GetxRCLineList( ("*" + labelGroup + "OnX1ClickUp") ) );
	ExtraModClicks.merge( GetxRCLineList( (one + "OnX1ClickUp") ) );
	FillModClickList(ExtraModClicks, X1ModClickUpList);

	ExtraModClicks.clear();
	ExtraModClicks = GetxRCLineList( ("*" + name + "OnX1ClickDown") );
	ExtraModClicks.merge( GetxRCLineList( ("*" + labelGroup + "OnX1ClickDown") ) );
	ExtraModClicks.merge( GetxRCLineList( (one + "OnX1ClickDown") ) );
	FillModClickList(ExtraModClicks, X1ModClickDownList);

	ExtraModClicks.clear();
	ExtraModClicks = GetxRCLineList( ("*" + name + "OnX1DoubleClick") );
	ExtraModClicks.merge( GetxRCLineList( ("*" + labelGroup + "OnX1DoubleClick") ) );
	ExtraModClicks.merge( GetxRCLineList( (one + "OnX1DoubleClick") ) );
	FillModClickList(ExtraModClicks, X1ModDoubleClickList);

	// X2
	ExtraModClicks.clear();
	ExtraModClicks = GetxRCLineList( ("*" + name + "OnX2Click") );
	ExtraModClicks.merge( GetxRCLineList( ("*" + labelGroup + "OnX2Click") ) );
	ExtraModClicks.merge( GetxRCLineList( (one + "OnX2Click") ) );
	FillModClickList(ExtraModClicks, X2ModClickList);

	ExtraModClicks.clear();
	ExtraModClicks = GetxRCLineList( ("*" + name + "OnX2ClickUp") );
	ExtraModClicks.merge( GetxRCLineList( ("*" + labelGroup + "OnX2ClickUp") ) );
	ExtraModClicks.merge( GetxRCLineList( (one + "OnX2ClickUp") ) );
	FillModClickList(ExtraModClicks, X2ModClickUpList);

	ExtraModClicks.clear();
	ExtraModClicks = GetxRCLineList( ("*" + name + "OnX2ClickDown") );
	ExtraModClicks.merge( GetxRCLineList( ("*" + labelGroup + "OnX2ClickDown") ) );
	ExtraModClicks.merge( GetxRCLineList( (one + "OnX2ClickDown") ) );
	FillModClickList(ExtraModClicks, X2ModClickDownList);

	ExtraModClicks.clear();
	ExtraModClicks = GetxRCLineList( ("*" + name + "OnX2DoubleClick") );
	ExtraModClicks.merge( GetxRCLineList( ("*" + labelGroup + "OnX2DoubleClick") ) );
	ExtraModClicks.merge( GetxRCLineList( (one + "OnX2DoubleClick") ) );
	FillModClickList(ExtraModClicks, X2ModDoubleClickList);


	wheelDownCommand = Getx2RCLine(name, labelGroup, "OnWheelDown", "");
	wheelUpCommand = Getx2RCLine(name, labelGroup, "OnWheelUp", "");
	enterCommand = Getx2RCLine(name, labelGroup, "OnEnter", "");
	leaveCommand = Getx2RCLine(name, labelGroup, "OnLeave", "");
	dropCommand = Getx2RCLine(name, labelGroup, "OnDrop", "");
	textChangeCommand = Getx2RCLine(name, labelGroup, "OnTextChange", "");
	hideCommand = Getx2RCLine(name, labelGroup, "OnHide", "");
	showCommand = Getx2RCLine(name, labelGroup, "OnShow", "");
	focusCommand = Getx2RCLine(name, labelGroup, "TextEditBoxOnFocus", "");
	unfocusCommand = Getx2RCLine(name, labelGroup, "TextEditBoxOnUnFocus", "");
	hoverCommand = Getx2RCLine(name, labelGroup, "OnHover", "");

	gainfocusCommand = Getx2RCLine(name, labelGroup, "OnFocusGained", "");
	loosefocusCommand = Getx2RCLine(name, labelGroup, "OnFocusLost", "");

	keyDownCommand = Getx2RCLine(name, labelGroup, "OnKeyDown", "");
	keyUpCommand = Getx2RCLine(name, labelGroup, "OnKeyUp", "");

	//Helper Settings
	focusOnEnter = Getx2RCBool(name, labelGroup, "FocusOnEnter");
	hoverTime = Getx2RCInt(name, labelGroup, "HoverTimeOut", 400, 100);

	//Event Regions
	FillLabel_Regions( GetxRCLineList( ("*" + name + "OnLeftClickRegion").c_str() ), labelLeftClickRegions);
	FillLabel_Regions( GetxRCLineList( ("*" + name + "OnLeftDoubleClickRegion").c_str() ), labelLeftDoubleClickRegions);
	
	FillLabel_Regions( GetxRCLineList( ("*" + name + "OnMiddleClickRegion").c_str() ), labelMiddleClickRegions);
	FillLabel_Regions( GetxRCLineList( ("*" + name + "OnMiddleDoubleClickRegion").c_str() ), labelMiddleDoubleClickRegions);

	FillLabel_Regions( GetxRCLineList( ("*" + name + "OnRightClickRegion").c_str() ), labelRightClickRegions);
	FillLabel_Regions( GetxRCLineList( ("*" + name + "OnRightDoubleClickRegion").c_str() ), labelRightDoubleClickRegions);

	FillLabel_Regions( GetxRCLineList( ("*" + name + "OnX1ClickRegion").c_str() ), labelX1ClickRegions);
	FillLabel_Regions( GetxRCLineList( ("*" + name + "OnX1DoubleClickRegion").c_str() ), labelX1DoubleClickRegions);

	FillLabel_Regions( GetxRCLineList( ("*" + name + "OnX2ClickRegion").c_str() ), labelX2ClickRegions);
	FillLabel_Regions( GetxRCLineList( ("*" + name + "OnX2DoubleClickRegion").c_str() ), labelX2DoubleClickRegions);

	FillLabel_Regions( GetxRCLineList( ("*" + name + "TooltipRegion").c_str() ), labelTooltipRegions, true);
	//--------------------------------------------------------------------------------
	//Docking/SnapTo/zOrder Setup
	//--------------------------------------------------------------------------------

	//Label Docking
	dockedToList.clear();

	dockedToList = GetxRCLineList( ("*" + name + "DockedTo") );

	if ( refresh )
	{
		for(StringListIterator sit = dockedToList.begin(); sit != dockedToList.end(); sit++)
		{
			numTokens = SplitString( (*sit), one, two, three, four, five, six, seven);
			
			if ( numTokens >= 1 )
			{
				if ( numTokens >= 2 )
				{
					if ( _stricmp(two.c_str(), "left") == NULL )
						dockedtoSide = 0;
					else if ( _stricmp(two.c_str(), "top") == NULL )
						dockedtoSide = 1;
					else if ( _stricmp(two.c_str(), "right") == NULL )
						dockedtoSide = 2;
					else if ( _stricmp(two.c_str(), "bottom") == NULL )
						dockedtoSide = 3;
					else if ( _stricmp(two.c_str(), "topleft") == NULL )
						dockedtoSide = 4;
					else if ( _stricmp(two.c_str(), "topright") == NULL )
						dockedtoSide = 5;
					else if ( _stricmp(two.c_str(), "bottomleft") == NULL )
						dockedtoSide = 6;
					else if ( _stricmp(two.c_str(), "bottomright") == NULL )
						dockedtoSide = 7;
					else
						dockedtoSide = -1;
				}

				bool found = false;
				bool alreadyFound = false;
				for(LabelListIterator plit = labelList.begin(); plit != labelList.end(); plit++)
				{
					if( _stricmp( one.c_str(), (*plit)->getName().c_str()) == NULL )
					{
						for(LabelListIterator lit = (*plit)->hasDockedLabelList.begin(); lit != (*plit)->hasDockedLabelList.end(); lit++)
						{
							if ( _stricmp( name.c_str(), (*lit)->getName().c_str() ) == NULL )
							{
								alreadyFound = true;
								break;
							}
						}

						if (!alreadyFound)
						{
							(*plit)->hasDockedLabelList.insert((*plit)->hasDockedLabelList.end(), this);
							found = true;
						}
						break;
					}
				}

				alreadyFound = false;
				if (!found)
				{
					for(LabelListIterator plit = labelList.begin(); plit != labelList.end(); plit++)
					{
						if( _stricmp( one.c_str(), (*plit)->getLabelGroup().c_str()) == NULL )
						{
							for(LabelListIterator lit = (*plit)->hasDockedLabelList.begin(); lit != (*plit)->hasDockedLabelList.end(); lit++)
							{
								if ( _stricmp( name.c_str(), (*lit)->getName().c_str() ) == NULL )
								{
									alreadyFound = true;
									break;
								}
							}

							if (!alreadyFound)
							{
								(*plit)->hasDockedLabelList.insert((*plit)->hasDockedLabelList.end(), this);
							}
						}
					}
				}
			}
		}
	}

	dockOnSnapTo = Getx2RCNamedValue(name, labelGroup, "DockOnSnap", dockValues, 0);

	//Snap To Settings
	//Use Split Helper Strings
	numTokens = SplitString(Getx2RCLine(name, labelGroup, "SnapTo", "0 desktop"), one, two, three, four, five, six, seven);

	if ( numTokens >= 1 )
	{
		snapDistance = atoi( one.c_str() );
		if (numTokens == 1)
			snapTo = "desktop";
		else if (numTokens >= 2)
			snapTo = two;
	}

	//zOrder Settings
	//Use Split Helper Strings
	numTokens = SplitString(Getx2RCLine(name, labelGroup, "BehindWindow", ""), one, two, three, four, five, six, seven);
	
	if ( numTokens >= 1 )
	{
		afterClass = one;
		if ( numTokens >= 2 )
			afterTitle = two;
	}

	//--------------------------------------------------------------------------------
	//Moving Setup
	//--------------------------------------------------------------------------------

	moveable = Getx2RCBool(name, labelGroup, "Moveable");
	moveRestriction = Getx2RCNamedValue(name, labelGroup, "Moveable", moverestrictionValues, 0);
	moveKey = Getx2RCNamedValue(name, labelGroup, "MoveModifierKey", keyValues, 1);
	moveButton = Getx2RCNamedValue(name, labelGroup, "MoveButton", buttonValues, 1);

	string workArea = Getx2RCLine(name, labelGroup, "MoveArea", "");
	if ( !workArea.empty() )
	{
		numTokens = SplitString(workArea, one, two, three, four, five, six, seven);
		
		if ( numTokens >= 4 )
		{
			customMoveArea.left = xParseCoord(one.c_str(), 0, maxX, hWnd);
			customMoveArea.top = xParseCoord(two.c_str(), 0, maxY, hWnd);
			customMoveArea.right = customMoveArea.left+xParseDimen(three.c_str(), maxX, maxX, hWnd);
			customMoveArea.bottom = customMoveArea.top+xParseDimen(four.c_str(), maxY, maxY, hWnd);

			useCustomMoveArea = true;
		}
	}

	//--------------------------------------------------------------------------------
	//Transparency/Alpha Features Setup
	//--------------------------------------------------------------------------------

	//Transparency Settings
	transparencyMode = Getx2RCNamedValue(name, labelGroup, "TransparencyMode", transparencyValues, 1);
	
	if ( !transparencyNeeded && transparencyMode == 1 )
		transparencyMode = 2;

	trackingIgnoreTrueTransparency = Getx2RCBool(name, labelGroup, "TrackingIgnoreTrueTransparency");

	alphaTransparency = Getx2RCInt(name, labelGroup, "AlphaTransparency", 255, 0, 255);
	alphaMap = Getx2RCBool(name, labelGroup, "AlphaMap");
	alphaFade = Getx2RCBool(name, labelGroup, "AlphaFade");

	//Use Split Helper Strings
	string tmpCustomFade = Getx2RCLine(name, labelGroup, "CustomAlphaFade", "");
	if ( !tmpCustomFade.empty() )
	{
		numTokens = SplitString(tmpCustomFade, one, two, three, four, five, six, seven);
		
		if (numTokens == 2)
		{
			alphaFade = true;
			alphaFadeStep = atoi( one.c_str() );
			if (alphaFadeStep < 1 || alphaFadeStep > 255)
				alphaFadeStep = 2;
		
			alphaFadeDelay = atoi( two.c_str() );
			if (alphaFadeDelay < 1 || alphaFadeDelay > 500)
				alphaFadeDelay = 10;
		}
	}

	if ( IsOS(OS_2KXP) )
		ghosted = Getx2RCBool(name, labelGroup, "Ghosted");
	else
		ghosted = false;

	//Final Check
	if ( !box && (alphaTransparency != 255 || alphaMap || alphaFade || ghosted) && IsOS(OS_2KXP) )
	{
		if ( (Getx2RCString(name, labelGroup, "Image", "")).empty() )
			alphaMap = false;

		if ( !refresh )
		{
			// MUST BE SET BEFORE setAlwaysOnTop()
			ModifyStyleEx(hWnd, 0, WS_EX_LAYERED);

			if (ghosted)
				ModifyStyleEx(hWnd, 0, WS_EX_TRANSPARENT);

			if ( !alphaMap && (alphaTransparency != 255 || alphaFade) && !hasHookedModules && Getx2RCLine(name, labelGroup, "TextEditBox", "").empty() )
			{
				fakealphaMap = true;
			}
			else
			{
				fakealphaMap = false;
				if ( !Getx2RCBool(name, labelGroup, "AlwaysOnTop") && !alphaMap )
					alwaysUpdateContent = true;
			}

			if ( alphaMap || fakealphaMap )
				SetWindowRgn(hWnd, NULL, true);

			if ( !alphaFade )
			{
				alphaFadeTransparencyBackup = alphaTransparency;
				UpdateLabelAlpha(alphaTransparency);
			}
			else
			{
				alphaFadeTransparencyBackup = alphaTransparency;
				alphaTransparency = 0;
				UpdateLabelAlpha(0);
			}
		}
	}
	else
	{
		alphaFadeTransparencyBackup = 255;
		alphaTransparency = 255;
		alphaMap = false;
		alphaFade = false;
		fakealphaMap = false;
		ghosted = false;
	}

	//--------------------------------------------------------------------------------
	//Special Features Setup
	//--------------------------------------------------------------------------------

	bUseFahrenheit = Getx2RCBool(name, labelGroup, "UseFahrenheit");

	//Cursor
	noMoveCursor = Getx2RCBool(name, labelGroup, "NoMoveCursor");
	cursorPath = Getx2RCString(name, labelGroup, "Cursor", "");
	if ( cursor )
	{
		DestroyCursor(cursor);
		cursor = NULL;
	}
	SendMessage(hWnd, LM_SETCURSOR, 0, 0);

	//File Actions
	if ( !dropCommand.empty() )
		DragAcceptFiles(hWnd, true);

	if ( !refresh )
	{	
		//AlwaysOnTop (before initial reposition due to MultiMonitor Support)
		setAlwaysOnTop(Getx2RCBool(name, labelGroup, "AlwaysOnTop"), true);
		
		//Initial Repositioning
		reposition(3, Getx2RCCoord(name, labelGroup, "X", 0, maxX, box, false),
						Getx2RCCoord(name, labelGroup, "Y", 0, maxY, box, true),
							Getx2RCDimen(name, labelGroup, "Width", 64, maxX, box, false),
								Getx2RCDimen(name, labelGroup, "Height", 64, maxY, box, true));
	}
	else
	{
		//Update internal Regions after a Refresh, since no Reposition was called!
		UpdateLabel_Regions( labelLeftClickRegions );
		UpdateLabel_Regions( labelLeftDoubleClickRegions );
		
		UpdateLabel_Regions( labelMiddleClickRegions );
		UpdateLabel_Regions( labelMiddleDoubleClickRegions );

		UpdateLabel_Regions( labelRightClickRegions );
		UpdateLabel_Regions( labelRightDoubleClickRegions );

		UpdateLabel_Regions( labelX1ClickRegions );
		UpdateLabel_Regions( labelX1DoubleClickRegions );

		UpdateLabel_Regions( labelX2ClickRegions );
		UpdateLabel_Regions( labelX2DoubleClickRegions );

		UpdateLabel_Regions( labelTooltipRegions );

		UpdateLabel_Regions( labelGhostRegions );
	}

	startHidden = Getx2RCBool(name, labelGroup, "StartHidden");

	//OnMove and OnResize AFTER first Positioning
	resizeCommand = Getx2RCLine(name, labelGroup, "OnResize", "");
	moveCommand = Getx2RCLine(name, labelGroup, "OnMove", "");

	//--------------------------------------------------------------------------------
	//Additional Extra Features Setup
	//--------------------------------------------------------------------------------

	//TextEdit Box
	numTokens = SplitString(Getx2RCLine(name, labelGroup, "TextEditBox", ""), one, two, three, four, five, six, seven);

	bool textboxStartHidden = false;
	int textboxAlign = 0;
	bool textboxPassword = false;
	bool textboxNumber = false;
	int textboxDigits = -1;

	if ( numTokens >= 3 )
	{
		textboxXOrg = one;
		textboxYOrg = two;
		textboxWidthOrg = three;

		if ( numTokens >= 4 )
		{
			textboxStartHidden = (strchr(four.c_str(), 'h') != 0 || strchr(four.c_str(), 'H') != 0);
			if ( strchr(four.c_str(), 'r') != 0 || strchr(four.c_str(), 'R') != 0 )
				textboxAlign = 2;
			else if ( strchr(four.c_str(), 'c') != 0 || strchr(four.c_str(), 'C') != 0)
				textboxAlign = 1;
			else
				textboxAlign = 0;
			textboxPassword = (strchr(four.c_str(), 'p') != 0 || strchr(four.c_str(), 'P') != 0);
			textboxNumber = (strchr(four.c_str(), 'n') != 0 || strchr(four.c_str(), 'N') != 0);

			if ( numTokens >= 5 )
				textboxDigits = atoi(five.c_str());
		}
	}

	enterAction = Getx2RCLine(name, labelGroup, "TextEditBoxEnterAction", ".execute");

	//Destroy on Refresh
	if ( refresh && hWndTextBox )
	{
		if ( textcursor )
		{
			DestroyCursor(textcursor);
			textcursor = 0;
		}

		// Remove the subclass from the edit control. 
        SetWindowLong(hWndTextBox, GWL_WNDPROC, (LONG) wpOrigEditProc); 

		if ( hTextBGBrush )
		{
			DeleteObject(hTextBGBrush);
			hTextBGBrush = 0;
		}

		if ( textboxhFont )
		{
			DeleteObject( (HGDIOBJ)(HFONT)textboxhFont );
			textboxhFont = 0;
		}
		
		DestroyWindow(hWndTextBox);
		hWndTextBox = 0;
	}
	
	if ( hWndTextBox == NULL && (!textboxXOrg.empty() || !textboxYOrg.empty() || !textboxWidthOrg.empty()) && !alphaMap && !fakealphaMap )
	{
		hWndTextBox = CreateWindow( "EDIT",
									TEXT(""),
									WS_CHILD | WS_VISIBLE | (textboxNumber ? ES_NUMBER : 0 ) | (textboxPassword ? ES_PASSWORD : 0) | ES_AUTOHSCROLL | (textboxAlign == 1 ? ES_CENTER : 0) | (textboxAlign == 2 ? ES_RIGHT : 0),
									0, 0,
									0, 0,
									hWnd,
									0,
									hInstance,
									0);
	}

	if ( hWndTextBox )
	{	
		if ( texteditboxbackground )
			delete texteditboxbackground;

		successCommand = Getx2RCLine(name, labelGroup, "TextEditBoxOnSuccess", "");
		failCommand = Getx2RCLine(name, labelGroup, "TextEditBoxOnFail", "");

		textcursorPath = Getx2RCString(name, labelGroup, "TextEditBoxCursor", "");
		if ( !textcursorPath.empty() )
		{
			if ( strstr(textcursorPath.c_str(), ":\\") != NULL )
			{
				textcursor = LoadCursorFromFile( textcursorPath.c_str() );
			}

			if ( !textcursor )
			{	
				char imagePath[MAX_LINE_LENGTH] = {0};
				VarExpansion( imagePath, "$lsimagefolder$" );
				strcat(imagePath, textcursorPath.c_str());

				textcursor = LoadCursorFromFile( imagePath );
			}
		}

		string strtextboxmode = Getx2RCString(name, labelGroup, "TextEditBoxPaintingMode", "transparent");
		if ( _stricmp(strtextboxmode.c_str(), "image") == NULL )
		{
			texteditboxbackground =	new ImageTexture();
			texteditboxbackground->configure((name + "TextEditBox"), (labelGroup + "TextEditBox"), false);
			textboxPaintingMode = 1;
		}
		else if ( _stricmp(strtextboxmode.c_str(), "solidcolor") == NULL )
		{
			texteditboxbackground = new SolidTexture();
			texteditboxbackground->configure((name + "TextEditBox"), (labelGroup + "TextEditBox"), false);
			textboxPaintingMode = 2;
		}
		else if ( _stricmp(strtextboxmode.c_str(), "plaincolor") == NULL )
		{
			textboxPaintingMode = -1;
			textboxPlainColor = Getx2RCColor(name, labelGroup, "TextEditBoxPlainColor", RGB(255,255,255));
		}

		if ( textboxDigits >= 1 )
		{
			SendMessage(hWndTextBox, EM_LIMITTEXT, textboxDigits, 0 );  
		}

		//Padding Calculation
		numTokens = SplitString(Getx2RCLine(name, labelGroup, "TextEditBoxPadding", "0"), one, two, three, four, five, six, seven);
		
		if (numTokens == 1)
		{
			tebleftBorder = tebrightBorder = tebtopBorder = tebbottomBorder = atoi(one.c_str());
		}
		else if (numTokens == 2)
		{
			tebleftBorder = tebrightBorder = atoi(one.c_str());
			tebtopBorder = tebbottomBorder = atoi(two.c_str());
		}
		else if (numTokens == 4)
		{
			tebleftBorder = atoi(one.c_str());
			tebrightBorder = atoi(two.c_str());
			tebtopBorder = atoi(three.c_str());
			tebbottomBorder = atoi(four.c_str());
		}

		if ( tebleftBorder <= 0 )
			tebleftBorder = 0;

		if ( tebrightBorder <= 0 )
			tebrightBorder = 0;

		if ( tebtopBorder <= 0 )
			tebtopBorder = 0;

		if ( tebbottomBorder <= 0 )
			tebbottomBorder = 0;

		// Buffer for Fontheight Calculation
		HDC	bufferDC = CreateCompatibleDC(NULL);
		HBITMAP bufferBitmap = CreateCompatibleBitmap(bufferDC, actualwidth, actualheight);
		HBITMAP oldBMP = (HBITMAP) SelectObject(bufferDC, bufferBitmap);

		textboxhFont = font->gethFont();

		SendMessage(hWndTextBox, WM_SETFONT, (WPARAM) textboxhFont, FALSE);

		RECT r = {0,0,0,0};

		HFONT hFont = (HFONT) SelectObject(bufferDC, textboxhFont);
				
		string testtext = "Test Text Calculation!";
		DrawTextEx(bufferDC, (char *) testtext.c_str(), testtext.length(), &r, DT_SINGLELINE | DT_CALCRECT, 0);
		textboxHeight = (r.bottom-r.top)+(tebtopBorder+tebbottomBorder);

		textboxhFont = (HFONT) SelectObject(bufferDC, hFont);

		DeleteObject(hFont);

		bufferBitmap = (HBITMAP) SelectObject(bufferDC, oldBMP);
		DeleteObject(bufferBitmap);
		DeleteDC(bufferDC);
 
		// Set margins
        SendMessage(hWndTextBox, EM_SETMARGINS, EC_LEFTMARGIN | EC_RIGHTMARGIN, MAKELONG(0,0));
       
		TextEditBoxSizeBorder.left = xParseCoord(textboxXOrg, 0, actualwidth, hWnd);
		TextEditBoxSizeBorder.right = TextEditBoxSizeBorder.left+xParseDimen(textboxWidthOrg, actualwidth, actualwidth, hWnd);
		TextEditBoxSizeBorder.top = xParseCoord(textboxYOrg, 0, actualheight, hWnd);
		TextEditBoxSizeBorder.bottom = TextEditBoxSizeBorder.top+textboxHeight;

		TextEditBoxSize.left = TextEditBoxSizeBorder.left+tebleftBorder;
		TextEditBoxSize.right = TextEditBoxSizeBorder.right-tebrightBorder;
		TextEditBoxSize.top = TextEditBoxSizeBorder.top+tebtopBorder;
		TextEditBoxSize.bottom = TextEditBoxSizeBorder.bottom-tebbottomBorder;

		SetWindowPos(hWndTextBox, NULL, 
						TextEditBoxSize.left,
						TextEditBoxSize.top,
						(TextEditBoxSize.right-TextEditBoxSize.left), 
						(TextEditBoxSize.bottom-TextEditBoxSize.top), 
						SWP_NOACTIVATE | SWP_NOZORDER);

		wpOrigEditProc = (WNDPROC) SetWindowLong(hWndTextBox, GWL_WNDPROC, (LONG) TextEditProc);

		if ( !textboxStartHidden )
		{
			ShowWindow(hWndTextBox, SW_SHOW);
			textboxVisible = true;

			PrefixedLSSetVariable(name, "TextEditBoxVisible", "true");
		}
		else
		{
			ShowWindow(hWndTextBox, SW_HIDE);
			textboxVisible = false;

			PrefixedLSSetVariable(name, "TextEditBoxVisible", "false");
		}
	}
	else
	{
		textboxXOrg = textboxYOrg = textboxWidthOrg = "";
		hWndTextBox = NULL;

		successCommand = "";
		failCommand = "";
	}

	//Tooltips
	InitializeTooltip();

	//--------------------------------------------------------------------------------
	//Final Routines Start (TextUpdate and AnimationStart)
	//--------------------------------------------------------------------------------

	if ( refresh )
	{
		update();
		repaint(true);
		if ( hasHookedModules )
			RedrawWindow(hWnd, NULL, NULL, RDW_UPDATENOW | RDW_ALLCHILDREN);
	}
	else
	{
		setText(Getx2RCString(name, labelGroup, "Text", ""), true);
	}
	
	if (normalAnim)
	{
		SetTimer(hWnd, TIMER_FRAMEANIMATION, imageFrameDelay, NULL);
	}
}

void Label::InitializeTooltip(bool onlyhrefTooltip)
{
	//Initialize Basic Helper Strings
	int numTokens = 0;
	string one = "";
	string two = "";
	string three = "";
	string four = "";
	string five = "";
	string six = "";
	string seven = "";

	//ToolTips
	bool balloontooltip = false;
	string balloonTitle = "";
	int balloonIcon = 0;

	numTokens = SplitString(Getx2RCLine(name, labelGroup, "BalloonTooltip", "false"), one, two, three, four, five, six, seven);

	if ( _stricmp( one.c_str(), "false") != NULL && (IsOS(OS_2KXP) || IsOS(OS_ME)) )
	{
		balloontooltip = true;

		if ( numTokens >= 1 && !one.empty() && _stricmp( one.c_str(), "true") != NULL )
		{
			balloonTitle = one;
			if ( numTokens >= 2 )
				balloonIcon = atoi( two.c_str() );
			if ( balloonIcon > 3 || balloonIcon < 0 )
				balloonIcon = 0;
		}
	}
	else
	{
		balloontooltip = false;
	}

	originalTooltip = Getx2RCString(name, labelGroup, "Tooltip", "");

	if ( !originalTooltip.empty() || !labelTooltipRegions.empty() || onlyhrefTooltip )
	{
		while ( strstr(originalTooltip.c_str(), "<br>") != NULL )
			originalTooltip.replace(originalTooltip.find("<br>"), 4, "\n", 1);
		
		while ( strstr(originalTooltip.c_str(), "\\t") != NULL )
			originalTooltip.replace(originalTooltip.find("\\t"), 2, "\t", 1);

		if ( !TooltipHints )
		{
			// struct specifying control classes to register
			INITCOMMONCONTROLSEX iccex; 

			/* INITIALIZE COMMON CONTROLS */
			iccex.dwICC = ICC_WIN95_CLASSES;
			iccex.dwSize = sizeof(INITCOMMONCONTROLSEX);

			InitCommonControlsEx(&iccex);

			TooltipHints = CreateWindowEx(WS_EX_TOOLWINDOW,
											TOOLTIPS_CLASS,
											TEXT(""),
											WS_POPUP | TTS_NOPREFIX | TTS_ALWAYSTIP | (balloontooltip ? TTS_BALLOON : 0),
											CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
											NULL,
											NULL,
											hInstance,
											NULL);

			if ( TooltipHints )
			{
				SetWindowPos(TooltipHints, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);  
		
				SendMessage(TooltipHints, TTM_SETMAXTIPWIDTH, 0, Getx2RCInt(name, labelGroup, "ToolTipMaxWidth", 300, 50));

				SendMessage( TooltipHints, (UINT) TTM_SETTIPBKCOLOR, (WPARAM) (COLORREF) Getx2RCColor(name, labelGroup, "TooltipColor", GetSysColor(COLOR_INFOBK)), 0 );   
				SendMessage( TooltipHints, (UINT) TTM_SETTIPTEXTCOLOR, (WPARAM) (COLORREF) Getx2RCColor(name, labelGroup, "ToolTipTextColor", GetSysColor(COLOR_INFOTEXT)), 0);  
			
				numTokens = SplitString(Getx2RCLine(name, labelGroup, "ToolTipDurations", "-1 -1"), one , two, three, four, five, six, seven);
			
				int initial = -1;
				int autopop = -1;

				if ( numTokens >= 1 )
				{
					initial = atoi(one.c_str());
					if ( numTokens >= 2 )
						autopop = atoi(two.c_str());
				}

				if ( initial < 0 )
					initial = -1;
				if ( autopop < 1000 )
					autopop = 1000;

				if ( initial != -1 )
					SendMessage( TooltipHints, (UINT) TTM_SETDELAYTIME,	(WPARAM) (DWORD) TTDT_INITIAL, (LPARAM) MAKELONG(initial, 0) );
				if ( autopop != -1 )
					SendMessage( TooltipHints, (UINT) TTM_SETDELAYTIME,	(WPARAM) (DWORD) TTDT_AUTOPOP, (LPARAM) MAKELONG(autopop, 0) );

				if ( balloontooltip && !balloonTitle.empty() )
					SendMessage( TooltipHints, TTM_SETTITLE, (WPARAM) (int) balloonIcon, (LPARAM) balloonTitle.c_str() );

				strcpy(tooltip, (systemInfo->processLabelText(originalTooltip, this, &dynamicTooltip)).c_str());
				addHint(hWnd, tooltip);
			}
		}
		else
		{
			strcpy(tooltip, (systemInfo->processLabelText(originalTooltip, this, &dynamicTooltip)).c_str());
			updateHint(hWnd, tooltip);
		}
	}
	else
	{
		if ( hWnd && TooltipHints )
		{
			removeHint(hWnd);
			DestroyWindow(TooltipHints);
			TooltipHints = NULL;
		}
	}
}

void Label::FillModClickList(StringList inputList, ModClickList &outputList)
{
	if ( inputList.empty() )
		return;

	char modifier[32] = {0};
	char action[MAX_LINE_LENGTH] = {0};
	char *buffers[] = {modifier};

	for(StringListIterator sit = inputList.begin(); sit != inputList.end(); sit++)
	{
		if(LCTokenize((*sit).c_str(), buffers, 1, action) >= 1)
		{
			ModClick *insert = new ModClick();

			if ( _stricmp( modifier, ".ctrl") == NULL )
				insert->modifier = 1;
			else if ( _stricmp( modifier, ".shift") == NULL )
				insert->modifier = 2;
			else if ( _stricmp( modifier, ".alt") == NULL )
				insert->modifier = 3;
			else if ( _stricmp( modifier, ".ctrl+shift") == NULL || _stricmp( modifier, ".shift+ctrl") == NULL )
				insert->modifier = 12;
			else if ( _stricmp( modifier, ".ctrl+alt") == NULL || _stricmp( modifier, ".alt+ctrl") == NULL )
				insert->modifier = 13;
			else if ( _stricmp( modifier, ".shift+alt") == NULL || _stricmp( modifier, ".alt+shift") == NULL )
				insert->modifier = 23;
			else if ( _stricmp( modifier, ".ctrl+shift+alt") == NULL 
						|| _stricmp( modifier, ".ctrl+alt+shift") == NULL
						|| _stricmp( modifier, ".shift+alt+ctrl") == NULL
						|| _stricmp( modifier, ".shift+ctrl+alt") == NULL 
						|| _stricmp( modifier, ".alt+ctrl+shift") == NULL
						|| _stricmp( modifier, ".alt+shift+ctrl") == NULL )
				insert->modifier = 123;
			else
			{
				delete insert;
				continue;
			}

			insert->action = action;
				
			outputList.insert(outputList.end(), insert);
		}
	}
}

bool Label::CheckModClick(ModClickList inputList)
{
	if ( inputList.empty() )
		return false;

	bool found = false;

	for(ModClickListIterator mlit = inputList.begin(); mlit != inputList.end(); mlit++)
	{
		if ( (*mlit)->modifier == 123 && (0x8000 & GetKeyState(VK_CONTROL)) && (0x8000 & GetKeyState(VK_SHIFT)) && (0x8000 & GetKeyState(VK_MENU)) )
		{
			xLSExecute(hWnd, (*mlit)->action.c_str(), SW_SHOWNORMAL);
			found = true;
			break;
		}
		else if ( (*mlit)->modifier == 12 && (0x8000 & GetKeyState(VK_CONTROL)) && (0x8000 & GetKeyState(VK_SHIFT)) && !(0x8000 & GetKeyState(VK_MENU)))
		{
			xLSExecute(hWnd, (*mlit)->action.c_str(), SW_SHOWNORMAL);
			found = true;
			break;
		}
		else if ( (*mlit)->modifier == 13 && (0x8000 & GetKeyState(VK_CONTROL)) && !(0x8000 & GetKeyState(VK_SHIFT)) && (0x8000 & GetKeyState(VK_MENU)) )
		{
			xLSExecute(hWnd, (*mlit)->action.c_str(), SW_SHOWNORMAL);
			found = true;
			break;
		}
		else if ( (*mlit)->modifier == 23 && !(0x8000 & GetKeyState(VK_CONTROL)) && (0x8000 & GetKeyState(VK_SHIFT)) && (0x8000 & GetKeyState(VK_MENU)) )
		{
			xLSExecute(hWnd, (*mlit)->action.c_str(), SW_SHOWNORMAL);
			found = true;
			break;
		}
		else if ( (*mlit)->modifier == 1 && (0x8000 & GetKeyState(VK_CONTROL)) && !(0x8000 & GetKeyState(VK_SHIFT)) && !(0x8000 & GetKeyState(VK_MENU)) )
		{
			xLSExecute(hWnd, (*mlit)->action.c_str(), SW_SHOWNORMAL);
			found = true;
			break;
		}
		else if ( (*mlit)->modifier == 2 && !(0x8000 & GetKeyState(VK_CONTROL)) && (0x8000 & GetKeyState(VK_SHIFT)) && !(0x8000 & GetKeyState(VK_MENU)) )
		{
			xLSExecute(hWnd, (*mlit)->action.c_str(), SW_SHOWNORMAL);
			found = true;
			break;
		}
		else if ( (*mlit)->modifier == 3 && !(0x8000 & GetKeyState(VK_CONTROL)) && !(0x8000 & GetKeyState(VK_SHIFT)) && (0x8000 & GetKeyState(VK_MENU)) )
		{
			xLSExecute(hWnd, (*mlit)->action.c_str(), SW_SHOWNORMAL);
			found = true;
			break;
		}
	}

	return found;
}

void Label::FillLabel_Regions(StringList inputList, Label_RegionList &outputList, bool tooltipregion)
{
	if ( inputList.empty() )
		return;

	char x[16] = {0};
	char y[16] = {0};
	char width[16] = {0};
	char height[16] = {0};
	char command[MAX_LINE_LENGTH] = {0};;
	char *buffers[] = {x, y, width, height};

	for(StringListIterator sit = inputList.begin(); sit != inputList.end(); sit++)
	{
		if(LCTokenize((*sit).c_str(), buffers, 4, command) >= 4)
		{
			Label_Region *insert = new Label_Region();

			insert->orgx = x;
			insert->orgy = y;
			insert->orgwidth = width;
			insert->orgheight = height;

			if ( !tooltipregion )
			{
				insert->action = command;
			}
			else
			{
				insert->originaltooltip = command;

				insert->originaltooltip = trimit(insert->originaltooltip);

				CString quotes = insert->originaltooltip.c_str();
				if ( quotes.Left(1) == "\"" && quotes.Right(1) == "\"" )
					insert->originaltooltip = quotes.Mid(1, quotes.GetLength()-2);

				while ( strstr(insert->originaltooltip.c_str(), "<br>") != NULL )
					insert->originaltooltip.replace(insert->originaltooltip.find("<br>"), 4, "\n", 1);
			}

			outputList.insert(outputList.end(), insert);
		}
	}
}

void Label::UpdateLabel_Regions(Label_RegionList inputList)
{
	if ( inputList.empty() )
		return;	

	for (Label_RegionListIterator lrit = inputList.begin(); lrit != inputList.end(); lrit++)
	{
		(*lrit)->region.left = xParseCoord((*lrit)->orgx, 0, actualwidth, hWnd);
		(*lrit)->region.right = (*lrit)->region.left+xParseDimen((*lrit)->orgwidth, actualwidth, actualwidth, hWnd);
		(*lrit)->region.top = xParseCoord((*lrit)->orgy, 0, actualheight, hWnd);
		(*lrit)->region.bottom = (*lrit)->region.top+xParseDimen((*lrit)->orgheight, actualheight, actualheight, hWnd);
	}
}

//------------------------------------------------------------
//"Modify Styles"
//------------------------------------------------------------

void Label::setAlwaysOnTop(bool setalwaysOnTop, bool firstSetup)
{
	if ( box != NULL )
		return;
	
	bool realChange = ( alwaysOnTop != setalwaysOnTop );

	alwaysOnTop = setalwaysOnTop;
	
	if ( realChange || firstSetup )
	{
		ModifyStyle(hWnd, WS_POPUP, WS_CHILD);
		SetParent(hWnd, alwaysOnTop ? 0 : GetLitestepDesktop());
		ModifyStyle(hWnd, WS_CHILD, WS_POPUP);
	}

	disablezOrderChange = false;

	if ( !afterClass.empty() )
	{
		HWND before = FindWindowEx2(afterClass, afterTitle);
		
		if ( before != NULL )
		{
			if ( alwaysOnTop )
			{
				SetWindowPos(hWnd, HWND_TOPMOST,
								0, 0, 0, 0,
								SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE );

				disablezOrderChange = false;
			}

			SetWindowPos(hWnd, before,
							0, 0, 0, 0,
							SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE );
		}
		else
		{
			SetWindowPos(hWnd, alwaysOnTop ? HWND_TOPMOST : HWND_NOTOPMOST,
							0, 0, 0, 0,
							SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE );
		}
	}
	else
	{
		SetWindowPos(hWnd, alwaysOnTop ? HWND_TOPMOST : HWND_NOTOPMOST,
						0, 0, 0, 0,
						SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE );
	}

	/*if ( realMonitors > 1 )
	{
		if ( setalwaysOnTop )
		{
			reposition(1, minscreenX+actualx, minscreenY+actualy, 0, 0);
		}
		else
		{
			reposition(1, actualx-minscreenX, actualy-minscreenY, 0, 0);
		}
	
		repaint();
	}*/
}

void Label::setGhosted(bool setghosted)
{	
	if (box != NULL)
		return;

	ghosted = setghosted;
	
	if (ghosted)
	{
		if ( !layeredFlag )
		{
			ModifyStyleEx(hWnd, 0, WS_EX_LAYERED | WS_EX_TRANSPARENT);
			UpdateLabelAlpha(alphaTransparency);
		}
		else
		{
			ModifyStyleEx(hWnd, 0, WS_EX_TRANSPARENT);
		}
	}
	else
	{
		ModifyStyleEx(hWnd, WS_EX_TRANSPARENT, 0);
	}
}

void Label::setAlpha(int newAlphaTransparency, bool hookingReset)
{
	if (box != NULL)
		return;

	if ( hookingReset )
	{
		UpdateLabelAlpha(alphaFadeTransparencyBackup);
		
		if ( !alwaysOnTop && !alphaMap )
			alwaysUpdateContent = true;

		return;
	}

	if (newAlphaTransparency >= 0 && newAlphaTransparency <= 255 && IsOS(OS_2KXP) )
	{
		if (!layeredFlag)
		{
			ModifyStyle(hWnd, WS_POPUP, WS_CHILD);
			SetParent(hWnd, 0);
			ModifyStyle(hWnd, WS_CHILD, WS_POPUP);

			ModifyStyleEx(hWnd, 0, WS_EX_LAYERED);

			ModifyStyle(hWnd, WS_POPUP, WS_CHILD);
			SetParent(hWnd, alwaysOnTop ? 0 : GetLitestepDesktop());
			ModifyStyle(hWnd, WS_CHILD, WS_POPUP);

			layeredFlag = true;
		}
		
		alphaTransparency = newAlphaTransparency;
		alphaFadeTransparencyBackup = newAlphaTransparency;
		
		UpdateLabelAlpha(newAlphaTransparency);
	}
}

void Label::setAlphaFade(int newAlphaTransparency, int step, int delay)
{
	if (box != NULL)
		return;

	if (newAlphaTransparency >= 0 && newAlphaTransparency <= 255 && IsOS(OS_2KXP) )
	{
		if (!layeredFlag)
		{
			ModifyStyle(hWnd, WS_POPUP, WS_CHILD);
			SetParent(hWnd, 0);
			ModifyStyle(hWnd, WS_CHILD, WS_POPUP);

			ModifyStyleEx(hWnd, 0, WS_EX_LAYERED);

			ModifyStyle(hWnd, WS_POPUP, WS_CHILD);
			SetParent(hWnd, alwaysOnTop ? 0 : GetLitestepDesktop());
			ModifyStyle(hWnd, WS_CHILD, WS_POPUP);

			layeredFlag = true;
		}
		
		KillTimer(hWnd, TIMER_ALPHAFADEUPBANG);
		KillTimer(hWnd, TIMER_ALPHAFADEDOWNBANG);

		KillTimer(hWnd, TIMER_ALPHAFADEIN);
		KillTimer(hWnd, TIMER_ALPHAFADEOUT);

		UpdateAlpha = alphaTransparency;
		
		alphaFadeBangEndValue = newAlphaTransparency;

		alphaTransparency = newAlphaTransparency;
		alphaFadeTransparencyBackup = newAlphaTransparency;

		alphaFadeBangStep = step;

		if ( UpdateAlpha > alphaFadeBangEndValue )
			SetTimer(hWnd, TIMER_ALPHAFADEDOWNBANG, delay, NULL);
		else if ( UpdateAlpha < alphaFadeBangEndValue )
			SetTimer(hWnd, TIMER_ALPHAFADEUPBANG, delay, NULL);
		else
			UpdateLabelAlpha(newAlphaTransparency);
	}
}

//------------------------------------------------------------
//Set Text Settings
//------------------------------------------------------------

void Label::setText(const string &text, bool updateBackup, int extraFeatureMode, GhostLabel *ghostlabel)
{
	if ( extraFeatureMode == 2 )
	{
		if ( ghostlabel != NULL )
		{
			ghostlabel->originaltext = text;
			update();
		}
	}
	else if ( extraFeatureMode == 1 )
	{
		if ( hWndTextBox )
		{
			SetWindowText( hWndTextBox, (systemInfo->processLabelText(text, this, NULL)).c_str() );
		}
	}
	else
	{
		//Backup User Changed TEXT <-> Internal Change
		if ( updateBackup )
		{
			originalTextBackup = text;
			labelHREFClickRegions.clear();
		}

		//originalText MUST be splitted, even if process is empty!
		if (noTextSplit)
			originalText = split(text, "%&*&%");
		else
			originalText = split(text, ";");
	
		if ( updateBackup )
		{
			current = 0;
			currentBackup = 0;
		}
		else
		{
			current = currentBackup;
		}

		update();
	}
}

void Label::previous()
{
	current--;

	if(current < 0)
		current = originalText.size()-1;

	currentBackup = current;

	update();
}

void Label::next()
{
	current++;

	if(current >= originalText.size())
		current = 0;

	currentBackup = current;

	update();
}

void Label::setScrolling(int scrolling)
{
	if (scrolling != 0 && scroll == 0)
	{
		scrollPosition = 0;
		skipScrollRepaint = false;
		repaint();
	}
	else if (scrolling == 0 && scroll != 0)
	{
		KillTimer(hWnd, TIMER_SCROLL);
		scrollPosition = 0;
		repaint();
	}

	scrollLimit = 0;
	scroll = scrolling;
	if (scrolling != 0)
		scrollBackup = scrolling;
}

void Label::setScrollLimit(int limit)
{
	if (limit >= 0)
	{
		setScrolling(scrollBackup);
		scrollLimit = limit + 1;
	}
	else
	{
		setScrolling(0);
	}
}

void Label::update()
{
	bool ghostUpdateContent = false;
	string oldText = text;

	//Tooltips
	if ( TooltipHints )
	{
		if ( !originalTooltip.empty() )
		{
			strcpy(tooltip, (systemInfo->processLabelText(originalTooltip, this, &dynamicTooltip)).c_str());	

			if (dynamicTooltip)
				updateHint(hWnd, tooltip);
		}

		if ( !labelTooltipRegions.empty() )
		{
			POINT pt;
			GetCursorPos(&pt);
			ScreenToClient(hWnd, &pt);
			bool tmpdynamicTooltip = false;

			for(Label_RegionListIterator lrit = labelTooltipRegions.begin(); lrit != labelTooltipRegions.end(); lrit++)
			{
				if ( !(*lrit)->originaltooltip.empty() )
				{
					string oldTooltipText = (*lrit)->tooltip;
					(*lrit)->tooltip = systemInfo->processLabelText((*lrit)->originaltooltip, this, &tmpdynamicTooltip);
				
					if ( !dynamicTooltip && tmpdynamicTooltip )
						dynamicTooltip = true;

					if( PtInRect( &(*lrit)->region, pt) )
					{	
						if ( !(*lrit)->tooltip.empty() && _stricmp(tooltip, (*lrit)->tooltip.c_str()) != NULL )
						{
							strcpy(tooltip, (*lrit)->tooltip.c_str());
							updateHint(hWnd, tooltip);
						}
					}
				}
			}
		}
	}

	if ( !ghostlabelList.empty() )
	{
		string oldGhostText = "";
		bool tmpdynamicText = false;

		for(GhostLabelListIterator glit = ghostlabelList.begin(); glit != ghostlabelList.end(); glit++)
		{
			if ( *glit != NULL )
			{
				if ( !(*glit)->originaltext.empty() )
				{
					oldGhostText = (*glit)->text;
					(*glit)->text = systemInfo->processLabelText((*glit)->originaltext, this, &tmpdynamicText);

					if ( !dynamicText && tmpdynamicText )
						dynamicText = true;

					if ( !ghostUpdateContent && strcmp((*glit)->text.c_str(), oldGhostText.c_str()) != NULL )
						ghostUpdateContent = true;
				}
			}
			else
			{
				break;
			}
		}
	}

	//Process Escape Sequence for ';'
	string process = originalText[current];
	while ( strstr(process.c_str(), "%~") != NULL )
		process.replace(process.find("%~"), 2, ";", 1);

	//Process [labelname]
	while ( strstr(process.c_str(), "[labelname]") != NULL )
		process.replace(process.find("[labelname]"), 11, name.c_str(), name.length());

	// Normal Text Processing
	if ( !noEscape )
	{
		bool tmpdynamicText = false;

		text = systemInfo->processLabelText(process, this, &tmpdynamicText);
		
		if ( !dynamicText && tmpdynamicText )
			dynamicText = true;

		while ( strstr(text.c_str(), "\t") != NULL )
			text.replace(text.find("\t"), 1, " ", 1);
	}
	// Skip Text Processing
	else
	{
		text = process;
	}

	if ( text.empty() )
		skipScrollRepaint = true;
	else 
		skipScrollRepaint = false;

	if ( strcmp(text.c_str(), oldText.c_str()) != NULL )
	{	
		// Buffer for Fontheight Calculation
		HDC	bufferDC = CreateCompatibleDC(NULL);
		HBITMAP bufferBitmap = CreateCompatibleBitmap(bufferDC, actualwidth, actualheight);
		HBITMAP oldBMP = (HBITMAP) SelectObject(bufferDC, bufferBitmap);
		
		POINT size = font->measure(bufferDC, leftBorder, actualwidth-rightBorder, text, 0/*normal*/);

		oldscrolltextwidth = size.x+6;
		oldscrolltextheight = size.y+6;
		
		bufferBitmap = (HBITMAP) SelectObject(bufferDC, oldBMP);
		DeleteObject(bufferBitmap);
		DeleteDC(bufferDC);
		
		if ( autoWidthMode != 0 || autoHeightMode != 0 )
			autoSize();

		if(!textChangeCommand.empty())
			xLSExecute(hWnd, textChangeCommand.c_str(), SW_SHOWNORMAL);
	
		repaint();
	}
	else if ( ghostUpdateContent || alwaysUpdateContent )
	{
		repaint();
	}

	if ( dynamicText || dynamicTooltip || alwaysUpdateContent )
	{ 
		SetTimer(hWnd, TIMER_UPDATE, updateInterval, 0);
	}
	else
	{
		KillTimer(hWnd, TIMER_UPDATE);
	}
}

//------------------------------------------------------------
//Set Animations
//------------------------------------------------------------

void Label::setAnimation(bool start, int loops, int startFrame)
{
	imageLoop = loops;
	normalAnim = start;

	if (start)
	{
		frameCounter = startFrame;
		loopCounter = 0;
		SetTimer(hWnd, TIMER_FRAMEANIMATION, imageFrameDelay, NULL);
	}
	else
	{
		frameCounter = 0;
		loopCounter = 0;
		KillTimer(hWnd, TIMER_FRAMEANIMATION);
		repaint(true);
	}	
}

//------------------------------------------------------------
//Move/Resize/Reposition
//------------------------------------------------------------
//Type == 1 -> Move
//Type == 2 -> Resize
//Type == 3 -> Reposition

void Label::reposition(int type, int newx, int newy, int newwidth, int newheight, int steps, int time, bool repositionBy, bool movecommand, Label* sender)
{
	//Prevents infinite looping of docked moving
	this->sender = sender;

	RECT rectPosition;
	rectPosition.left = actualx;
	rectPosition.right = actualx+actualwidth;
	rectPosition.top = actualy;
	rectPosition.bottom = actualy+actualheight;

	POINT posHolder;
	posHolder.x = rectPosition.left;
	posHolder.y = rectPosition.top;

	POINT sizeHolder;
	sizeHolder.x = (rectPosition.right-rectPosition.left);
	sizeHolder.y = (rectPosition.bottom-rectPosition.top);

	if (steps > 0 && time > 0)
	{
		if (time > 50)
			time = 50;

		if (repositionBy)
		{
			if (type != 2)
			{
				newx = rectPosition.left + newx;
				newy = rectPosition.top + newy;
			}
			if (type != 1)
			{
				newwidth = (rectPosition.right-rectPosition.left) + newwidth;
				newheight = (rectPosition.bottom-rectPosition.top) + newheight;
			}
		}

		int counter = 0;
		int deltax = newx - rectPosition.left;
		int deltay = newy - rectPosition.top;
		int deltasizex = newwidth - (rectPosition.right-rectPosition.left);
		int deltasizey = newheight - (rectPosition.bottom-rectPosition.top);

		bool growx;
		bool growy;
		bool growsizex;
		bool growsizey;
		
		if (deltax < 0)
			growx = false;
		else
			growx = true;
		if (deltay < 0)
			growy = false;
		else
			growy = true;
		
		if (deltasizex < 0)
			growsizex = false;
		else
			growsizex = true;
		if (deltasizey < 0)
			growsizey = false;
		else
			growsizey = true;

		int restx = 0;
		int resty = 0;
		int restw = 0;
		int resth = 0;

		div_t div_result;
		div_result = div( deltax, steps );
		deltax = div_result.quot;
		restx = div_result.rem;
		div_result = div( deltay, steps );
		deltay = div_result.quot;
		resty = div_result.rem;

		div_result = div( deltasizex, steps );
		deltasizex = div_result.quot;
		restw = div_result.rem;
		div_result = div( deltasizey, steps );
		deltasizey = div_result.quot;
		resth = div_result.rem;

		if (deltax == 0 && growx)
			deltax = 1;
		else if (deltax == 0 && !growx)
			deltax = -1;
		if (deltay == 0 && growy)
			deltay = 1;
		else if (deltay == 0 && !growy)
			deltay = -1;

		if (deltasizex == 0 && growsizex)
			deltasizex = 1;
		else if (deltasizex == 0 && !growsizex)
			deltasizex = -1;
		if (deltasizey == 0 && growsizey)
			deltasizey = 1;
		else if (deltasizey == 0 && !growsizey)
			deltasizey = -1;

		int dynx = rectPosition.left;
		int dyny = rectPosition.top;
		int dynwidth = (rectPosition.right-rectPosition.left);
		int dynheight = (rectPosition.bottom-rectPosition.top);

		int stopper = 0;
		while (counter <= steps)
		{
			stopper = 0; // Stop if x, y, width and height are finished, stopper = 4
			if ( restx > 0 && deltax > 0 )
			{
				dynx = dynx+deltax+1;
				restx--;
			}
			else if ( restx < 0 && deltax < 0 )
			{
				dynx = dynx+deltax-1;
				restx++;
			}
			else
			{
				dynx = dynx+deltax;
			}

			if (growx && dynx > newx)
			{
				dynx = newx;
				stopper++;
			}
			else if (!growx && dynx < newx)
			{
				dynx = newx;
				stopper++;
			}

			if ( resty > 0 && deltay > 0 )
			{
				dyny = dyny+deltay+1;
				resty--;
			}
			else if ( resty < 0 && deltay < 0 )
			{
				dyny = dyny+deltay-1;
				resty++;
			}
			else
			{
				dyny = dyny+deltay;
			}

			if (growy && dyny > newy)
			{
				dyny = newy;
				stopper++;
			}
			else if (!growy && dyny < newy)
			{
				dyny = newy;
				stopper++;
			}

			if ( restw > 0 && deltasizex > 0 )
			{
				dynwidth = dynwidth+deltasizex+1;
				restw--;
			}
			else if ( restw < 0 && deltasizex < 0 )
			{
				dynwidth = dynwidth+deltasizex-1;
				restw++;
			}
			else
			{
				dynwidth = dynwidth+deltasizex;
			}

			if (growsizex && dynwidth > newwidth)
			{
				dynwidth = newwidth;
				stopper++;
			}
			else if (!growsizex && dynwidth < newwidth)
			{
				dynwidth = newwidth;
				stopper++;
			}

			if ( resth > 0 && deltasizey > 0 )
			{
				dynheight = dynheight+deltasizey+1;
				resth--;
			}
			else if ( resth < 0 && deltasizey < 0 )
			{
				dynheight = dynheight+deltasizey-1;
				resth++;
			}
			else
			{
				dynheight = dynheight+deltasizey;
			}

			if (growsizey && dynheight > newheight)
			{
				dynheight = newheight;
				stopper++;
			}
			else if (!growsizey && dynheight < newheight)
			{
				dynheight = newheight;
				stopper++;
			}

			if ( type == 2 || type == 3 )
				resizeBorderResize = false;
			
			SetWindowPos(hWnd, 0, dynx, dyny, dynwidth, dynheight,
					SWP_NOACTIVATE | (type == 2 ? SWP_NOMOVE : 0) | (type == 1 ? SWP_NOSIZE : 0 ) | SWP_NOZORDER);

			if (box != NULL)
				RedrawWindow(box, NULL, NULL, RDW_INVALIDATE | RDW_UPDATENOW | RDW_ALLCHILDREN);

			repaint();
		
			if (stopper == 4)
				counter = steps;

			counter++;
			
			Sleep(time);
		}

		if ( type == 1 )
		{
			SetWindowPos(hWnd, 0, newx, newy, 0, 0,
						SWP_NOACTIVATE | SWP_NOSIZE | SWP_NOZORDER);
		}
		else if ( type == 2 )
		{
			resizeBorderResize = false;

			SetWindowPos(hWnd, 0, 0, 0, newwidth, newheight,
						SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOZORDER);

			if ( !fireOnResizeDuringResize && !resizeCommand.empty() && (newwidth != sizeHolder.x || newheight != sizeHolder.y) )
				xLSExecute(hWnd, resizeCommand.c_str(), SW_SHOWNORMAL);

		}
		else if ( type == 3 )
		{
			resizeBorderResize = false;

			SetWindowPos(hWnd, 0, newx, newy, newwidth, newheight,
						SWP_NOACTIVATE | SWP_NOZORDER);

			if ( !fireOnResizeDuringResize && !resizeCommand.empty() && (newwidth != sizeHolder.x || newheight != sizeHolder.y) )
				xLSExecute(hWnd, resizeCommand.c_str(), SW_SHOWNORMAL);
		}
		
		if (box != NULL)
			repaint();
	}
	else
	{
		if (repositionBy)
		{
			if (type != 2)
			{
				newx = rectPosition.left + newx;
				newy = rectPosition.top + newy;
			}
			if (type != 1)
			{
				newwidth = (rectPosition.right-rectPosition.left) + newwidth;
				newheight = (rectPosition.bottom-rectPosition.top) + newheight;
			}
		}

		if ( type == 1 )
		{
			SetWindowPos(hWnd, 0, newx, newy, 0, 0,
						SWP_NOACTIVATE | SWP_NOSIZE | SWP_NOZORDER);
		}
		else if ( type == 2 )
		{
			resizeBorderResize = false;

			SetWindowPos(hWnd, 0, 0, 0, newwidth, newheight,
						SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOZORDER);

			if ( !fireOnResizeDuringResize && !resizeCommand.empty() && (newwidth != sizeHolder.x || newheight != sizeHolder.y) )
				xLSExecute(hWnd, resizeCommand.c_str(), SW_SHOWNORMAL);
		}
		else if ( type == 3 )
		{
			resizeBorderResize = false;

			SetWindowPos(hWnd, 0, newx, newy, newwidth, newheight,
						SWP_NOACTIVATE | SWP_NOZORDER);

			if ( !fireOnResizeDuringResize && !resizeCommand.empty() && (newwidth != sizeHolder.x || newheight != sizeHolder.y) )
				xLSExecute(hWnd, resizeCommand.c_str(), SW_SHOWNORMAL);
		}

		if (box != NULL)
			repaint();
	}

	if (type != 2 && movecommand && labelInitialized && !moveCommand.empty() && (newx != posHolder.x || newy != posHolder.y))
		xLSExecute(hWnd, moveCommand.c_str(), SW_SHOWNORMAL);
}

//------------------------------------------------------------
//Show/Hide
//------------------------------------------------------------
void Label::show(bool execute, int extraFeatureMode, GhostLabel *ghostlabel)
{
	if ( ParentDestroyNotificationReceived )
		return;

	if ( extraFeatureMode == 2 )
	{
		if ( ghostlabel != NULL )
		{
			ghostlabel->visible = true;
			repaint(true);
		}
	}
	else if ( extraFeatureMode == 1 )
	{
		if ( hWndTextBox )
		{
			ShowWindow(hWndTextBox, SW_SHOW);
			if ( !textboxVisible )
			{
				textboxVisible = true;

				SetForegroundWindow(hWndTextBox);
				SetFocus(hWndTextBox);

				repaint();

				PrefixedLSSetVariable(name, "TextEditBoxVisible", "true");
			}
			else
			{
				SetForegroundWindow(hWndTextBox);
				SetFocus(hWndTextBox);
			}
		}
	}
	else
	{
		KillTimer(hWnd, TIMER_AUTOHIDE);
		autoHideIsRunning = false;

		if (alphaFade && !visible)
		{
			KillTimer(hWnd, TIMER_ALPHAFADEUPBANG);
			KillTimer(hWnd, TIMER_ALPHAFADEDOWNBANG);

			KillTimer(hWnd, TIMER_ALPHAFADEOUT);
			UpdateAlpha = alphaTransparency;
			SetTimer(hWnd, TIMER_ALPHAFADEIN, alphaFadeDelay, NULL);
		}
		else
		{
			ShowWindow(hWnd, SW_SHOWNOACTIVATE);
		}

		if (!showCommand.empty() && execute && !visible)
		{
			xLSExecute(hWnd, showCommand.c_str(), SW_SHOWNORMAL);
		}

		visible = true;

		PrefixedLSSetVariable(name, "Visible", "true");

		if (normalAnim)
		{
			frameCounter = 1;
			loopCounter = 0;
			SetTimer(hWnd, TIMER_FRAMEANIMATION, imageFrameDelay, NULL);
		}

		//Very important, but i don't know why :)
		repaint(true);

		//Call Initial Zorder
		if ( !afterClass.empty() )
		{
			HWND before = FindWindowEx2(afterClass, afterTitle);
			
			if ( before != NULL )
			{
				disablezOrderChange = false;

				SetWindowPos(hWnd, before,
								0, 0, 0, 0,
								SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE );
			}
		}
	}
}

void Label::showHide(int timeout, bool execute)
{
	if ( autoHideIsRunning )
	{
		KillTimer(hWnd, TIMER_AUTOHIDE);
		autoHideIsRunning = false;
	}
	else
	{
		if ( !visible )
			show(execute);
		else
			hide(execute);
	}

	SetTimer(hWnd, TIMER_AUTOHIDE, timeout, NULL);

	autoHideIsRunning = true;
}

void Label::hide(bool execute, int extraFeatureMode, GhostLabel *ghostlabel)
{
	if ( extraFeatureMode == 2 )
	{
		if ( ghostlabel != NULL )
		{
			ghostlabel->visible = false;
			repaint(true);
		}
	}
	else if ( extraFeatureMode == 1 )
	{
		if ( hWndTextBox )
		{
			SetWindowText(hWndTextBox,"");
			ShowWindow(hWndTextBox, SW_HIDE);
			textboxVisible = false;
			repaint();

			PrefixedLSSetVariable(name, "TextEditBoxVisible", "false");
		}
	}
	else
	{
		KillTimer(hWnd, TIMER_AUTOHIDE);
		autoHideIsRunning = false;

		if (alphaFade && visible)
		{
			KillTimer(hWnd, TIMER_ALPHAFADEUPBANG);
			KillTimer(hWnd, TIMER_ALPHAFADEDOWNBANG);

			KillTimer(hWnd, TIMER_ALPHAFADEIN);
			UpdateAlpha = alphaTransparency;
			SetTimer(hWnd, TIMER_ALPHAFADEOUT, alphaFadeDelay, NULL);
		}
		else
		{
			ShowWindow(hWnd, SW_HIDE);
		}

		if (!hideCommand.empty() && execute && visible)
		{
			xLSExecute(hWnd, hideCommand.c_str(), SW_SHOWNORMAL);
		}

		visible = false;

		PrefixedLSSetVariable(name, "Visible", "false");

		if (normalAnim || hoverAnim)
		{
			KillTimer(hWnd, TIMER_FRAMEANIMATION);
		}
	}
}

//------------------------------------------------------------
//Clipboard Functions
//------------------------------------------------------------

void Label::clipboardCopy(string prefix)
{
	string toCopy = prefix.empty() ? text : (prefix + " " + text);
	SetClipboardText(hWnd, toCopy.c_str());
}

void Label::clipboardPaste(string prefix)
{
	char buffer[256] = {0};

	if(GetClipboardText(hWnd, buffer, 256))
	{
		string toPaste = prefix.empty() ? string(buffer) : (prefix + " " + string(buffer));
		setText(toPaste);
	}
}

//------------------------------------------------------------
//Region Events (Checking)
//------------------------------------------------------------

bool Label::checkHREFRegions(int x, int y, StringList inputlist)
{
	bool found = false;
	if (!inputlist.empty())
	{	
		char left[16], top[16], right[16], bottom[16], hreftooltip[MAX_LINE_LENGTH], command[MAX_LINE_LENGTH];
		char *buffers[] = {left, top, right, bottom, hreftooltip};

		int intleft = 0;
		int intright = 0;
		int inttop = 0;
		int intbottom = 0;
		string strTooltip = "";

		for(StringListIterator sit = inputlist.begin(); sit != inputlist.end(); sit++)
		{
			if(LCTokenize((*sit).c_str(), buffers, 5, command) >= 5)
			{
				intleft = atoi(left);
				intright = atoi(right);
				inttop = atoi(top);
				intbottom  = atoi(bottom);
				strTooltip = hreftooltip;

				if( x >= intleft && x <= intright && y >= inttop && y <= intbottom )
				{
					if (!handcursor)
					{
						if ( !IsOS(OS_2KXP) )
						{
							// Get the windows directory
							CString strWndDir;
							GetWindowsDirectory(strWndDir.GetBuffer(MAX_LINE_LENGTH), MAX_LINE_LENGTH);
							strWndDir.ReleaseBuffer();

							strWndDir += _T("\\winhlp32.exe");

							// This retrieves cursor #106 from winhlp32.exe, which is a hand pointer
							HMODULE hModule = LoadLibrary(strWndDir);
								
							if (hModule)
								handcursor = ::LoadCursor(hModule, MAKEINTRESOURCE(106));
					
							FreeLibrary(hModule);
						}
						else
						{
							handcursor = LoadCursor( 0, MAKEINTRESOURCE(32649) );
						}

						SetCursor(handcursor);
					}
					else
					{
						SetCursor(handcursor);
					}

					if ( !TooltipHints && !strTooltip.empty() )
						InitializeTooltip(true);

					if ( TooltipHints && !strTooltip.empty() )
					{
						strcpy(tooltip, strTooltip.c_str());
						updateHint(hWnd, tooltip);
					}
						
					//Processing Hover Change
					if (hrefconfig.size() == 2 && !mouseInsideHREF)
					{
						CString textprocess = originalText[current].c_str();
						CString textoutput = originalText[current].c_str();
						CString commandprocess = command;

						textprocess.MakeLower();
						commandprocess.MakeLower();

						int pos = 0;
						textprocess.Replace("&quot;", "\"");
						textprocess.Replace("&lt;", "<");
						textprocess.Replace("&gt;", ">");
						textprocess.Replace("&amp;", "&");
						//textprocess.Replace("%~", ";");
						//textprocess.Replace("[semicolon]", ";"); 
						textoutput.Replace("&quot;", "\"");
						textoutput.Replace("&lt;", "<");
						textoutput.Replace("&gt;", ">");
						textoutput.Replace("&amp;", "&");
						//textoutput.Replace("%~", ";");
						//textoutput.Replace("[semicolon]", ";"); 

						commandprocess = commandprocess.Right(commandprocess.GetLength()-commandprocess.Find("'>"));
						pos = textprocess.Find(commandprocess);
						if ( pos != -1 )
						{
							textprocess.Insert(pos+2, hrefconfig[0].c_str());
							textoutput.Insert(pos+2, hrefconfig[0].c_str());
							pos = textprocess.Find("</a>", pos+2);
							if ( pos != -1 )
							{
								textprocess.Insert(pos, hrefconfig[1].c_str());
								textoutput.Insert(pos, hrefconfig[1].c_str());

								textoutput.Replace("\"", "");

								setText((string)textoutput);
									
								mouseInsideHREF = true;
							}
						}
					}
					//End

					found = true;
					break;
				}
			}
		}
	}

	if ( !found && TooltipHints )
	{
		bool tmpdynamicTooltip = false;
		string restore = systemInfo->processLabelText(originalTooltip, this, &tmpdynamicTooltip);
		if ( _stricmp(tooltip, restore.c_str()) != NULL )
		{
			strcpy(tooltip, restore.c_str());
			updateHint(hWnd, tooltip);
		}
	}

	return found;
}

bool Label::checkHREFClickRegions(int x, int y, StringList inputlist)
{
	bool found = false;
	if (!inputlist.empty())
	{	
		char left[16], top[16], right[16], bottom[16], command[MAX_LINE_LENGTH];
		char *buffers[] = {left, top, right, bottom};

		int intleft = 0;
		int intright = 0;
		int inttop = 0;
		int intbottom = 0;

		for(StringListIterator sit = inputlist.begin(); sit != inputlist.end(); sit++)
		{
			if(LCTokenize((*sit).c_str(), buffers, 4, command) >= 4)
			{
				intleft = atoi(left);
				intright = atoi(right);
				inttop = atoi(top);
				intbottom  = atoi(bottom);

				if( x >= intleft && x <= intright && y >= inttop && y <= intbottom )
				{	
					CString process = command;
					process = process.Right(process.GetLength()-process.Find("<'")-2);
					process = process.Left(process.Find("'>"));

					process.Replace("&quot;", "\"");
					process.Replace("&lt;", "<");
					process.Replace("&gt;", ">");
					process.Replace("&amp;", "&");
			
					xLSExecute(hWnd, process, SW_SHOWNORMAL);

					found = true;
					break;	
				}
			}
		}
	}
	return found;
}

bool Label::checkRegions(int x, int y, Label_RegionList inputlist, bool btooltip)
{
	bool found = false;
	if (!inputlist.empty())
	{
		if ( btooltip && TooltipHints )
		{	
			POINT pt = {x,y};
			for(Label_RegionListIterator lrit = inputlist.begin(); lrit != inputlist.end(); lrit++)
			{
				if( PtInRect( &(*lrit)->region, pt) )
				{	
					if ( !(*lrit)->tooltip.empty() && _stricmp(tooltip, (*lrit)->tooltip.c_str()) != NULL )
					{
						strcpy(tooltip, (*lrit)->tooltip.c_str());
						updateHint(hWnd, tooltip);
					}

					found = true;
					break;	
				}
			}

			if ( !found )
			{
				bool tmpdynamicTooltip = false;
				string restore = systemInfo->processLabelText(originalTooltip, this, &tmpdynamicTooltip);
				if ( _stricmp(tooltip, restore.c_str()) != NULL )
				{
					strcpy(tooltip, restore.c_str());
					updateHint(hWnd, tooltip);
				}
			}
		}
		else if ( !btooltip )
		{
			POINT pt = {x,y};
			for(Label_RegionListIterator lrit = inputlist.begin(); lrit != inputlist.end(); lrit++)
			{
				if( PtInRect( &(*lrit)->region, pt) )
				{	
					if ( (*lrit)->ghostlabel != NULL )
					{
						if ( (*lrit)->ghostlabel->visible && !(*lrit)->ghostlabel->isHover && !(*lrit)->ghostlabel->isPressed )
						{
							(*lrit)->ghostlabel->isHover = true;
							found = true;
							if ( !(*lrit)->ghostlabel->mouseInside )
							{
								(*lrit)->ghostlabel->mouseInside = true;
								if ( !(*lrit)->ghostlabel->enterCommand.empty() )
									xLSExecute(hWnd, (*lrit)->ghostlabel->enterCommand.c_str(), SW_SHOWNORMAL);
							}
						}
					}
				}
				else 
				{
					if ( (*lrit)->ghostlabel != NULL )
					{
						if ( (*lrit)->ghostlabel->visible && ((*lrit)->ghostlabel->isHover || (*lrit)->ghostlabel->isPressed) )
						{
							(*lrit)->ghostlabel->isHover = false;
							(*lrit)->ghostlabel->isPressed = false;
							found = true;
							if ( (*lrit)->ghostlabel->mouseInside )
							{
								(*lrit)->ghostlabel->mouseInside = false;
								if ( !(*lrit)->ghostlabel->leaveCommand.empty() )
									xLSExecute(hWnd, (*lrit)->ghostlabel->leaveCommand.c_str(), SW_SHOWNORMAL);
							}
						}
					}
				}
			}
		}
	}
	return found;
}


bool Label::checkClickRegions(int x, int y, Label_RegionList inputlist, int checkPressedState)
{
	bool found = false;
	if (!inputlist.empty())
	{	
		POINT pt = {x,y};
		for(Label_RegionListIterator lrit = inputlist.begin(); lrit != inputlist.end(); lrit++)
		{
			if( PtInRect( &(*lrit)->region, pt) )
			{	
				if ( checkPressedState != -1 )
				{
					if ( (*lrit)->ghostlabel == NULL || ((*lrit)->ghostlabel != NULL && (*lrit)->ghostlabel->visible) )
					{
						if ( !(*lrit)->action.empty() )
							xLSExecute(hWnd, (*lrit)->action.c_str(), SW_SHOWNORMAL);
					}
				}

				if ( checkPressedState != 0)
				{
					if ( (*lrit)->ghostlabel != NULL )
					{
						if ( !(*lrit)->ghostlabel->isPressed && checkPressedState == -1 )
							(*lrit)->ghostlabel->isPressed = true;
						else if ( (*lrit)->ghostlabel->isPressed && checkPressedState == 1 )
							(*lrit)->ghostlabel->isPressed = false;
					}
				}

				found = true;
				if ( checkPressedState == 0 )
					break;	
			}
			else if ( checkPressedState != 0)
			{
				if ( (*lrit)->ghostlabel != NULL )
				{
					if ( (*lrit)->ghostlabel->isPressed )
						(*lrit)->ghostlabel->isPressed = false;
				}
			}
		}
	}
	return found;
}

//------------------------------------------------------------
//Mouse Related Actions (Clicks, Enter, Leave, Move, ...)
//------------------------------------------------------------

LRESULT Label::NcHitTest(int x, int y)
{
	if ( y >= 0 && y <= resizeBorderTop && x >= 0 && x <= resizeBorderLeft )
	{
		return HTTOPLEFT;
	}
	else if ( y >= 0 && y <= resizeBorderTop && x >= actualwidth-resizeBorderRight && x <= actualwidth )
	{	
		return HTTOPRIGHT;
	}	
	else if ( y >= 0 && y <= resizeBorderTop )
	{	
		return HTTOP;
	}	
	else if ( y >= actualheight-resizeBorderBottom && y <= actualheight && x >= 0 && x <= resizeBorderLeft )
	{	
		return HTBOTTOMLEFT;
	}	
	else if ( y >= actualheight-resizeBorderBottom && y <= actualheight && x >= actualwidth-resizeBorderRight && x <= actualwidth )
	{	
		return HTBOTTOMRIGHT;
	}	
	else if ( y >= actualheight-resizeBorderBottom && y <= actualheight )
	{	
		return HTBOTTOM;
	}	
	else if ( x >= 0 && x <= resizeBorderLeft )
	{	
		return HTLEFT;
	}	
	else if ( x >= actualwidth-resizeBorderRight && x <= actualwidth )
	{	
		return HTRIGHT;
	}	
	else
	{
		return 0;
	}
}

void Label::onLButtonDblClk(int x, int y)
{
	if ( !checkClickRegions(x, y, labelLeftDoubleClickRegions) ) 
	{
		if ( !CheckModClick(leftModDoubleClickList) )
		{
			if(!leftDoubleClickCommand.empty())
				xLSExecute(hWnd, leftDoubleClickCommand.c_str(), SW_SHOWNORMAL);
		}
	}
}

void Label::onLButtonDown(int x, int y)
{
	// 1 = LeftButton
	ontemplateButtonDown(x, y, 1, labelLeftClickRegions, leftClickDownCommand);
}

void Label::onLButtonUp(int x, int y)
{	
	// 1 = LeftButton
	ontemplateButtonUp(x, y, 1, labelLeftClickRegions, leftClickCommand, leftClickUpCommand);
}

void Label::onMButtonDblClk(int x, int y)
{
	if ( !checkClickRegions(x, y, labelMiddleDoubleClickRegions) ) 
	{
		if ( !CheckModClick(middleModDoubleClickList) )
		{
			if(!middleDoubleClickCommand.empty())
				xLSExecute(hWnd, middleDoubleClickCommand.c_str(), SW_SHOWNORMAL);
		}
	}
}

void Label::onMButtonDown(int x, int y)
{
	// 3 = MiddleButton
	ontemplateButtonDown(x, y, 3, labelMiddleClickRegions, middleClickDownCommand);
}

void Label::onMButtonUp(int x, int y)
{
	// 3 = MiddleButton
	ontemplateButtonUp(x, y, 3, labelMiddleClickRegions, middleClickCommand, middleClickUpCommand);
}

void Label::onRButtonDblClk(int x, int y)
{
	if ( !checkClickRegions(x, y, labelRightDoubleClickRegions) ) 
	{
		if ( !CheckModClick(rightModDoubleClickList) )
		{
			if(!rightDoubleClickCommand.empty())
				xLSExecute(hWnd, rightDoubleClickCommand.c_str(), SW_SHOWNORMAL);
		}
	}
}

void Label::onRButtonDown(int x, int y)
{
	// 2 = RightButton
	ontemplateButtonDown(x, y, 2, labelRightClickRegions, rightClickDownCommand);
}

void Label::onRButtonUp(int x, int y)
{
	// 2 = RightButton
	ontemplateButtonUp(x, y, 2, labelRightClickRegions, rightClickCommand, rightClickUpCommand);
}

void Label::onX1ButtonDblClk(int x, int y)
{
	if ( !checkClickRegions(x, y, labelX1DoubleClickRegions) ) 
	{
		if ( !CheckModClick(X1ModDoubleClickList) )
		{
			if(!X1DoubleClickCommand.empty())
				xLSExecute(hWnd, X1DoubleClickCommand.c_str(), SW_SHOWNORMAL);
		}
	}
}

void Label::onX1ButtonDown(int x, int y)
{
	// 4 = X1Button
	ontemplateButtonDown(x, y, 4, labelX1ClickRegions, X1ClickDownCommand);
}

void Label::onX1ButtonUp(int x, int y)
{
	// 4 = X1Button
	ontemplateButtonUp(x, y, 4, labelX1ClickRegions, X1ClickCommand, X1ClickUpCommand);
}

void Label::onX2ButtonDblClk(int x, int y)
{
	if ( !checkClickRegions(x, y, labelX2DoubleClickRegions) ) 
	{
		if ( !CheckModClick(X2ModDoubleClickList) )
		{
			if(!X2DoubleClickCommand.empty())
				xLSExecute(hWnd, X2DoubleClickCommand.c_str(), SW_SHOWNORMAL);
		}
	}
}

void Label::onX2ButtonDown(int x, int y)
{
	// 5 = X2Button
	ontemplateButtonDown(x, y, 5, labelX2ClickRegions, X2ClickDownCommand);
}

void Label::onX2ButtonUp(int x, int y)
{
	// 5 = X2Button
	ontemplateButtonUp(x, y, 5, labelX2ClickRegions, X2ClickCommand, X2ClickUpCommand);
}

void Label::onWheelDown(int x, int y)
{
	if(!wheelDownCommand.empty())
		xLSExecute(hWnd, wheelDownCommand.c_str(), SW_SHOWNORMAL);
}

void Label::onWheelUp(int x, int y)
{
	if(!wheelUpCommand.empty())
		xLSExecute(hWnd, wheelUpCommand.c_str(), SW_SHOWNORMAL);
}

void Label::onMouseEnter()
{
	SendMessage(hWnd, LM_SETCURSOR, 0, 0);

	if (useHover)
	{
		//Kill AnimTimer from Normal Animation
		if (normalAnim)
		{
			KillTimer(hWnd, TIMER_FRAMEANIMATION);
		}

		//If HoverAnim is set (hoverframecount != 0)
		if (hoverAnim)
		{
			frameCounter = 1;
			hoverActive = true;
			pressedActive = false;

			//AutoSize due to possible Font change
			if ( !text.empty() && (autoWidthMode != 0 || autoHeightMode != 0) )
				autoSize();

			loopCounter = 0;
			//Start AnimTimer from Hover Animation
			SetTimer(hWnd, TIMER_FRAMEANIMATION, hoverimageFrameDelay, NULL);
		}
		//If HoverAnim is not set (hoverframecount == 0)
		else
		{
			frameCounter = 0;
			hoverActive = true;
			pressedActive = false;

			//AutoSize due to possible Font change
			if ( !text.empty() && (autoWidthMode != 0 || autoHeightMode != 0) )
				autoSize();

			repaint(true);
		}
	}
	else
	{
		if ( font != NULL && font->useHoverText )
		{
			//AutoSize due to possible Font change
			if ( !text.empty() )
			{
				hoverActive = true;
				pressedActive = false;

				if ( autoWidthMode != 0 || autoHeightMode != 0 )
					autoSize();

				repaint();
			}
		}
	}
	
	if (!enterCommand.empty())
	{
		xLSExecute(hWnd, enterCommand.c_str(), SW_SHOWNORMAL);
	}
		
	if ( !hoverCommand.empty() )
	{
		TRACKMOUSEEVENT mouseEvent = {0};
		mouseEvent.cbSize = sizeof(TRACKMOUSEEVENT);
		mouseEvent.hwndTrack = hWnd;
		mouseEvent.dwFlags = TME_HOVER;
		mouseEvent.dwHoverTime = hoverTime;

		_TrackMouseEvent(&mouseEvent);
	}

	if (focusOnEnter)
	{	
		skipButton = true;
		mouse_event( MOUSEEVENTF_ABSOLUTE | MOUSEEVENTF_MIDDLEDOWN,
						NULL, NULL, NULL, NULL );
		mouse_event( MOUSEEVENTF_ABSOLUTE | MOUSEEVENTF_MIDDLEUP,
						NULL, NULL, NULL, NULL );
	}
}

void Label::onMouseLeave()
{
	bool repaintneeded = false;

	if (useHover || usePressed)
	{
		//If HoverAnim was active (hoverframecount != 0)
		if (hoverAnim)
		{	
			frameCounter = 0;
			hoverActive = false;
			pressedActive = false;

			//AutoSize due to possible Font change
			if ( !text.empty() && (autoWidthMode != 0 || autoHeightMode != 0) )
				autoSize();
		
			//Kill AnimTimer from Hover Animation
			KillTimer(hWnd, TIMER_FRAMEANIMATION);

			//If NormalAnim is set (framecount != 0)
			if (normalAnim)
			{
				frameCounter = 1;
				loopCounter = 0;
				SetTimer(hWnd, TIMER_FRAMEANIMATION, imageFrameDelay, NULL);
			}
			//If NormalAnim is not set (framecount == 0)
			else
			{
				repaintneeded = true;
			}
		}
		else
		{
			hoverActive = false;
			pressedActive = false;

			//AutoSize due to possible Font change
			if ( !text.empty() && (autoWidthMode != 0 || autoHeightMode != 0) )
				autoSize();

			//If NormalAnim is set (framecount != 0)
			if (normalAnim)
			{
				frameCounter = 1;
				SetTimer(hWnd, TIMER_FRAMEANIMATION, imageFrameDelay, NULL);
			}
			//If NormalAnim is not set (framecount == 0)
			else
			{
				repaintneeded = true;
			}
		}
	}
	else
	{
		if ( font != NULL && font->useHoverText )
		{
			//AutoSize due to possible Font change
			if ( !text.empty() )
			{	
				hoverActive = false;
				pressedActive = false;

				if ( autoWidthMode != 0 || autoHeightMode != 0 )
					autoSize();

				repaintneeded = true;
			}
		}
	}

	if ( !labelGhostRegions.empty() )
	{
		for (Label_RegionListIterator lrit = labelGhostRegions.begin(); lrit != labelGhostRegions.end(); lrit++)
		{
			if ( (*lrit)->ghostlabel->isHover || (*lrit)->ghostlabel->isPressed )
			{
				(*lrit)->ghostlabel->isHover = false;
				(*lrit)->ghostlabel->isPressed = false;
				if ( (*lrit)->ghostlabel->mouseInside )
				{
					(*lrit)->ghostlabel->mouseInside = false;
					if ( !(*lrit)->ghostlabel->leaveCommand.empty() )
						xLSExecute(hWnd, (*lrit)->ghostlabel->leaveCommand.c_str(), SW_SHOWNORMAL);
				}

				repaintneeded = true;
			}
		}
	}
	
	if ( repaintneeded )
		repaint(true);

	if (!leaveCommand.empty())
	{
		xLSExecute(hWnd, leaveCommand.c_str(), SW_SHOWNORMAL);
	}
}

void Label::onMouseMove(int x, int y)
{
	if (!moveable)
	{
		if (!mouseInside)
		{
			mouseInside = true;
			onMouseEnter();
			SetTimer(hWnd, TIMER_MOUSETRACK, 100, 0);
		}

		if (mouseInside2)
		{
			mouseInside2 = false;
			SendMessage(hWnd, LM_SETCURSOR, 0, 0);

			if ( !hoverCommand.empty() )
			{
				TRACKMOUSEEVENT mouseEvent = {0};
				mouseEvent.cbSize = sizeof(TRACKMOUSEEVENT);
				mouseEvent.hwndTrack = hWnd;
				mouseEvent.dwFlags = TME_HOVER;
				mouseEvent.dwHoverTime = hoverTime;

				_TrackMouseEvent(&mouseEvent);
			}
		}

		if (!labelHREFClickRegions.empty())
		{
			if (!checkHREFRegions(x, y, labelHREFClickRegions) && mouseInsideHREF)
			{
				if (!originalTextBackup.empty())
				{
					current = currentBackup;
					setText(originalTextBackup);
				}
				SendMessage(hWnd, LM_SETCURSOR, 0, 0);
				mouseInsideHREF = false;
			}
		}
		if (!labelTooltipRegions.empty() && TooltipHints)
		{
			checkRegions(x, y, labelTooltipRegions);
		}
		if (!ghostlabelList.empty())
		{
			if ( checkRegions(x, y, labelGhostRegions, false) )
				repaint(true);
		}
	}
	else if(moveable)
	{
		if (mousePressed && movemodifierkeyPressed && moveButtonPressed == moveButton)
		{
			KillTimer(hWnd, TIMER_MOUSETRACK);

			POINT newpt = {x,y};
			ClientToScreen(hWnd, &newpt);
			if ( box )
				ScreenToClient(box, &newpt);
			newpt.x = newpt.x-savedpt.x;
			newpt.y = newpt.y-savedpt.y;

			if ( !hasMoved && (newpt.x != actualx || newpt.y != actualy) )
				hasMoved = true;

			if ( fireOnMoveDuringMove )
				reposition(1, newpt.x, newpt.y, 0, 0);
			else
				reposition(1, newpt.x, newpt.y, 0, 0, 0, 0, false, false);

			if ( hasMoved && !noMoveCursor)
				SendMessage(hWnd, LM_SETCURSOR, 0, (LONG) IDC_SIZEALL);
		}
		else
		{
			SetTimer(hWnd, TIMER_MOUSETRACK, 100, 0);

			if (!mouseInside)
			{
				mouseInside = true;
				onMouseEnter();
			}

			if (mouseInside2)
			{
				mouseInside2 = false;
				SendMessage(hWnd, LM_SETCURSOR, 0, 0);

				if ( !hoverCommand.empty() )
				{
					TRACKMOUSEEVENT mouseEvent = {0};
					mouseEvent.cbSize = sizeof(TRACKMOUSEEVENT);
					mouseEvent.hwndTrack = hWnd;
					mouseEvent.dwFlags = TME_HOVER;
					mouseEvent.dwHoverTime = hoverTime;

					_TrackMouseEvent(&mouseEvent);
				}
			}

			if (!labelHREFClickRegions.empty())
			{
				if (!checkHREFRegions(x, y, labelHREFClickRegions) && mouseInsideHREF)
				{
					if (!originalTextBackup.empty())
					{
						current = currentBackup;
						setText(originalTextBackup);
					}
					SendMessage(hWnd, LM_SETCURSOR, 0, 0);
					mouseInsideHREF = false;
				}
			}
			if (!labelTooltipRegions.empty() && TooltipHints)
			{
				checkRegions(x, y, labelTooltipRegions);
			}
			if (!ghostlabelList.empty())
			{
				if ( checkRegions(x, y, labelGhostRegions, false) )
					repaint(true);
			}
		}
	}
}

//------------------------------------------------------------
//AutoWidthMode and AutoHeightMode Implementation
//------------------------------------------------------------
void Label::autoSize()
{ 
	RECT rectPosition;
	rectPosition.left = actualx;
	rectPosition.right = actualx+actualwidth;
	rectPosition.top = actualy;
	rectPosition.bottom = actualy+actualheight;

	int width = (rectPosition.right-rectPosition.left);
	int height = (rectPosition.bottom-rectPosition.top);

	if ( width < 0 || height < 0 )
		return;

	int textWidth = 0;
	int textHeight = 0;

	int imageType = 0;
	if (pressedActive && font->usePressedText)
		imageType = 2;
	else if (hoverActive && font->useHoverText)
		imageType = 1;

	//Create Tmp BufferDC
	HDC	bufferDC = CreateCompatibleDC(NULL);
	HBITMAP bufferBitmap = CreateCompatibleBitmap(bufferDC, width, height);
	HBITMAP oldBMP = (HBITMAP) SelectObject(bufferDC, bufferBitmap);
		
	POINT size = font->measure(bufferDC, leftBorder, width-rightBorder, text, imageType);

	textWidth = size.x+6;
	textHeight = size.y+6;

	if ( font->getLineBreak() && textWidth+leftBorder+rightBorder >= autoMaxWidth )
	{
		size = font->measure(bufferDC, leftBorder, autoMaxWidth-rightBorder, text, imageType);
	
		textWidth = size.x+6;
		textHeight = size.y+6;
	}
	else if ( font->getLineBreak() && textWidth+leftBorder+rightBorder <= autoMinWidth )
	{
		size = font->measure(bufferDC, leftBorder, autoMinWidth-rightBorder, text, imageType);
	
		textWidth = size.x+6;
		textHeight = size.y+6;
	}

	bufferBitmap = (HBITMAP) SelectObject(bufferDC, oldBMP);
	DeleteObject(bufferBitmap);
	DeleteDC(bufferDC);
	
	if ( oldWidth == textWidth && oldHeight == textHeight )
	{
		return;
	}
	else
	{
		oldWidth = textWidth;
		oldHeight = textHeight;
	}

	//For AUTOWIDTH/HEIGHT MODE	
	int newx, newy, newwidth, newheight, deltax, deltay;

	if (autoWidthMode != 0)
		deltax = textWidth - (width - leftBorder - rightBorder);
	else
		deltax = 0;

	if (autoHeightMode != 0)
		deltay = textHeight - (height - topBorder - bottomBorder);
	else
		deltay = 0;

	newx = rectPosition.left;
	newy = rectPosition.top;
	newwidth = width + deltax;
	newheight = height + deltay;

	//Check MIN/MAX
	if (newwidth < autoMinWidth)
	{
		deltax = deltax + (autoMinWidth-newwidth);
		newwidth = autoMinWidth;
	}
	else if (newwidth > autoMaxWidth)
	{
		deltax = deltax - (newwidth-autoMaxWidth);
		newwidth = autoMaxWidth;
	}

	if (newheight < autoMinHeight)
	{
		deltay = deltay + (autoMinHeight-newheight);
		newheight = autoMinHeight;
	}
	else if (newheight > autoMaxHeight)
	{
		deltay = deltay - (newheight-autoMaxHeight);
		newheight = autoMaxHeight;
	}

	if ((autoWidthMode == 0 || autoWidthMode == 1) && (autoHeightMode == 0 || autoHeightMode == 1))
	{
		newx = rectPosition.left;
		newy = rectPosition.top;
	}
	else if ((autoWidthMode == 0 || autoWidthMode == 1) && autoHeightMode == 2)
	{
		newx = rectPosition.left;
		newy = rectPosition.top-deltay;
	}
	else if ((autoWidthMode == 0 || autoWidthMode == 1) && autoHeightMode == 3)
	{
		newx = rectPosition.left;
		newy = startCenterY-(newheight/2);
	}
	else if (autoWidthMode == 2 && (autoHeightMode == 0 || autoHeightMode == 1))
	{
		newx = rectPosition.left-deltax;
		newy = rectPosition.top;
	}
	else if (autoWidthMode == 2 && autoHeightMode == 2)
	{
		newx = rectPosition.left-deltax;
		newy = rectPosition.top-deltay;
	}
	else if (autoWidthMode == 2 && autoHeightMode == 3)
	{
		newx = rectPosition.left-deltax;
		newy = startCenterY-(newheight/2);
	}
	else if (autoWidthMode == 3 && (autoHeightMode == 0 || autoHeightMode == 1))
	{
		newx = startCenterX-(newwidth/2);
		newy = rectPosition.top;
	}
	else if (autoWidthMode == 3 && autoHeightMode == 2)
	{
		newx = startCenterX-(newwidth/2);
		newy = rectPosition.top-deltay;
	}
	else if (autoWidthMode == 3 && autoHeightMode == 3)
	{
		newx = startCenterX-(newwidth/2);
		newy = startCenterY-(newheight/2);
	}

	reposition(3, newx, newy, newwidth, newheight, autoAnimSteps, autoAnimDelay);
}

void Label::repaint(bool deleteScrollDC)
{
	if ( deleteScrollDC )
	{
		if ( scrollBitmap )
		{
			scrollBitmap = (HBITMAP) SelectObject(scrollDC, oldscrollBMP);
			DeleteObject(scrollBitmap);
		}
		if ( scrollDC )
			DeleteObject(scrollDC);

		scrollBitmap = 0;
		scrollDC = 0;
	}

	if ( backupBitmap )
	{
		backupBitmap = (HBITMAP) SelectObject(backupDC, oldbackupBMP);
		DeleteObject(backupBitmap);
	}
	if ( backupDC )
		DeleteObject(backupDC);

	backupBitmap = 0;
	backupDC = 0;

	if ( alphaMap || fakealphaMap ) 
	{
		//Attention!! Background MUST exist
		if ( background && font )
			UpdateLabelAlpha(alphaTransparency);
	}
	else
	{
		RedrawWindow(hWnd, NULL, NULL, RDW_INVALIDATE | RDW_UPDATENOW);

		if ( transparencyMode == 1 && transparencyNeeded && labelInitialized )
			updateRegion();

		if ( hWndTextBox && textboxVisible )
		{
			if ( hTextBGBrush )
			{
				DeleteObject( hTextBGBrush );
				hTextBGBrush = 0;
			}
		
			InvalidateRect(hWndTextBox, NULL, true);
		}
	}
}

void Label::updateRegion()
{	
	RECT rectPosition;
	GetClientRect(hWnd, &rectPosition);

	HDC tmpDC = CreateCompatibleDC(NULL);
	HBITMAP tmpBmp = CreateBitmap(rectPosition.right, rectPosition.bottom, 1, bitsperpixel, NULL);
	HBITMAP oldBMP = (HBITMAP) SelectObject(tmpDC, tmpBmp);

	//Draw Image
	onPaint(tmpDC);

	tmpBmp = (HBITMAP) SelectObject(tmpDC, oldBMP);

	HRGN region = BitmapToRegion(tmpBmp, RGB(255, 0, 255), 0, 0, 0);
	HRGN windowRgn = CreateRectRgn(0, 0, 0, 0);
	CombineRgn(windowRgn, region, NULL, RGN_COPY);

	SetWindowRgn(hWnd, windowRgn, true);	

	DeleteObject(windowRgn);
	DeleteObject(region);
	DeleteObject(tmpBmp);
	DeleteDC(tmpDC);
}

bool Label::FindDockedLabel(Label* findLabel)
{
	for(LabelListIterator lit = hasDockedLabelList.begin(); lit != hasDockedLabelList.end(); lit++)
	{
		if ( (*lit) == findLabel )
			return true;
	}

	return false;
}

void Label::onWindowPosChanging(WINDOWPOS *windowPos)
{
	if ( (windowPos->flags & SWP_NOZORDER) == 0 )
	{
		if ( disablezOrderChange )
		{
			windowPos->flags = windowPos->flags | SWP_NOZORDER;
		}	
	}
	
	disablezOrderChange = true;

	if ( useresizeBorder && resizeBorderResize && (windowPos->flags & SWP_NOSIZE) == 0 )
	{
		if ( resizeStepSize.x != 0 && (windowPos->cx-actualwidth)%resizeStepSize.x != 0 )
		{
			windowPos->x = actualx;
			windowPos->cx = actualwidth;
		}
		if ( resizeStepSize.y != 0 &&(windowPos->cy-actualheight)%resizeStepSize.y != 0 )
		{
			windowPos->y = actualy;
			windowPos->cy = actualheight;
		}
	}

	if ( (windowPos->flags & SWP_NOMOVE) != 0 )
		return;

	int result = 0;

	if ( mousePressed && movemodifierkeyPressed )
	{
		if ( moveRestriction == 1 )
		{
			// top edge
			if ( windowPos->y != actualy )
				windowPos->y = actualy;
		}
		else if ( moveRestriction == 2 )
		{
			// left edge
			if ( windowPos->x != actualx )
				windowPos->x = actualx;
		}

		//Moving of hooked Label out of the Parent isn't allowed
		if ( box || useCustomMoveArea )
		{
			int minx = 0;
			int miny = 0;
			int maxx = 0;
			int maxy = 0;
			if ( box && !useCustomMoveArea )
			{
				RECT boxPosition;
				GetClientRect(box, &boxPosition);
				maxx = boxPosition.right;
				maxy = boxPosition.bottom;
			}
			else if ( useCustomMoveArea )
			{
				minx = customMoveArea.left;
				miny = customMoveArea.top;
				maxx = customMoveArea.right;
				maxy = customMoveArea.bottom;
			}

			// left edge
			if ( windowPos->x < minx )
				windowPos->x = minx;

			// top edge
			if ( windowPos->y < miny )
				windowPos->y = miny;

			// bottom edge
			if ( windowPos->y+actualheight > maxy)
				windowPos->y = maxy-actualheight;

			// right edge
			if ( windowPos->x+actualwidth > maxx)
				windowPos->x = maxx-actualwidth;

			if (box)
				return;
		}

		//Check Snap to Borders
		if ( snapDistance != 0 )
		{
			POINT pos = {windowPos->x, windowPos->y};
			POINT size = {actualwidth, actualheight};

			//Check First WorkArea Border
			if ( (_stricmp("alllabels", snapTo.c_str()) == NULL || _stricmp("desktop", snapTo.c_str()) == NULL) )
			{
				RECT rScreen;
				if ( realMonitors > 1 )
				{
					MONITORINFO mi = { sizeof(MONITORINFO) };
					GetMonitorInfo(MonitorFromWindow(hWnd, MONITOR_DEFAULTTONEAREST), &mi);

					rScreen = mi.rcWork;
				}
				else
				{
					SystemParametersInfo(SPI_GETWORKAREA, 0, &rScreen, NULL);
				}

				result = isCloseTo( pos, size, rScreen.left, rScreen.top, rScreen.right-rScreen.left, rScreen.bottom-rScreen.top, snapDistance);
			
				if (result == 1)
				{
					windowPos->x = pos.x;
				}
				else if (result == 2)
				{	
					windowPos->y = pos.y;
				}
				else if (result == 3)
				{
					windowPos->x = pos.x;
					windowPos->y = pos.y;
				}
							
				if ( _stricmp("desktop", snapTo.c_str()) == NULL )
					return;
			}

			result = 0;
			string Bang = "";

			bool skipX = false;
			bool skipY = false;

			if ( moveRestriction == 2 )
				skipX = true;
			else if ( moveRestriction == 1 )
				skipY = true;

			for( LabelListIterator lit = labelList.begin(); lit != labelList.end(); lit++)
			{
				if ( (*lit)->getBox() == NULL && (*lit)->isVisible() && _stricmp(name.c_str(), (*lit)->getName().c_str()) != NULL)
				{
					if ( _stricmp((*lit)->getName().c_str(), snapTo.c_str()) == NULL )
					{
						result = isCloseTo( pos, size, (*lit)->actualx, (*lit)->actualy, (*lit)->actualwidth, (*lit)->actualheight, snapDistance, skipX, skipY);
						if ( result == 0 )
						{
							if ( dockOnSnapTo == 1 )
							{
								if ( (*lit)->FindDockedLabel(this) )
								{
									Bang = (name + " " + (*lit)->getName() + " false");
									snapToList.insert(snapToList.end(), Bang);
								}
							}
							else if ( dockOnSnapTo == 2 )
							{
								if ( FindDockedLabel( (*lit) ) )
								{
									Bang = ((*lit)->getName() + " " + name + " false");
									snapToList.insert(snapToList.end(), Bang);
								}
							}
							else if ( dockOnSnapTo == 3 )
							{
								if ( (*lit)->FindDockedLabel(this) )
								{
									Bang = (name + " " + (*lit)->getName() + " false");
									snapToList.insert(snapToList.end(), Bang);
								}
								if ( FindDockedLabel( (*lit) ) )
								{
									Bang = ((*lit)->getName() + " " + name + " false");
									snapToList.insert(snapToList.end(), Bang);	
								}
							}
						}
					}
					if ( _stricmp((*lit)->labelGroup.c_str(), snapTo.c_str()) == NULL || _stricmp("alllabels", snapTo.c_str()) == NULL )
					{
						result = isCloseTo( pos, size, (*lit)->actualx, (*lit)->actualy, (*lit)->actualwidth, (*lit)->actualheight, snapDistance, skipX, skipY);
						if ( result == 0 )
						{
							if ( dockOnSnapTo == 1 )
							{
								if ( (*lit)->FindDockedLabel(this) )
								{
									Bang = (name + " " + (*lit)->getName() + " false");
									snapToList.insert(snapToList.end(), Bang);
								}
							}
							else if ( dockOnSnapTo == 2 )
							{
								if ( FindDockedLabel( (*lit) ) )
								{
									Bang = ((*lit)->getName() + " " + name + " false");
									snapToList.insert(snapToList.end(), Bang);
								}
							}
							else if ( dockOnSnapTo == 3 )
							{
								if ( (*lit)->FindDockedLabel(this) )
								{
									Bang = (name + " " + (*lit)->getName() + " false");
									snapToList.insert(snapToList.end(), Bang);
								}
								if ( FindDockedLabel( (*lit) ) )
								{
									Bang = ((*lit)->getName() + " " + name + " false");
									snapToList.insert(snapToList.end(), Bang);
								}
							}
						}
					}
					if (result != 0)
					{
						if ( dockOnSnapTo == 1 )
						{
							Bang = (name + " " + (*lit)->getName());
							snapToList.insert(snapToList.end(), Bang);
						}
						else if ( dockOnSnapTo == 2 )
						{
							Bang = ((*lit)->getName() + " " + name);
							snapToList.insert(snapToList.end(), Bang);
						}
						else if ( dockOnSnapTo == 3 )
						{
							Bang = (name + " " + (*lit)->getName());
							snapToList.insert(snapToList.end(), Bang);
							
							Bang = ((*lit)->getName() + " " + name);
							snapToList.insert(snapToList.end(), Bang);
						}

						if ( result == 1 )
						{
							skipX = true;
						}
						else if (result == 2 )
						{
							skipY = true;
						}
						else if (result == 3 )
						{
							skipX = true;
							skipY = true;
						}
					}

					if (skipX && skipY)	//Only Break, if both direction have snapped
						break;
				}
			}
				
			if (result == 1)
			{
				windowPos->x = pos.x;
			}
			else if (result == 2)
			{	
				windowPos->y = pos.y;
			}
			else if (result == 3)
			{
				windowPos->x = pos.x;
				windowPos->y = pos.y;
			}
		}
	}
}

void Label::onMove(int x, int y)
{
	//Docked Label Must Move (MoveBy)
	if ( !hasDockedLabelList.empty() )
	{
		for(LabelListIterator lit = hasDockedLabelList.begin(); lit != hasDockedLabelList.end(); lit++)
		{
			if ( (*lit) == this || (*lit) == sender )
				continue;
			
			(*lit)->reposition(1, x-oldPositionBackup.x, y-oldPositionBackup.y, 0, 0, 0, 0, true, true, this);
		}
	}

	oldPositionBackup.x = x;
	oldPositionBackup.y = y;
	actualx = x;
	actualy = y;

	if (autoWidthMode != 0 || autoHeightMode != 0)
	{
		startCenterX = actualx+(actualwidth/2);
		startCenterY = actualy+(actualheight/2);
	}

	if ( labelInitialized )
	{
		PrefixedLSSetVariable(name, "CurrentX", IntToString( actualx ));
		PrefixedLSSetVariable(name, "CurrentY", IntToString( actualy ));
	}

	if ( transparencyMode == 2 && transparencyNeeded )
	{
		repaint(true);
	}

	if ( hasHookedModules )
		RedrawWindow(hWnd, NULL, NULL, RDW_UPDATENOW | RDW_ALLCHILDREN);
}

void Label::onSize(int width, int height)
{
	//Docked Label Must Move (MoveBy)
	if ( !hasDockedLabelList.empty() )
	{
		for(LabelListIterator lit = hasDockedLabelList.begin(); lit != hasDockedLabelList.end(); lit++)
		{
			if ( (*lit) == this || (*lit) == sender || (*lit)->dockedtoSide == 0 || (*lit)->dockedtoSide == 1 || (*lit)->dockedtoSide == 4 )
				continue;

			if ( (*lit)->dockedtoSide == 2 && width == actualwidth )
				continue;
			else if ( (*lit)->dockedtoSide == 3 && height == actualheight )
				continue;
			else if ( (*lit)->dockedtoSide == 5 && width == actualwidth )
				continue;
			else if ( (*lit)->dockedtoSide == 6 && height == actualheight )
				continue;
			else if ( (*lit)->dockedtoSide == 7 && width == actualwidth && height == actualheight)
				continue;
			
			(*lit)->reposition(1, width-oldSizeBackup.x, height-oldSizeBackup.y, 0, 0, 0, 0, true, true, this);
		}
	}

	oldSizeBackup.x = width;
	oldSizeBackup.y = height;
	actualwidth = width;
	actualheight = height;

	if (autoWidthMode != 0 || autoHeightMode != 0)
	{
		startCenterX = actualx+(actualwidth/2);
		startCenterY = actualy+(actualheight/2);
	}

	if ( hWndTextBox )
	{
		TextEditBoxSizeBorder.left = xParseCoord(textboxXOrg, 0, actualwidth, hWnd);
		TextEditBoxSizeBorder.right = TextEditBoxSizeBorder.left+xParseDimen(textboxWidthOrg, actualwidth, actualwidth, hWnd);
		TextEditBoxSizeBorder.top = xParseCoord(textboxYOrg, 0, actualheight, hWnd);
		TextEditBoxSizeBorder.bottom = TextEditBoxSizeBorder.top+textboxHeight;

		TextEditBoxSize.left = TextEditBoxSizeBorder.left+tebleftBorder;
		TextEditBoxSize.right = TextEditBoxSizeBorder.right-tebrightBorder;
		TextEditBoxSize.top = TextEditBoxSizeBorder.top+tebtopBorder;
		TextEditBoxSize.bottom = TextEditBoxSizeBorder.bottom-tebbottomBorder;

		SetWindowPos(hWndTextBox, NULL, 
						TextEditBoxSize.left,
						TextEditBoxSize.top,
						(TextEditBoxSize.right-TextEditBoxSize.left), 
						(TextEditBoxSize.bottom-TextEditBoxSize.top), 
						SWP_NOACTIVATE | SWP_NOZORDER);
	}

	if (labelInitialized)
	{
		PrefixedLSSetVariable(name, "CurrentWidth", IntToString( width ));
		PrefixedLSSetVariable(name, "CurrentHeight", IntToString( height ));

		if ( !ghostlabelList.empty() )
		{
			for(GhostLabelListIterator glit = ghostlabelList.begin(); glit != ghostlabelList.end(); glit++)
			{
				if ( *glit != NULL )
				{
					(*glit)->x = xParseCoord((*glit)->orgx, 0, actualwidth, hWnd);
					(*glit)->y = xParseCoord((*glit)->orgy, 0, actualheight, hWnd);

					(*glit)->width = xParseDimen((*glit)->orgwidth, 0, actualwidth, hWnd);
					(*glit)->height = xParseDimen((*glit)->orgheight, 0, actualheight, hWnd);
				}
				else
				{
					break;
				}
			}
		}

		UpdateLabel_Regions( labelLeftClickRegions );
		UpdateLabel_Regions( labelLeftDoubleClickRegions );
		
		UpdateLabel_Regions( labelMiddleClickRegions );
		UpdateLabel_Regions( labelMiddleDoubleClickRegions );

		UpdateLabel_Regions( labelRightClickRegions );
		UpdateLabel_Regions( labelRightDoubleClickRegions );

		UpdateLabel_Regions( labelX1ClickRegions );
		UpdateLabel_Regions( labelX1DoubleClickRegions );

		UpdateLabel_Regions( labelX2ClickRegions );
		UpdateLabel_Regions( labelX2DoubleClickRegions );

		UpdateLabel_Regions( labelTooltipRegions );

		UpdateLabel_Regions( labelGhostRegions );
	}

	repaint(true);

	//Set new Rect for ToolTip Tracker
	if ( TooltipHints )
		updateHint(hWnd, tooltip, true);

	if ( fireOnResizeDuringResize && !resizeCommand.empty() )
		xLSExecute(hWnd, resizeCommand.c_str(), SW_SHOWNORMAL);

	if ( hasHookedModules )
		RedrawWindow(hWnd, NULL, NULL, RDW_UPDATENOW | RDW_ALLCHILDREN);

	resizeBorderResize = true;
}

void Label::relayMouseMessageToBox(UINT message, WPARAM wParam, LPARAM lParam)
{
	POINT pt;
	pt.x = GET_X_LPARAM(lParam);
	pt.y = GET_Y_LPARAM(lParam);

	if (box != NULL)
	{
		MapWindowPoints(hWnd, box, &pt, 1);
		PostMessage(box, message, wParam, MAKELPARAM((short) pt.x, (short) pt.y));
	}
	else
	{
		MapWindowPoints(hWnd, GetLitestepDesktop(), &pt, 1);
		PostMessage(GetLitestepDesktop(), message, wParam, MAKELPARAM((short) pt.x, (short) pt.y));
	}
}

void Label::mouseTrack()
{
	POINT pt;
	GetCursorPos(&pt);
			
	RECT rc;
	GetWindowRect(hWnd, &rc);

	if ( transparencyMode == 1 && transparencyNeeded && !trackingIgnoreTrueTransparency )
	{
		HRGN hrgn = CreateRectRgn(0,0,0,0);
		int regionType = GetWindowRgn(hWnd, hrgn);
		if (regionType != ERROR) 
		{ 
			/* hrgn contains window region */ 
			ScreenToClient(hWnd, &pt);
			if(!PtInRegion(hrgn, pt.x, pt.y))
			{
				KillTimer(hWnd, TIMER_MOUSETRACK);
				mouseInside = false;
				mousePressed = false;
				if (!originalTextBackup.empty() && mouseInsideHREF)
				{
					current = currentBackup;
					setText(originalTextBackup);
				}
				mouseInsideHREF = false;
				onMouseLeave();	

				DeleteObject(hrgn); /* finished with region */

				return;
			}
			else if ( ChildWindowFromPoint(hWnd, pt) != NULL )
			{
				mouseInside2 = true;

				DeleteObject(hrgn); /* finished with region */

				return;
			}
		}
		DeleteObject(hrgn); /* finished with region */
	}
	
	if(!PtInRect(&rc, pt))
	{
		KillTimer(hWnd, TIMER_MOUSETRACK);
		mouseInside = false;
		mousePressed = false;
		if (!originalTextBackup.empty() && mouseInsideHREF)
		{
			current = currentBackup;
			setText(originalTextBackup);
		}
		mouseInsideHREF = false;
		onMouseLeave();	
	}
	else
	{
		ScreenToClient(hWnd, &pt);
		if ( ChildWindowFromPoint(hWnd, pt) != NULL )
		{
			mouseInside2 = true;
		}
	}
}

void Label::onTimer(int timerID)
{
	switch(timerID)
	{
		case TIMER_MOUSETRACK:
		{
			mouseTrack();
		}
		break;

		case TIMER_UPDATE:
		{
			if ( alwaysUpdateContent )
				update();
			else if ( visible && IsWindowVisible(hWnd) )
				update();
		}
		break;
		
		case TIMER_SCROLL:
		{
			if ( skipScrollRepaint )
				break;

			KillTimer(hWnd, TIMER_SCROLL);

			int useSteps;
			if ( scrollStepsInitial > 0 )
				useSteps = scrollStepsInitial;
			else
				useSteps = scrollSteps;

			if ( visible && IsWindowVisible(hWnd) && !frozen && (scroll == 2 || scroll == 4) && scrollPause > 0 && useSteps > 0 )
			{
				if ( scrollActualStep < useSteps )
				{
					scrollActualStep++;
				
					if (scroll == 2)
						scrollPosition -= scrollSpeed;
					else if (scroll == 4)
						scrollPosition += scrollSpeed;

					repaint();
				}
				else
				{
					scrollActualStep = 0;
					scrollStepsInitial = -1;
					frozen = true;
					SetTimer(hWnd, TIMER_SCROLLPAUSE, scrollPause, 0);
				}
			}
			//Normal Scroll (Horizontal or without Pause)
			else if ( visible && IsWindowVisible(hWnd) && !frozen )
			{
				if (scroll == 1 || scroll == 2)
					scrollPosition -= scrollSpeed;
				else if (scroll == 3 || scroll == 4)
					scrollPosition += scrollSpeed;

				repaint();
			}
		}
		break;

		case TIMER_AUTOHIDE:
		{
			autoHideIsRunning = false;
			KillTimer(hWnd, TIMER_AUTOHIDE);

			if ( visible )
				hide();
			else
				show();
		}
		break;

		case TIMER_ALPHAFADEIN:
		{
			if ( UpdateAlpha == 0 )
				ShowWindow(hWnd, SW_SHOWNOACTIVATE);

			if (UpdateAlpha < alphaFadeTransparencyBackup-alphaFadeStep)
			{
				UpdateAlpha += alphaFadeStep;
				UpdateLabelAlpha(UpdateAlpha);

				alphaTransparency = UpdateAlpha;
			}
			else
			{
				KillTimer(hWnd, TIMER_ALPHAFADEIN);
				if ( UpdateAlpha != alphaFadeTransparencyBackup)
				{
					UpdateAlpha = alphaFadeTransparencyBackup;
					UpdateLabelAlpha(alphaFadeTransparencyBackup);
				}

				alphaTransparency = alphaFadeTransparencyBackup;
			}
		}
		break;

		case TIMER_ALPHAFADEOUT:
		{
			if (UpdateAlpha > alphaFadeStep)
			{
				UpdateAlpha -= alphaFadeStep;
				UpdateLabelAlpha(UpdateAlpha);

				alphaTransparency = UpdateAlpha;
			}
			else
			{
				KillTimer(hWnd, TIMER_ALPHAFADEOUT);
				if (UpdateAlpha != 0)
				{
					UpdateAlpha = 0;
					UpdateLabelAlpha(0);
				}

				alphaTransparency = 0;
			}	
			
			if ( UpdateAlpha == 0 )
				ShowWindow(hWnd, SW_HIDE);
		}
		break;

		case TIMER_ALPHAFADEUPBANG:
		{
			if (UpdateAlpha < alphaFadeBangEndValue-alphaFadeBangStep)
			{
				UpdateAlpha += alphaFadeBangStep;
				UpdateLabelAlpha(UpdateAlpha);

				alphaTransparency = UpdateAlpha;
			}
			else
			{
				KillTimer(hWnd, TIMER_ALPHAFADEUPBANG);
				if (UpdateAlpha != alphaFadeBangEndValue)
				{
					UpdateAlpha = alphaFadeBangEndValue;
					UpdateLabelAlpha(alphaFadeBangEndValue);
				}

				alphaTransparency = alphaFadeBangEndValue;
			}
		}
		break;

		case TIMER_ALPHAFADEDOWNBANG:
		{
			if (UpdateAlpha > alphaFadeBangEndValue)
			{
				UpdateAlpha -= alphaFadeBangStep;
				UpdateLabelAlpha(UpdateAlpha);

				alphaTransparency = UpdateAlpha;
			}
			else
			{
				KillTimer(hWnd, TIMER_ALPHAFADEDOWNBANG);
				if (UpdateAlpha != alphaFadeBangEndValue)
				{
					UpdateAlpha = alphaFadeBangEndValue;
					UpdateLabelAlpha(alphaFadeBangEndValue);
				}

				alphaTransparency = alphaFadeBangEndValue;
			}
		}
		break;

		case TIMER_FRAMEANIMATION:
		{
			if ( visible && IsWindowVisible(hWnd) )
			{
				if (hoverActive && useHover)
				{
					if (frameCounter == hoverimageFrameCount)
					{
						frameCounter = 0;
						loopCounter++;
					}
					if (loopCounter == hoverimageLoop)
					{
						frameCounter = -1;
						loopCounter = 0;
						KillTimer(hWnd, TIMER_FRAMEANIMATION);
					}
				}
				else
				{
					if (frameCounter == imageFrameCount)
					{
						frameCounter = 0;
						loopCounter++;
					}
					if (loopCounter == imageLoop)
					{
						frameCounter = -1;
						loopCounter = 0;
						KillTimer(hWnd, TIMER_FRAMEANIMATION);
					}
				}
				frameCounter++;

				repaint(true);
			}
		}
		break;

		case TIMER_SCROLLPAUSE:
		{
			if ( visible && IsWindowVisible(hWnd) && frozen )
			{
				KillTimer(hWnd, TIMER_SCROLLPAUSE);
				frozen = false;
				repaint();
			}
		}
		break;
	}
}

bool Label::onWindowMessage(UINT message, WPARAM wParam, LPARAM lParam, LRESULT &lResult)
{
	switch(message)
	{
		case LM_TEXTEDITBOX:
		{
			enterActionProcessing( (string) (const char *) lParam );
			return true;
		}

		case LM_SETLABELTEXT:
		{
			setText(string((const char *) lParam), true);
			return true;
		}

		case LM_SETCURSOR:
		{
			if (lParam != NULL)
			{
				SetCursor( LoadCursor(0, (const char *) lParam ) );
			}
			else if (!cursor && !cursorPath.empty())
			{
				if (strstr(cursorPath.c_str(), ":\\") != NULL)
				{
					cursor = LoadCursorFromFile( cursorPath.c_str() );
				}

				if (!cursor)
				{	
					char imagePath[MAX_LINE_LENGTH] = {0};
					VarExpansion( imagePath, "$lsimagefolder$" );
					strcat(imagePath, cursorPath.c_str());

					cursor = LoadCursorFromFile( imagePath );
				}

				if (!cursor)
				{
					cursor = LoadCursor( 0, IDC_ARROW );
				}

				SetCursor( cursor );
			}
			else if ( cursor )
			{
				SetCursor( cursor );
			}
			else
			{
				SetCursor( LoadCursor( 0, IDC_ARROW ) );
			}

			return true;
		}

		case WM_KEYDOWN:
		{
			if ( !keyDownCommand.empty() )
			{
				CString fullCommand = keyDownCommand.c_str();
				if ( fullCommand.Find("%[keycode]%") != -1 )
				{
					fullCommand.Replace("%[keycode]%", IntToString((INT) wParam).c_str());	
				}
				xLSExecute(hWnd, fullCommand, SW_SHOWNORMAL);
			}

			return true;
		}

		case WM_KEYUP:
		{
			if ( !keyUpCommand.empty() )
			{
				CString fullCommand = keyUpCommand.c_str();
				if ( fullCommand.Find("%[keycode]%") != -1 )
				{
					fullCommand.Replace("%[keycode]%", IntToString((INT) wParam).c_str());
				}
				xLSExecute(hWnd, fullCommand, SW_SHOWNORMAL);
			}

			return true;
		}

		case WM_ACTIVATE:
		{
			if (wParam == WA_INACTIVE && focusActive)
			{
				if ( !loosefocusCommand.empty() )
					xLSExecute(hWnd, loosefocusCommand.c_str(), SW_SHOWNORMAL);

				focusActive = false;
			}
			else if ( (wParam == WA_CLICKACTIVE || wParam == WA_ACTIVE) && !focusActive )
			{
				if ( !gainfocusCommand.empty() )
					xLSExecute(hWnd, gainfocusCommand.c_str(), SW_SHOWNORMAL);

				focusActive = true;
			}

			return true;
		}


		case WM_MOUSEHOVER:
		{
			if ( !hoverCommand.empty() )
				xLSExecute(hWnd, hoverCommand.c_str(), SW_SHOWNORMAL);

			return true;
		}

		case WM_CLOSE:
		{
			lResult = 0;
			return true;
		}

		case WM_DESTROY:
		{
			if (box != NULL)
			{
				if ( TooltipHints )
				{
					removeHint(hWnd);
					DestroyWindow(TooltipHints);
					TooltipHints = 0;
				}

				if (hWndTextBox)
				{
					if ( textcursor )
					{
						DestroyCursor(textcursor);
						textcursor = 0;
					}
					// Remove the subclass from the edit control. 
					SetWindowLong(this->hWndTextBox, GWL_WNDPROC, (LONG) this->wpOrigEditProc); 
					
					if ( hTextBGBrush )
					{
						DeleteObject(hTextBGBrush);
						hTextBGBrush = 0;
					}

					if ( textboxhFont )
					{
						DeleteObject( (HGDIOBJ)(HFONT) textboxhFont );
						textboxhFont = 0;
					}
					
					DestroyWindow(hWndTextBox);
					hWndTextBox = 0;
				}

				KillTimer(hWnd, TIMER_MOUSETRACK);
				KillTimer(hWnd, TIMER_UPDATE);
				KillTimer(hWnd, TIMER_SCROLL);
				KillTimer(hWnd, TIMER_AUTOHIDE);
				KillTimer(hWnd, TIMER_ALPHAFADEIN);
				KillTimer(hWnd, TIMER_ALPHAFADEOUT);
				KillTimer(hWnd, TIMER_FRAMEANIMATION);
				KillTimer(hWnd, TIMER_SCROLLPAUSE);
				KillTimer(hWnd, TIMER_ALPHAFADEUPBANG);
				KillTimer(hWnd, TIMER_ALPHAFADEDOWNBANG);
			
				box = NULL;
				hide();
				
				ParentDestroyNotificationReceived = true;
			}			
			
			return false;
		}

		case WM_DROPFILES:
		{
			if (dropCommand.empty())	// we shouldn't get here... but somehow we did :p
				break;

			int intFileAction = -1;

			char action[32] = {0};
			char szTarget[MAX_LINE_LENGTH] = {0};
			char *buffers[] = {action};

			LCTokenize(dropCommand.c_str(), buffers, 1, szTarget);

			if (_stricmp(action, "move") == NULL)
				intFileAction = 1;
			else if (_stricmp(action, "copy") == NULL)
				intFileAction = 2;
			else if (_stricmp(action, "delete") == NULL)
				intFileAction = 0;
			else if (_stricmp(action, "shortcut") == NULL)
				intFileAction = 3;
			else
				intFileAction = -1;

			if (intFileAction == -1)
			{
				int numDropped, i;
				char szFile[MAX_LINE_LENGTH];
				
				numDropped = DragQueryFile((HDROP) wParam, 0xFFFFFFFF, NULL, 0);
				for (i = 0; i < numDropped; i++)
				{
					DragQueryFile((HDROP) wParam, i, (char *)&szFile, MAX_LINE_LENGTH);
					if (szFile && szFile[0])
					{
						CString fullCommand = dropCommand.c_str();
						if ( fullCommand.Find("%[droppath]%") != -1 )
						{
							fullCommand.Replace("%[droppath]%", szFile);
						}
						else
						{
							fullCommand += " \"";
							fullCommand += szFile;
							fullCommand += "\"";
						}
						xLSExecute(hWnd, fullCommand, SW_SHOWNORMAL);
					}
				}

				DragFinish((HDROP) wParam);
				break;
			}
		
			//Test Path for Existence
			CString strPath = szTarget;
			strPath.Replace("\"", "");
			if ( strPath.Right(1) != "\\" )
				strPath += "\\";

			bool pathExists = PathFileExists(strPath);

			if (intFileAction >= 0 && intFileAction < 3)
			{
				char *szSourceFiles = NULL;
				char buffer[MAX_LINE_LENGTH] = {0};
				int numFiles;
				UINT len,scursor = 0;

				numFiles = DragQueryFile((HDROP) wParam, 0xFFFFFFFF, NULL, NULL);

				szSourceFiles = new char[MAX_LINE_LENGTH*numFiles+1];
				scursor = 0;

				for (int i=0;i<numFiles;i++)
				{
					len = DragQueryFile((HDROP) wParam, i, buffer, MAX_LINE_LENGTH);
					
					strcpy(szSourceFiles+scursor,buffer);
					scursor += len + 1;
				}
					
				DragFinish((HDROP) wParam);

				if( pathExists || intFileAction == 0 )
				{
					szSourceFiles[scursor] = 0;
					strcpy(buffer, strPath);
					buffer[strlen(buffer)+1] = 0;
					
					SHFILEOPSTRUCT shfop;
					shfop.hwnd = NULL;
					if (intFileAction == 1)
						shfop.wFunc = FO_MOVE;
					else if (intFileAction == 2)
						shfop.wFunc = FO_COPY;
					else if (intFileAction == 0)
						shfop.wFunc = FO_DELETE;
					shfop.pFrom = szSourceFiles;
					shfop.pTo = buffer;
					shfop.fFlags = FOF_NOCONFIRMMKDIR | FOF_ALLOWUNDO;
					SHFileOperation(&shfop);
				}
				
				delete[] szSourceFiles;
			}
			else if (pathExists && intFileAction == 3)
			{
				int numDropped;
				char szFile[MAX_LINE_LENGTH];
				
				numDropped = DragQueryFile((HDROP) wParam, 0xFFFFFFFF, NULL, 0);
				for (int i = 0; i < numDropped; i++)
				{
					DragQueryFile((HDROP) wParam, i, (char *)&szFile, MAX_LINE_LENGTH);

					HRESULT hres; 
					IShellLink* psl; 

					hres = CoCreateInstance(CLSID_ShellLink, NULL,
					CLSCTX_INPROC_SERVER, IID_IShellLink, (LPVOID *) &psl); 
					if (SUCCEEDED(hres))
					{
						IPersistFile* ppf; 
						psl->SetPath(szFile); 
						psl->SetDescription("");
 
						hres = psl->QueryInterface(IID_IPersistFile, (LPVOID*)&ppf); 
						if (SUCCEEDED(hres))
						{
							CString work = szFile;
							work = work.Right(work.GetLength()-work.ReverseFind('\\')-1);
							work = work.Left(work.Find("."));
							work += ".lnk";

							WCHAR wsz[MAX_PATH]; 
							MultiByteToWideChar(CP_ACP, 0, (strPath + work), -1, wsz, MAX_LINE_LENGTH);  
							hres = ppf->Save(wsz, TRUE); 
							ppf->Release(); 
						}
						psl->Release(); 
					}
				}
				DragFinish((HDROP) wParam);
			}
			
			return true;
		}
		
		case WM_LBUTTONDBLCLK:
		{
			if(labelInitialized && leftDoubleClickCommand.empty() && labelLeftDoubleClickRegions.empty() && leftModDoubleClickList.empty())
			{
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onLButtonDblClk(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			return true;
		}

		case WM_LBUTTONDOWN:
		{	
			if(labelInitialized && leftClickDownCommand.empty() && leftClickUpCommand.empty() && leftClickCommand.empty() && labelLeftClickRegions.empty() && labelHREFClickRegions.empty() && leftModClickDownList.empty() && (!moveable || (moveable && moveButton != 1)) )
			{
				if ( usePressed )
				{
					pressedActive = true;
					repaint(true);
				}

				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onLButtonDown(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			return true;
		}

		case WM_LBUTTONUP:
		{
			if(labelInitialized && leftClickDownCommand.empty() && leftClickUpCommand.empty() && leftClickCommand.empty() && labelLeftClickRegions.empty() && labelHREFClickRegions.empty() && leftModClickUpList.empty() && leftModClickList.empty() && (!moveable || (moveable && moveButton != 1)) )
			{
				if ( usePressed )
				{
					pressedActive = false;
					repaint(true);
				}
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onLButtonUp(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			return true;
		}

		case WM_MBUTTONDBLCLK:
		{
			if(labelInitialized && middleDoubleClickCommand.empty() && labelMiddleDoubleClickRegions.empty() && middleModDoubleClickList.empty())
			{
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onMButtonDblClk(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			return true;
		}

		case WM_MBUTTONDOWN:
		{
			if ( skipButton )
				return true;

			if(labelInitialized && middleClickDownCommand.empty() && middleClickUpCommand.empty() && middleClickCommand.empty() && labelMiddleClickRegions.empty() && middleModClickDownList.empty() && (!moveable || (moveable && moveButton != 3)) )
			{
				if ( usePressed )
				{
					pressedActive = true;
					repaint(true);
				}
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onMButtonDown(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			return true;
		}

		case WM_MBUTTONUP:
		{
			if ( skipButton )
			{
				skipButton = false;
				return true;
			}

			if(labelInitialized && middleClickDownCommand.empty() && middleClickUpCommand.empty() && middleClickCommand.empty() && labelMiddleClickRegions.empty() && middleModClickUpList.empty() && middleModClickList.empty() && (!moveable || (moveable && moveButton != 3)) )
			{
				if ( usePressed )
				{
					pressedActive = false;
					repaint(true);
				}
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onMButtonUp(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			return true;
		}

		case WM_RBUTTONDBLCLK:
		{
			if(labelInitialized && rightDoubleClickCommand.empty() && labelRightDoubleClickRegions.empty() && rightModDoubleClickList.empty())
			{
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onRButtonDblClk(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			return true;
		}

		case WM_RBUTTONDOWN:
		{
			if(labelInitialized && rightClickDownCommand.empty() && rightClickUpCommand.empty() && rightClickCommand.empty() && labelRightClickRegions.empty() && rightModClickDownList.empty() && (!moveable || (moveable && moveButton != 2)) )
			{
				if ( usePressed )
				{
					pressedActive = true;
					repaint(true);
				}
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onRButtonDown(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			return true;
		}

		case WM_RBUTTONUP:
		{
			if(labelInitialized && rightClickDownCommand.empty() && rightClickUpCommand.empty() && rightClickCommand.empty() && labelRightClickRegions.empty() && rightModClickUpList.empty() && rightModClickList.empty() && (!moveable || (moveable && moveButton != 2)) )
			{
				if ( usePressed )
				{
					pressedActive = false;
					repaint(true);
				}
				relayMouseMessageToBox(message, wParam, lParam);
				return true;
			}

			onRButtonUp(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			return true;
		}

		case WM_XBUTTONDOWN:
		{
			if (HIWORD(wParam) == XBUTTON1)
			{
				if(labelInitialized && X1ClickDownCommand.empty() && X1ClickUpCommand.empty() && X1ClickCommand.empty() && labelX1ClickRegions.empty() && X1ModClickDownList.empty() )
				{
					if ( usePressed )
					{
						pressedActive = true;
						repaint(true);
					}
					relayMouseMessageToBox(message, wParam, lParam);
					return true;
				}

				onX1ButtonDown(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			}
			else if (HIWORD(wParam) == XBUTTON2)
			{
				if(labelInitialized && X2ClickDownCommand.empty() && X2ClickUpCommand.empty() && X2ClickCommand.empty() && labelX2ClickRegions.empty() && X2ModClickDownList.empty() )
				{
					if ( usePressed )
					{
						pressedActive = true;
						repaint(true);
					}
					relayMouseMessageToBox(message, wParam, lParam);
					return true;
				}

				onX2ButtonDown(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			}
						
			return true;
		}

		case WM_XBUTTONUP:
		{
			if (HIWORD(wParam) == XBUTTON1)
			{
				if(labelInitialized && X1ClickDownCommand.empty() && X1ClickUpCommand.empty() && X1ClickCommand.empty() && labelX1ClickRegions.empty() && X1ModClickUpList.empty() && X1ModClickList.empty() )
				{
					if ( usePressed )
					{
						pressedActive = false;
						repaint(true);
					}
					relayMouseMessageToBox(message, wParam, lParam);
					return true;
				}

				onX1ButtonUp(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			}
			else if (HIWORD(wParam) == XBUTTON2)
			{
				if(labelInitialized && X2ClickDownCommand.empty() && X2ClickUpCommand.empty() && X2ClickCommand.empty() && labelX2ClickRegions.empty() && X2ModClickUpList.empty() && X2ModClickList.empty() )
				{
					if ( usePressed )
					{
						pressedActive = false;
						repaint(true);
					}
					relayMouseMessageToBox(message, wParam, lParam);
					return true;
				}

				onX2ButtonUp(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			}
						
			return true;
		}

		case WM_XBUTTONDBLCLK:
		{
			if (HIWORD(wParam) == XBUTTON1)
			{
				if(labelInitialized && X1DoubleClickCommand.empty() && labelX1DoubleClickRegions.empty() && X1ModDoubleClickList.empty())
				{
					relayMouseMessageToBox(message, wParam, lParam);
					return true;
				}

				onX1ButtonDblClk(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			}
			else if (HIWORD(wParam) == XBUTTON2)
			{
				if(labelInitialized && X2DoubleClickCommand.empty() && labelX2DoubleClickRegions.empty() && X2ModDoubleClickList.empty())
				{
					relayMouseMessageToBox(message, wParam, lParam);
					return true;
				}

				onX2ButtonDblClk(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			}

			return true;
		}

		case WM_MOUSEWHEEL:
		{
			POINT pt;
			RECT childPosition;

			GetCursorPos(&pt);

			char tmpClass[MAX_LINE_LENGTH] = {0};				
			
			if ( hasHookedModules )
			{
				for(HWND hWndChild = ::GetTopWindow(hWnd); hWndChild != NULL; hWndChild = ::GetNextWindow(hWndChild, GW_HWNDNEXT))
				{
					GetWindowRect(hWndChild, &childPosition);
					if ( PtInRect(&childPosition, pt) )
					{
						GetClassName(hWndChild, tmpClass, MAX_LINE_LENGTH);
						if (_stricmp(tmpClass, "xlabel") == NULL	|| _stricmp(tmpClass, "xlabelxp") == NULL || _stricmp(tmpClass, "xtray") == NULL || _stricmp(tmpClass, "xtaskbar") == NULL)
						{
							ScreenToClient(hWndChild, &pt);
							::PostMessage(hWndChild, WM_MOUSEWHEEL, wParam, MAKELPARAM((short) pt.x, (short) pt.y));
								
							return true;
						}

						return true;
					}
				}
			}

			ScreenToClient(hWnd, &pt);
			if ((short)(HIWORD(wParam)) < 0)
			{
				onWheelDown(pt.x, pt.y);

				return true;
			}
			else
			{
				onWheelUp(pt.x, pt.y);

				return true;
			}
		}

		case WM_MOUSEMOVE:
		{
			if ( TooltipHints )
			{
				MSG hintMessage = {hWnd, message, wParam, lParam};
				SendMessage( TooltipHints, (UINT) TTM_RELAYEVENT, 0, (LPARAM) (LPMSG) &hintMessage ); 
			}
			
			if (moveable && moveKey != -1)
			{
				if (moveKey == 3)
					movemodifierkeyPressed = (bool) (0x8000 & GetKeyState(VK_CONTROL)) && (0x8000 & GetKeyState(VK_SHIFT));
				else if (moveKey == 2)
					movemodifierkeyPressed = (bool) (0x8000 & GetKeyState(VK_SHIFT));
				else if (moveKey == 1)
					movemodifierkeyPressed = (bool) (0x8000 & GetKeyState(VK_CONTROL));
				else 
					movemodifierkeyPressed = (bool) (0x8000 & GetKeyState(VK_CONTROL));
			}
			else if (moveable && moveKey == -1)
			{
				movemodifierkeyPressed = true;
			}

			onMouseMove(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			return true;
		}

		case WM_CTLCOLOREDIT:
		case WM_CTLCOLORSTATIC:
		{
			const HDC hDC = (HDC)wParam;
			SetTextColor(hDC, font->getTextColor());

			if ( textboxPaintingMode == -1 )
			{
				SetBkColor(hDC, textboxPlainColor);
				SetBkMode(hDC, OPAQUE);
			}
			else
			{
				SetBkMode(hDC, TRANSPARENT);
			}

			lResult = (int) GetBGBrush();

			return true;
		}

		case WM_PAINT:
		{	
			PAINTSTRUCT ps;
			HDC hDC;

			if(!wParam)
				hDC = BeginPaint(hWnd, &ps);
			else
				hDC = (HDC) wParam;

			onPaint(hDC);

			if(!wParam)
				EndPaint(hWnd, &ps);

			if ( restartScrollRepaint )
			{
				restartScrollRepaint = false;
				SetTimer(hWnd, TIMER_SCROLL, scrollInterval, 0);
			}
					
			return true;	
		}

		case WM_GETMINMAXINFO:
		{
			POINT maxsize = {autoMaxWidth, autoMaxHeight};
			POINT minsize = {autoMinWidth, autoMinHeight};

			minsize.x = max(minsize.x, resizeBorderLeft+resizeBorderRight);
			minsize.y = max(minsize.y, resizeBorderTop+resizeBorderBottom);

			((MINMAXINFO FAR*) lParam)->ptMaxTrackSize = maxsize;
			((MINMAXINFO FAR*) lParam)->ptMinTrackSize = minsize;

			return true;
		}

		case WM_WINDOWPOSCHANGING:
		{
			onWindowPosChanging((WINDOWPOS *) lParam);
			return true;
		}

		case WM_MOVE:
		{
			onMove(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			return true;
		}

		case WM_SIZE:
		{
			onSize(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
			return true;
		}

		case WM_TIMER:
		{
			onTimer((int) wParam);
			return true;
		}

		case WM_NCHITTEST:
		{
			if ( useresizeBorder )
			{
				POINT point = {GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam)};
				ScreenToClient(hWnd, &point);

				lResult = NcHitTest(point.x, point.y);
				
				if ( lResult == 0 )
					return false;
				else
					return true;
			}
			else
			{
				return false;
			}
		}

		default:
		{
			return false;
		}
	}

	return false;
}

HBRUSH Label::GetBGBrush()
{
	if ( !hTextBGBrush )
	{
		HDC tmpDC = CreateCompatibleDC(NULL);
		HBITMAP tmpBmp = CreateBitmap((TextEditBoxSize.right-TextEditBoxSize.left), (TextEditBoxSize.bottom-TextEditBoxSize.top), 1, bitsperpixel, NULL);
		HBITMAP oldBMP = (HBITMAP) SelectObject(tmpDC, tmpBmp);

		//Draw Image
		onPaint(tmpDC, &TextEditBoxSize);

		hTextBGBrush = CreatePatternBrush(tmpBmp);

		tmpBmp = (HBITMAP) SelectObject(tmpDC, oldBMP);
		DeleteObject(tmpBmp);
		DeleteDC(tmpDC);
	}
  
	return hTextBGBrush;
}

LRESULT Label::windowProcedure(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	Label *label = NULL;

	if (message == WM_NCCREATE)
	{
		label = (Label *) ((CREATESTRUCT *) lParam)->lpCreateParams;
		label->hWnd = hWnd;
		SetWindowLong(hWnd, 0, (LONG) label);
	}
	else
	{
		label = (Label *) GetWindowLong(hWnd, 0);
	}

	if (label)
	{
		LRESULT lResult = 0;

		if ( label->onWindowMessage(message, wParam, lParam, lResult) )
			return lResult;
	}

	return DefWindowProc(hWnd, message, wParam, lParam);
}

//------------------------------------------------------------
//Tooltips (Add, Update, Remove)
//------------------------------------------------------------

void Label::addHint(HWND hWnd, LPSTR caption)
{
	if ( !TooltipHints )
		return;
	
	RECT clientRect;
	GetClientRect(hWnd, &clientRect);

	TOOLINFO ti;
	ti.cbSize = sizeof(TOOLINFO);
	ti.uFlags = TTF_TRANSPARENT;
	ti.hwnd = hWnd;
	ti.uId = 0;
	ti.rect = clientRect;
	ti.hinst = hInstance;
	ti.lpszText = caption;
	ti.lParam = 0;

	SendMessage(TooltipHints, TTM_ADDTOOL, 0, (LPARAM) (LPTOOLINFO) &ti);
}

void Label::updateHint(HWND hWnd, LPSTR caption, bool updatesize)
{
	if ( !TooltipHints )
		return;	
	
	RECT clientRect;
	GetClientRect(hWnd, &clientRect);

	TOOLINFO ti;
	ti.cbSize = sizeof(TOOLINFO);
	ti.uFlags = TTF_TRANSPARENT;
	ti.hwnd = hWnd;
	ti.uId = 0;
	ti.rect = clientRect;
	ti.hinst = hInstance;
	ti.lpszText = caption;
	ti.lParam = 0;

	if (updatesize)
	{
		SendMessage(TooltipHints, TTM_NEWTOOLRECT, 0, (LPARAM) (LPTOOLINFO) &ti);
	}
	else
	{
		SendMessage(TooltipHints, TTM_UPDATETIPTEXT, 0, (LPARAM) (LPTOOLINFO) &ti);
	}
}

void Label::removeHint(HWND hWnd)
{
	if ( !TooltipHints )
		return;	
	
	TOOLINFO ti;
	RECT emptyRect = {0, 0, 0, 0};

	ti.cbSize = sizeof(TOOLINFO);
	ti.uFlags = 0;
	ti.hwnd = hWnd;
	ti.uId = 0;
	ti.rect = emptyRect;
	ti.hinst = hInstance;
	ti.lpszText = NULL;
	ti.lParam = 0;

	SendMessage(TooltipHints, TTM_DELTOOL, 0, (LPARAM) (LPTOOLINFO) &ti);
}

int Label::isCloseTo(POINT &pos, const POINT size, const int x, const int y, const int cx, const int cy, const int span, bool skipX, bool skipY)
{
	bool internalX = false;
	bool internalY = false;

	int tmpdockx = -1;
	int tmpdocky = -1;

	POINT tmppos = {pos.x, pos.y};

	if ( !skipX && (tmppos.x > x-span) && (tmppos.x < x+span) )
	{
		if ( tmppos.y > y-span && tmppos.y < y+span )
		{
			pos.x = x;
			internalX = true;
			tmpdockx = 0;
		}
		else if ( ( tmppos.y > y-span && tmppos.y < y+span ) 
			|| ( tmppos.y+size.y > y-span && tmppos.y+size.y < y+span ) 
			|| ( tmppos.y > y+cy-span && tmppos.y < y+cy+span ) 
			|| ( tmppos.y+size.y > y+cy-span && tmppos.y+size.y < y+cy+span ) )
		{
			pos.x = x;
			internalX = true;
			tmpdockx = 0;
		}
		else if ( tmppos.y > y && tmppos.y < y+cy 
				&& tmppos.y+size.y > y && tmppos.y+size.y < y+cy )
		{
			pos.x = x;
			internalX = true;
			tmpdockx = 0;
		}
	}
	else if ( !skipX && (tmppos.x > x+cx-span) && (tmppos.x < x+cx+span) )
	{
		if ( ( tmppos.y > y-span && tmppos.y < y+span ) 
			|| ( tmppos.y+size.y > y-span && tmppos.y+size.y < y+span ) 
			|| ( tmppos.y > y+cy-span && tmppos.y < y+cy+span ) 
			|| ( tmppos.y+size.y > y+cy-span && tmppos.y+size.y < y+cy+span ) )
		{
			pos.x = x+cx;
			internalX = true;
			tmpdockx = 2;
		}
		else if ( tmppos.y > y && tmppos.y < y+cy 
				&& tmppos.y+size.y > y && tmppos.y+size.y < y+cy )
		{
			pos.x = x+cx;
			internalX = true;
			tmpdockx = 2;
		}
	}
	else if ( !skipX && (tmppos.x+size.x > x-span ) && (tmppos.x+size.x < x+span) )
	{
		if ( ( tmppos.y > y-span && tmppos.y < y+span ) 
			|| ( tmppos.y+size.y > y-span && tmppos.y+size.y < y+span ) 
			|| ( tmppos.y > y+cy-span && tmppos.y < y+cy+span ) 
			|| ( tmppos.y+size.y > y+cy-span && tmppos.y+size.y < y+cy+span ) )
		{
			pos.x = x-size.x;
			internalX = true;
			tmpdockx = 0;
		}
		else if ( tmppos.y > y && tmppos.y < y+cy 
				&& tmppos.y+size.y > y && tmppos.y+size.y < y+cy )
		{
			pos.x = x-size.x;
			internalX = true;
			tmpdockx = 0;
		}
	}
	else if ( !skipX && (tmppos.x+size.x > x+cx-span ) && (tmppos.x+size.x < x+cx+span) )
	{
		if ( ( tmppos.y > y-span && tmppos.y < y+span ) 
			|| ( tmppos.y+size.y > y-span && tmppos.y+size.y < y+span ) 
			|| ( tmppos.y > y+cy-span && tmppos.y < y+cy+span ) 
			|| ( tmppos.y+size.y > y+cy-span && tmppos.y+size.y < y+cy+span ) )
		{
			pos.x = x+cx-size.x;
			internalX = true;
			tmpdockx = 2;
		}
		else if ( tmppos.y > y && tmppos.y < y+cy 
				&& tmppos.y+size.y > y && tmppos.y+size.y < y+cy )
		{
			pos.x = x+cx-size.x;
			internalX = true;
			tmpdockx = 2;
		}
	}

	if ( !skipY && (tmppos.y > y-span) && (tmppos.y < y+span) )
	{
		if ( tmppos.x > x-span && tmppos.x < x+span )
		{
			pos.y = y;
			internalY = true;
			tmpdocky = 1;
		}
		else if ( ( tmppos.x > x-span && tmppos.x < x+span ) 
			|| ( tmppos.x+size.x > x-span && tmppos.x+size.x < x+span ) 
			|| ( tmppos.x > x+cx-span && tmppos.x < x+cx+span ) 
			|| ( tmppos.x+size.x > x+cx-span && tmppos.x+size.x < x+cx+span ) )
		{
			pos.y = y;
			internalY = true;
			tmpdocky = 1;
		}
		else if ( tmppos.x > x && tmppos.x < x+cx 
				&& tmppos.x+size.x > x && tmppos.x+size.x < x+cx )
		{
			pos.y = y;
			internalY = true;
			tmpdocky = 1;
		}
	}
	else if ( !skipY && tmppos.y > y+cy-span && tmppos.y < y+cy+span )
	{
		if ( ( tmppos.x > x-span && tmppos.x < x+span ) 
			|| ( tmppos.x+size.x > x-span && tmppos.x+size.x < x+span ) 
			|| ( tmppos.x > x+cx-span && tmppos.x < x+cx+span ) 
			|| ( tmppos.x+size.x > x+cx-span && tmppos.x+size.x < x+cx+span ) )
		{
			pos.y = y+cy;
			internalY = true;
			tmpdocky = 3;
		}
		else if ( tmppos.x > x && tmppos.x < x+cx  
				&& tmppos.x+size.x > x && tmppos.x+size.x < x+cx )
		{
			pos.y = y+cy;
			internalY = true;
			tmpdocky = 3;
		}
	}
	else if ( !skipY && tmppos.y+size.y > y-span && tmppos.y+size.y < y+span )
	{
		if ( ( tmppos.x > x-span && tmppos.x < x+span ) 
			|| ( tmppos.x+size.x > x-span && tmppos.x+size.x < x+span ) 
			|| ( tmppos.x > x+cx-span && tmppos.x < x+cx+span ) 
			|| ( tmppos.x+size.x > x+cx-span && tmppos.x+size.x < x+cx+span ) )
		{
			pos.y = y-size.y;
			internalY = true;
			tmpdocky = 1;
		}
		else if ( tmppos.x > x && tmppos.x < x+cx 
				&& tmppos.x+size.x > x && tmppos.x+size.x < x+cx )
		{
			pos.y = y-size.y;
			internalY = true;
			tmpdocky = 1;
		}
	}
	else if ( !skipY && tmppos.y+size.y > y+cy-span && tmppos.y+size.y < y+cy+span )
	{
		if ( ( tmppos.x > x-span && tmppos.x < x+span ) 
			|| ( tmppos.x+size.x > x-span && tmppos.x+size.x < x+span ) 
			|| ( tmppos.x > x+cx-span && tmppos.x < x+cx+span ) 
			|| ( tmppos.x+size.x > x+cx-span && tmppos.x+size.x < x+cx+span ) )
		{
			pos.y = y+cy-size.y;
			internalY = true;
			tmpdocky = 3;
		}
		else if ( tmppos.x > x && tmppos.x < x+cx 
				&& tmppos.x+size.x > x && tmppos.x+size.x < x+cx )
		{
			pos.y = y+cy-size.y;
			internalY = true;
			tmpdocky = 3;
		}
	}

	if (!internalX && !internalY)
	{
		return 0;
	}
	else if (internalX && !internalY)
	{
		dockedtoSide = tmpdockx;
		return 1;
	}
	else if (!internalX && internalY)
	{
		dockedtoSide = tmpdocky;
		return 2;
	}
	else
	{
		if ( tmpdockx == 0 && tmpdocky == 1 )
			dockedtoSide = 4;
		else if ( tmpdockx == 2 && tmpdocky == 1 )
			dockedtoSide = 5;
		else if ( tmpdockx == 0 && tmpdocky == 3 )
			dockedtoSide = 6;
		else if ( tmpdockx == 2 && tmpdocky == 3 )
			dockedtoSide = 7;
		return 3;
	}
}

void Label::ontemplateButtonDown(int x, int y, int actualMouseClick, Label_RegionList &labelClickRegions, string clickDownCommand)
{
	bool skipClick = false;
	//Stop Animation Timer (Normal & Hover)
	KillTimer(hWnd, TIMER_FRAMEANIMATION);

	if(moveable && moveButton == actualMouseClick)
	{
		if(!mousePressed && !movemodifierkeyPressed)
		{
			mousePressed = true;
		}
		else if (!mousePressed && movemodifierkeyPressed)
		{
			mousePressed = true;
			moveButtonPressed = actualMouseClick;
			GetCursorPos(&savedpt);
			ScreenToClient(hWnd, &savedpt);	
		}
	}
	else
	{
		if(!mousePressed)
			mousePressed = true;
	}

	bool repaintNeeded = false;

	pressedActive = true;

	if ( font != NULL && font->usePressedText )
	{
		//AutoSize due to possible Font change
		if ( !text.empty() && (autoWidthMode != 0 || autoHeightMode != 0) )
			autoSize();
	}

	if (usePressed)
		repaintNeeded = true;
		
	//Last TRUE specifies Check for Clicks in GhostLabel
	if ( !ghostlabelList.empty() )
	{
		if ( checkClickRegions(x, y, labelClickRegions, -1) )
		{
			repaintNeeded = true;
			skipClick = true;
		}
	}

	//Repaint due to possible FontChange
	if ( !originalText.empty() || !text.empty() )
		repaint(repaintNeeded);
	else if (usePressed)
		repaint(true);
	
	SetCapture(hWnd);

	if ( !skipClick )
	{
		if ( actualMouseClick == 1 )
			skipClick = CheckModClick(leftModClickDownList);
		else if ( actualMouseClick == 2 )
			skipClick = CheckModClick(rightModClickDownList);
		else if ( actualMouseClick == 3 )
			skipClick = CheckModClick(middleModClickDownList);
		else if ( actualMouseClick == 4 )
			skipClick = CheckModClick(X1ModClickDownList);
		else if ( actualMouseClick == 5 )
			skipClick = CheckModClick(X2ModClickDownList);
		
		if( !skipClick && !clickDownCommand.empty() )
			xLSExecute(hWnd, clickDownCommand.c_str(), SW_SHOWNORMAL);
	}
}

void Label::ontemplateButtonUp(int x, int y, int actualMouseClick, Label_RegionList &labelClickRegions, string clickCommand, string clickUpCommand)
{
	bool repaintNeeded = false;
	bool skipClick = false;

	if ( moveable && moveButton == actualMouseClick )
	{	
		//To be on the save side
		ReleaseCapture();

		if ( mousePressed )
		{
			SendMessage(hWnd, LM_SETCURSOR, 0, 0);
		
			if (actualMouseClick == 1)
				skipClick = checkHREFClickRegions(x, y, labelHREFClickRegions);

			if ( !skipClick )
				skipClick = checkClickRegions(x, y, labelClickRegions, 1);

			mousePressed = false;
			moveButtonPressed = 0;

			//Docking
			if ( hasMoved )
			{
				for(StringListIterator sit = snapToList.begin(); sit != snapToList.end(); sit++)
				{
					ParseBangCommand(hWnd, "!labeldock", (*sit).c_str());
				}
				snapToList.clear();
			}

			if(hasMoved && labelInitialized && !moveCommand.empty())
				xLSExecute(hWnd, moveCommand.c_str(), SW_SHOWNORMAL);
			
			hasMoved = false;
		}
	}
	else 
	{
		//To be on the save side
		ReleaseCapture();

		if( mousePressed )
		{	
			if ( actualMouseClick == 1)
				skipClick = checkHREFClickRegions(x, y, labelHREFClickRegions);
			
			if ( !skipClick )
				skipClick = checkClickRegions(x, y, labelClickRegions, 1);
			
			mousePressed = false;
			moveButtonPressed = 0;
		}
	}

	hoverActive = true;
	pressedActive = false;

	if ( font != NULL && font->usePressedText )
	{
		//AutoSize due to possible Font change
		if ( !text.empty() && (autoWidthMode != 0 || autoHeightMode != 0) )
			autoSize();
	}

	if (usePressed)
		repaintNeeded = true;
	
	if ( skipClick && !ghostlabelList.empty() )
	{
		repaintNeeded = true;
	}

	//Animation RESTART
	//Restart Hover Anim
	if (useHover && hoverAnim)
	{
		frameCounter = 1;
		loopCounter = 0;
		SetTimer(hWnd, TIMER_FRAMEANIMATION, hoverimageFrameDelay, NULL);
	}
	//Restart Normal Anim
	else if (normalAnim)
	{
		frameCounter = 1;
		loopCounter = 0;
		SetTimer(hWnd, TIMER_FRAMEANIMATION, imageFrameDelay, NULL);
	}

	//Repaint due to possible FontChange
	if ( !originalText.empty() || !text.empty() )
		repaint(repaintNeeded);
	else if (repaintNeeded)
		repaint(true);

	if ( !skipClick )
	{
		bool tmpskipClick = false;

		if ( actualMouseClick == 1 )
		{
			tmpskipClick = CheckModClick(leftModClickList);
			skipClick = CheckModClick(leftModClickUpList);
		}
		else if ( actualMouseClick == 2 )
		{
			tmpskipClick = CheckModClick(rightModClickList);
			skipClick = CheckModClick(rightModClickUpList);
		}
		else if ( actualMouseClick == 3 )
		{
			tmpskipClick = CheckModClick(middleModClickList);
			skipClick = CheckModClick(middleModClickUpList);
		}
		else if ( actualMouseClick == 4 )
		{
			tmpskipClick = CheckModClick(X1ModClickList);
			skipClick = CheckModClick(X1ModClickUpList);
		}
		else if ( actualMouseClick == 5 )
		{
			tmpskipClick = CheckModClick(X2ModClickList);
			skipClick = CheckModClick(X2ModClickUpList);
		}

		if ( !skipClick && tmpskipClick )
			skipClick = true;
	}

	if( !skipClick && !clickCommand.empty() )
	{
		if ( hWndTextBox && (_stricmp(clickCommand.c_str(), ".editbox") == NULL || _stricmp(clickCommand.c_str(), "\".editbox\"") == NULL) )
		{
			ShowWindow(hWndTextBox, SW_SHOW);
			textboxVisible = true;
			SetForegroundWindow(hWndTextBox);
			SetFocus(hWndTextBox);
		}
		else
		{
			xLSExecute(hWnd, clickCommand.c_str(), SW_SHOWNORMAL);
		}
	}

	if( !skipClick && !clickUpCommand.empty() )
		xLSExecute(hWnd, clickUpCommand.c_str(), SW_SHOWNORMAL);
}

void Label::enterActionProcessing(string input)
{
	HINSTANCE result = HINSTANCE(32);

	if ( !input.empty() )
	{
		if ( _stricmp(enterAction.c_str(), ".execute") == NULL || _stricmp(enterAction.c_str(), "\".execute\"") == NULL )
		{
			result = xLSExecute(hWnd, input.c_str(), SW_SHOWNORMAL);
		}
		else if ( _stricmp(enterAction.c_str(), ".settext") == NULL || _stricmp(enterAction.c_str(), "\".settext\"") == NULL )
		{
			SendMessage(hWnd, LM_SETLABELTEXT, 0, (LONG) input.c_str());
			result = HINSTANCE(33);
		}
		else
		{
			CString worker = enterAction.c_str();
			worker.Replace("%[textinput]%", input.c_str());
			result = xLSExecute(hWnd, worker, SW_SHOWNORMAL);
		}
	}

	if ( (int) result > 32 )
	{
		if ( !successCommand.empty() )
			xLSExecute(hWnd, successCommand.c_str(), SW_SHOWNORMAL);
	}
	else
	{
		if ( !failCommand.empty() )
		{	
			CString worker = failCommand.c_str();
			worker.Replace("%[textinput]%", input.c_str());
			xLSExecute(hWnd, worker, SW_SHOWNORMAL);
		}
	}
}


//------------------------------------------------------------
//Initialize and Update AlphaTransparency
//------------------------------------------------------------

void Label::UpdateLabelAlpha(int newAlphaTransparency)
{
	//Initialize LayeredWindow Function
	if (!alphaInitialized)
	{
		HMODULE hDLL = LoadLibrary ("user32");

		pUpdateLayeredWindow = 
					(PULW) GetProcAddress(hDLL, "UpdateLayeredWindow");

		pSetLayeredWindowAttributes = 
					(PSLWA) GetProcAddress(hDLL, "SetLayeredWindowAttributes");

		//Alpha Transparency Features Not Allowed
		if ( pUpdateLayeredWindow == NULL || pSetLayeredWindowAttributes == NULL )
		{
			alphaFadeTransparencyBackup = 255;
			alphaTransparency = 255;
			alphaMap = false;
			fakealphaMap = false;
			alphaFade = false;

			ModifyStyleEx(hWnd, WS_EX_LAYERED, 0 );

			layeredFlag = false;

			return;
		}

		alphaInitialized = true;
	}

	if ( alphaMap && !hasHookedModules )
	{
		UpdateLabelAlphaMap(newAlphaTransparency);
			
		if ( restartScrollRepaint )
		{
			restartScrollRepaint = false;
			SetTimer(hWnd, TIMER_SCROLL, scrollInterval, 0);
		}
	}
	else if ( fakealphaMap && !hasHookedModules )
	{
		UpdateLabelFakeAlphaMap(newAlphaTransparency);

		if ( restartScrollRepaint )
		{
			restartScrollRepaint = false;
			SetTimer(hWnd, TIMER_SCROLL, scrollInterval, 0);
		}
	}
	else
	{	
		pSetLayeredWindowAttributes(hWnd, 0, newAlphaTransparency, LWA_ALPHA);
	}

	if ( !afterClass.empty() )
	{
		HWND before = FindWindowEx2(afterClass, afterTitle);
		
		if ( before != NULL )
		{
			disablezOrderChange = false;

			SetWindowPos(hWnd, before,
							0, 0, 0, 0,
							SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE );
		}
	}
}

void Label::UpdateLabelAlphaMap(int newAlphaTransparency)
{
	RECT rectPosition;
	rectPosition.left = actualx;
	rectPosition.right = actualx+actualwidth;
	rectPosition.top = actualy;
	rectPosition.bottom = actualy+actualheight;
			
	int x = rectPosition.left;
	int y = rectPosition.top;
	int width = (rectPosition.right-rectPosition.left);
	int height = (rectPosition.bottom-rectPosition.top);

	rectPosition.left = 0;
	rectPosition.right = width;
	rectPosition.top = 0;
	rectPosition.bottom = height;

	if ( width <= 0 || height <= 0 )
		return;
		
	HDC backgroundDC = CreateCompatibleDC(NULL);

	BITMAPINFO alpha;
	ZeroMemory( &alpha.bmiHeader, sizeof(BITMAPINFOHEADER) );
	alpha.bmiHeader.biWidth=width;      // Set size you need
	alpha.bmiHeader.biHeight=height;    // Set size you need
	alpha.bmiHeader.biPlanes=1;
	alpha.bmiHeader.biBitCount=32;            // Can be 8, 16, 32 bpp or other number
	alpha.bmiHeader.biSizeImage=0;
	alpha.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
	alpha.bmiHeader.biClrUsed= 0;
	alpha.bmiHeader.biClrImportant= 0;
	VOID *pvBits;

	HBITMAP	backgroundBitmap = CreateDIBSection( backgroundDC,
													&alpha,
													DIB_RGB_COLORS,
													&pvBits,
													NULL,
													0 );

	HBITMAP oldBMP = (HBITMAP) SelectObject(backgroundDC, backgroundBitmap);
		
	// Use Pure Background (without Text) Backup for Normal Repaint (Scrolling, Text Update, ...)
	if ( scrollBitmap != 0 && scrollDC != 0 )
	{
		BitBlt(backgroundDC, 0, 0, width, height, scrollDC, 0, 0, SRCCOPY);
	}
	else
	//1*********************************************************************
	{
	//1*********************************************************************

		//PrePaint Background (Fake Transparency)
		if(transparencyMode == 2)
		{
			// paint desktop on display DC and then into background buffer
			PaintDesktopEx(backgroundDC, 0, 0, width, height, x, y, false);
		}
		//Default Background
		else if (transparencyMode == 3)
		{
			DrawEdge(backgroundDC, &rectPosition, EDGE_RAISED, BF_RECT | BF_SOFT | BF_MIDDLE);
		}

		//BG Painting
		if ( (!normalAnim && !hoverAnim) || frameCounter == 0 || (usePressed && pressedActive) )
		{
			if (pressedActive && usePressed)
			{
				background->apply(backgroundDC, 0, 0, width, height, 1, 2);
			}
			else if (hoverActive && useHover)
			{
				background->apply(backgroundDC, 0, 0, width, height, 1, 1);
			}
			else
			{
				background->apply(backgroundDC, 0, 0, width, height, 1, 0);
			}
		}
		else
		{
			if (pressedActive && usePressed)
			{
				background->applyAnim(backgroundDC, 0, 0, width, height, 1, 1, frameCounter);
			}
			else if (hoverActive && useHover)
			{
				background->applyAnim(backgroundDC, 0, 0, width, height, 1, 1, frameCounter);
			}
			else
			{
				background->applyAnim(backgroundDC, 0, 0, width, height, 1, 0, frameCounter);
			}
		}

		if ( !ghostlabelList.empty() )
		{
			for(GhostLabelListIterator glit = ghostlabelList.begin(); glit != ghostlabelList.end(); glit++)
			{
				if ( *glit != NULL )
				{
					if ( (*glit)->visible )
					{
						if ( (*glit)->pressed && ((pressedActive && usePressed && !(*glit)->standalone) || ((*glit)->isPressed && (*glit)->standalone)) )
							(*glit)->background->applyOverlay(backgroundDC, (*glit)->x, (*glit)->y, (*glit)->width, (*glit)->height, (*glit)->pressedalpha, 2);
						else if ( (*glit)->hover && ((hoverActive && useHover && !(*glit)->standalone) || ((*glit)->isHover && (*glit)->standalone)) )
							(*glit)->background->applyOverlay(backgroundDC, (*glit)->x, (*glit)->y, (*glit)->width, (*glit)->height, (*glit)->hoveralpha, 1);
						else if ( (*glit)->normal ) 
							(*glit)->background->applyOverlay(backgroundDC, (*glit)->x, (*glit)->y, (*glit)->width, (*glit)->height, (*glit)->alpha, 0);
					}
				}
				else
				{ 
					break;
				}
			}
		}

	//1*********************************************************************
	}
	//1*********************************************************************

	//Create ScrollBitmap, if it doesn't exist!
	//**********************************************************************
	if ( scroll != 0 && scrollBitmap == 0 && scrollDC == 0 )
	{
		if ( scrollDC == 0 )
			scrollDC = CreateCompatibleDC(NULL);

		BITMAPINFO alpha;
		ZeroMemory( &alpha.bmiHeader, sizeof(BITMAPINFOHEADER) );
		alpha.bmiHeader.biWidth=width;      // Set size you need
		alpha.bmiHeader.biHeight=height;    // Set size you need
		alpha.bmiHeader.biPlanes=1;
		alpha.bmiHeader.biBitCount=32;      // Can be 8, 16, 32 bpp or other number
		alpha.bmiHeader.biSizeImage=0;
		alpha.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
		alpha.bmiHeader.biClrUsed= 0;
		alpha.bmiHeader.biClrImportant= 0;
		VOID *pvBits;

		scrollBitmap = CreateDIBSection( scrollDC,
												&alpha,
												DIB_RGB_COLORS,
												&pvBits,
												NULL,
												0 );

		oldscrollBMP = (HBITMAP) SelectObject(scrollDC, scrollBitmap);

		BitBlt(scrollDC, 0, 0, width, height, backgroundDC, 0, 0, SRCCOPY);
	}
	//**********************************************************************

	//Text Painting Function		
	if ( !text.empty() )
	{
		if ( embossText != 0 )
		{
			HDC textDC = CreateCompatibleDC(NULL);
			HBITMAP textBitmap = CreateCompatibleBitmap(textDC, width, height);
			HBITMAP old0Bitmap = (HBITMAP) SelectObject(textDC, textBitmap);

			::SetBkColor(textDC, RGB(255, 0, 255));
			::ExtTextOut(textDC, 0, 0, ETO_OPAQUE, &rectPosition, NULL, 0, NULL);

			onTextPaint(textDC, width, height, true);

			textBitmap = (HBITMAP) SelectObject(textDC, old0Bitmap);

			HBITMAP workBGBitmap = (HBITMAP) SelectObject(backgroundDC, oldBMP);

			backgroundBitmap = Emboss(textBitmap, workBGBitmap, (embossText == 1 ? true : false), 0, 0, lightembossColor, darkembossColor);

			oldBMP = (HBITMAP) SelectObject(backgroundDC, backgroundBitmap);

			DeleteObject(workBGBitmap);
			DeleteObject(textBitmap);
			DeleteDC(textDC);
		}
		else if ( useTextFade )
		{
			TextFadeProcessing(backgroundDC, width, height);
		}
		else
		{
			onTextPaint(backgroundDC, width, height, true);
		}
	}

	if ( !ghostlabelList.empty() )
	{
		for(GhostLabelListIterator glit = ghostlabelList.begin(); glit != ghostlabelList.end(); glit++)
		{
			if ( *glit != NULL )
			{
				if ( (*glit)->visible )
				{
					if ( (*glit)->font != 0 )
					{
						int glleft = (*glit)->x+(*glit)->font->leftBorder;
						int glright = (*glit)->x+(*glit)->width-(*glit)->font->rightBorder;
						int gltop = (*glit)->y+(*glit)->font->topBorder;
						int glbottom = (*glit)->y+(*glit)->height-(*glit)->font->bottomBorder;

						//Create Region in which text drawing is allowed
						HRGN allowedArea = CreateRectRgn( glleft, gltop, glright, glbottom );
						SelectClipRgn(backgroundDC, allowedArea);
						DeleteObject(allowedArea);

						StringList tmplabelHREFClickRegions;
						if (pressedActive && (*glit)->font->usePressedText)
						{
							(*glit)->font->apply(backgroundDC,
													glleft,
													glright,
													gltop,
													glbottom,
													(*glit)->text,
													tmplabelHREFClickRegions,
													2,
													(*glit)->alphamap);
						}
						else if (hoverActive && (*glit)->font->useHoverText)
						{
							(*glit)->font->apply(backgroundDC,
													glleft,
													glright,
													gltop,
													glbottom,
													(*glit)->text,
													tmplabelHREFClickRegions,
													1,
													(*glit)->alphamap);
						}
						else
						{
							(*glit)->font->apply(backgroundDC,
													glleft,
													glright,
													gltop,
													glbottom,
													(*glit)->text,
													tmplabelHREFClickRegions,
													0,
													(*glit)->alphamap);
						}

						//Reset Region in which text drawing is allowed
						SelectClipRgn(backgroundDC, NULL);
					}
				}
			}
			else
			{ 
				break;
			}
		}
	}

	//Prepare Combined Image
	if (labelInitialized)
	{
		BITMAP bm;
		GetObject(backgroundBitmap, sizeof(BITMAP), &bm);

		BYTE *pPixel = (LPBYTE)bm.bmBits;

		int iSize = bm.bmHeight * bm.bmWidth;

		for (int j = 0; j < iSize; ++j)
		{
			//Convert Text to Full Opacity
			if ( pPixel[3] == 0 )
				pPixel[3] = 255;

			//Magic Pink
			if ( pPixel[0] == 255 && pPixel[1] == 0 && pPixel[2] == 255 )
			{
				pPixel[0] = 0;
				pPixel[1] = 0;
				pPixel[2] = 0;
				pPixel[3] = 0;
			}

			//Remove Prepared Alpha (1 back to 0)
			if (pPixel[3] == 1)
				pPixel[3] = 0;

			pPixel+= 4;
		}
	}	

	POINT ptWindowScreenPosition = {x, y};
	SIZE szWindow = {width, height};

	POINT ptSrc = {0,0};

	BLENDFUNCTION blendPixelFunction = { AC_SRC_OVER, 0, newAlphaTransparency, AC_SRC_ALPHA };

	pUpdateLayeredWindow(hWnd, NULL, &ptWindowScreenPosition, &szWindow, backgroundDC, &ptSrc, 0, &blendPixelFunction, ULW_ALPHA );

	if ( backgroundBitmap != 0 && backgroundDC != 0 )
	{
		backgroundBitmap = (HBITMAP) SelectObject(backgroundDC, oldBMP);
		DeleteObject(backgroundBitmap);
		DeleteDC(backgroundDC);
	}
}

void Label::UpdateLabelFakeAlphaMap(int newAlphaTransparency)
{
	RECT rectPosition;
	rectPosition.left = actualx;
	rectPosition.right = actualx+actualwidth;
	rectPosition.top = actualy;
	rectPosition.bottom = actualy+actualheight;
			
	int x = rectPosition.left;
	int y = rectPosition.top;
	int width = (rectPosition.right-rectPosition.left);
	int height = (rectPosition.bottom-rectPosition.top);
	
	rectPosition.left = 0;
	rectPosition.right = width;
	rectPosition.top = 0;
	rectPosition.bottom = height;

	if ( width <= 0 || height <= 0 )
		return;
				
	HDC backgroundDC = CreateCompatibleDC(NULL);

	BITMAPINFO alpha;
	ZeroMemory( &alpha.bmiHeader, sizeof(BITMAPINFOHEADER) );
	alpha.bmiHeader.biWidth=width;      // Set size you need
	alpha.bmiHeader.biHeight=height;    // Set size you need
	alpha.bmiHeader.biPlanes=1;
	alpha.bmiHeader.biBitCount=32;            // Can be 8, 16, 32 bpp or other number
	alpha.bmiHeader.biSizeImage=0;
	alpha.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
	alpha.bmiHeader.biClrUsed= 0;
	alpha.bmiHeader.biClrImportant= 0;
	VOID *pvBits;

	HBITMAP backgroundBitmap = CreateDIBSection( backgroundDC,
													&alpha,
													DIB_RGB_COLORS,
													&pvBits,
													NULL,
													0 );

	HBITMAP oldBMP = (HBITMAP) SelectObject(backgroundDC, backgroundBitmap);

	// Use Pure Background (without Text) Backup for Normal Repaint (Scrolling, Text Update, ...)
	if ( scrollBitmap != 0 && scrollDC != 0 )
	{
		BitBlt(backgroundDC, 0, 0, width, height, scrollDC, 0, 0, SRCCOPY);
	}
	else
	//1*********************************************************************
	{
	//1*********************************************************************

		//True Transparency
		if (transparencyMode == 1)
		{
			::SetBkColor(backgroundDC, RGB(255, 0, 255));
			::ExtTextOut(backgroundDC, 0, 0, ETO_OPAQUE, &rectPosition, NULL, 0, NULL);
		}
		//PrePaint Background (Fake Transparency)
		else if(transparencyMode == 2)
		{
			// paint desktop on display DC and then into background buffer
			PaintDesktopEx(backgroundDC, 0, 0, width, height, x, y, false);
		}
		//Default Background
		else if (transparencyMode == 3)
		{
			DrawEdge(backgroundDC, &rectPosition, EDGE_RAISED, BF_RECT | BF_SOFT | BF_MIDDLE);
		}

		//BG Painting
		if ( (!normalAnim && !hoverAnim) || frameCounter == 0 || (usePressed && pressedActive) )
		{
			 if (pressedActive && usePressed)
			 {
				background->apply(backgroundDC, 0, 0, width, height, 1, 2);
			 }
			 else if (hoverActive && useHover)
			 {
				background->apply(backgroundDC, 0, 0, width, height, 1, 1);
			 }
			 else
			 {
				background->apply(backgroundDC, 0, 0, width, height, 1, 0);
			 }
		}
		else
		{
			if (pressedActive && usePressed)
			{
				background->applyAnim(backgroundDC, 0, 0, width, height, 1, 1, frameCounter);
			}
			else if (hoverActive && useHover)
			{
				background->applyAnim(backgroundDC, 0, 0, width, height, 1, 1, frameCounter);
			}
			else
			{
				background->applyAnim(backgroundDC, 0, 0, width, height, 1, 0, frameCounter);
			}
		}

		if ( !ghostlabelList.empty() )
		{
			for(GhostLabelListIterator glit = ghostlabelList.begin(); glit != ghostlabelList.end(); glit++)
			{
				if ( *glit != NULL )
				{
					if ( (*glit)->visible )
					{
						if ( (*glit)->pressed && ((pressedActive && usePressed && !(*glit)->standalone) || ((*glit)->isPressed && (*glit)->standalone)) )
							(*glit)->background->applyOverlay(backgroundDC, (*glit)->x, (*glit)->y, (*glit)->width, (*glit)->height, (*glit)->pressedalpha, 2);
						else if ( (*glit)->hover && ((hoverActive && useHover && !(*glit)->standalone) || ((*glit)->isHover && (*glit)->standalone)) )
							(*glit)->background->applyOverlay(backgroundDC, (*glit)->x, (*glit)->y, (*glit)->width, (*glit)->height, (*glit)->hoveralpha, 1);
						else if ( (*glit)->normal ) 
							(*glit)->background->applyOverlay(backgroundDC, (*glit)->x, (*glit)->y, (*glit)->width, (*glit)->height, (*glit)->alpha, 0);
					}
				}
				else
				{ 
					break;
				}
			}
		}

	//1*********************************************************************
	}
	//1*********************************************************************

	//Create ScrollBitmap, if it doesn't exist!
	//**********************************************************************
	if ( scroll != 0 && scrollBitmap == 0 && scrollDC == 0 )
	{
		if ( scrollDC == 0 )
			scrollDC = CreateCompatibleDC(NULL);

		BITMAPINFO alpha;
		ZeroMemory( &alpha.bmiHeader, sizeof(BITMAPINFOHEADER) );
		alpha.bmiHeader.biWidth=width;      // Set size you need
		alpha.bmiHeader.biHeight=height;    // Set size you need
		alpha.bmiHeader.biPlanes=1;
		alpha.bmiHeader.biBitCount=32;      // Can be 8, 16, 32 bpp or other number
		alpha.bmiHeader.biSizeImage=0;
		alpha.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
		alpha.bmiHeader.biClrUsed= 0;
		alpha.bmiHeader.biClrImportant= 0;
		VOID *pvBits;

		scrollBitmap = CreateDIBSection( scrollDC,
												&alpha,
												DIB_RGB_COLORS,
												&pvBits,
												NULL,
												0 );

		oldscrollBMP = (HBITMAP) SelectObject(scrollDC, scrollBitmap);

		BitBlt(scrollDC, 0, 0, width, height, backgroundDC, 0, 0, SRCCOPY);
	}
	//**********************************************************************

	//Text Painting Function
	if ( !text.empty() )
	{
		if ( embossText != 0 )
		{
			HDC textDC = CreateCompatibleDC(NULL);
			HBITMAP textBitmap = CreateCompatibleBitmap(textDC, width, height);
			HBITMAP old0Bitmap = (HBITMAP) SelectObject(textDC, textBitmap);

			::SetBkColor(textDC, RGB(255, 0, 255));
			::ExtTextOut(textDC, 0, 0, ETO_OPAQUE, &rectPosition, NULL, 0, NULL);

			onTextPaint(textDC, width, height, true);

			textBitmap = (HBITMAP) SelectObject(textDC, old0Bitmap);

			HBITMAP workBGBitmap = (HBITMAP) SelectObject(backgroundDC, oldBMP);

			backgroundBitmap = Emboss(textBitmap, workBGBitmap, (embossText == 1 ? true : false), 0, 0, lightembossColor, darkembossColor);

			oldBMP = (HBITMAP) SelectObject(backgroundDC, backgroundBitmap);

			DeleteObject(workBGBitmap);
			DeleteObject(textBitmap);
			DeleteDC(textDC);
		}
		else if ( useTextFade )
		{
			TextFadeProcessing(backgroundDC, width, height);
		}
		else
		{
			onTextPaint(backgroundDC, width, height);
		}
	}

	if ( !ghostlabelList.empty() )
	{
		for(GhostLabelListIterator glit = ghostlabelList.begin(); glit != ghostlabelList.end(); glit++)
		{
			if ( *glit != NULL )
			{
				if ( (*glit)->visible )
				{
					if ( (*glit)->font != 0 )
					{
						int glleft = (*glit)->x+(*glit)->font->leftBorder;
						int glright = (*glit)->x+(*glit)->width-(*glit)->font->rightBorder;
						int gltop = (*glit)->y+(*glit)->font->topBorder;
						int glbottom = (*glit)->y+(*glit)->height-(*glit)->font->bottomBorder;

						//Create Region in which text drawing is allowed
						HRGN allowedArea = CreateRectRgn( glleft, gltop, glright, glbottom );
						SelectClipRgn(backgroundDC, allowedArea);
						DeleteObject(allowedArea);

						StringList tmplabelHREFClickRegions;
						if (pressedActive && (*glit)->font->usePressedText)
						{
							(*glit)->font->apply(backgroundDC,
													glleft,
													glright,
													gltop,
													glbottom,
													(*glit)->text,
													tmplabelHREFClickRegions,
													2,
													(*glit)->alphamap);
						}
						else if (hoverActive && (*glit)->font->useHoverText)
						{
							(*glit)->font->apply(backgroundDC,
													glleft,
													glright,
													gltop,
													glbottom,
													(*glit)->text,
													tmplabelHREFClickRegions,
													1,
													(*glit)->alphamap);
						}
						else
						{
							(*glit)->font->apply(backgroundDC,
													glleft,
													glright,
													gltop,
													glbottom,
													(*glit)->text,
													tmplabelHREFClickRegions,
													0,
													(*glit)->alphamap);
						}

						//Reset Region in which text drawing is allowed
						SelectClipRgn(backgroundDC, NULL);	
					}
				}
			}
			else
			{
				break;
			}
		}
	}

	//Prepare Combined Image
	if (labelInitialized)
	{
		BITMAP bm;
		GetObject(backgroundBitmap, sizeof(BITMAP), &bm);

		BYTE *pPixel = (LPBYTE)bm.bmBits;

		int iSize = bm.bmHeight * bm.bmWidth;

		for (int j = 0; j < iSize; ++j)
		{
			//Convert Text to Full Opacity
			if ( pPixel[3] == 0 )
				pPixel[3] = 255;

			//Magic Pink
			if ( pPixel[0] == 255 && pPixel[1] == 0 && pPixel[2] == 255 )
			{
				pPixel[0] = 0;
				pPixel[1] = 0;
				pPixel[2] = 0;
				pPixel[3] = 0;
			}
			else
			{
				pPixel[3] = 255;
			}

			pPixel+= 4;
		}
	}	

	POINT ptWindowScreenPosition = {x, y};
	SIZE szWindow = {width, height};

	POINT ptSrc = {0,0};

	BLENDFUNCTION blendPixelFunction = { AC_SRC_OVER, 0, newAlphaTransparency, AC_SRC_ALPHA };

	pUpdateLayeredWindow(hWnd, NULL, &ptWindowScreenPosition, &szWindow, backgroundDC, &ptSrc, 0, &blendPixelFunction, ULW_ALPHA );
	
	backgroundBitmap = (HBITMAP) SelectObject(backgroundDC, oldBMP);
	DeleteObject(backgroundBitmap);
	DeleteDC(backgroundDC);
}

void Label::onPaint(HDC hDC, RECT *specialRect)
{
	bool useSpecialRectPainting = false;

	if ( specialRect != NULL )
	{
		useSpecialRectPainting = true;
	}

	RECT rectPosition;
	rectPosition.left = actualx;
	rectPosition.right = actualx+actualwidth;
	rectPosition.top = actualy;
	rectPosition.bottom = actualy+actualheight;

	int x = rectPosition.left;
	int y = rectPosition.top;
	int width = (rectPosition.right-rectPosition.left);
	int height = (rectPosition.bottom-rectPosition.top);

	if ( width < 0 || height < 0 )
		return;

	rectPosition.left = 0;
	rectPosition.top = 0;
	rectPosition.right = width;
	rectPosition.bottom = height;

	HDC backgroundDC = 0;
	HBITMAP backgroundBitmap = 0;
	HBITMAP oldBMP = 0;

	// Use Complete Backup for WindowsInternal Repaint (Covering Windows, ...)
	if ( !useSpecialRectPainting && backupBitmap != 0 && backupDC != 0 )
	{
		BitBlt(hDC, 0, 0, width, height, backupDC, 0, 0, SRCCOPY);
	}
	else
	//1----------------------------------------------------------------
	{
	//1----------------------------------------------------------------

		backgroundDC = CreateCompatibleDC(NULL);

		BITMAPINFO alpha;
		ZeroMemory( &alpha.bmiHeader, sizeof(BITMAPINFOHEADER) );
		alpha.bmiHeader.biWidth=width;      // Set size you need
		alpha.bmiHeader.biHeight=height;    // Set size you need
		alpha.bmiHeader.biPlanes=1;
		alpha.bmiHeader.biBitCount=32;      // Can be 8, 16, 32 bpp or other number
		alpha.bmiHeader.biSizeImage=0;
		alpha.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
		alpha.bmiHeader.biClrUsed= 0;
		alpha.bmiHeader.biClrImportant= 0;
		VOID *pvBits;

		backgroundBitmap = CreateDIBSection( backgroundDC,
												&alpha,
												DIB_RGB_COLORS,
												&pvBits,
												NULL,
												0 );

		oldBMP = (HBITMAP) SelectObject(backgroundDC, backgroundBitmap);

		// Use Pure Background (without Text) Backup for Normal Repaint (Scrolling, Text Update, ...)
		if ( !useSpecialRectPainting && scrollBitmap != 0 && scrollDC != 0 )
		{
			BitBlt(backgroundDC, 0, 0, width, height, scrollDC, 0, 0, SRCCOPY);
		}
		else
		//2*********************************************************************
		{
		//2*********************************************************************

			//True Transparency
			if (transparencyMode == 1)
			{
				::SetBkColor(backgroundDC, RGB(255, 0, 255));
				::ExtTextOut(backgroundDC, 0, 0, ETO_OPAQUE, &rectPosition, NULL, 0, NULL);
			}
			//PrePaint Background (Fake Transparency)
			else if(transparencyMode == 2)
			{
				// paint desktop on display DC and then into background buffer
				PaintDesktopEx(backgroundDC, 0, 0, width, height, x, y, FALSE);
			}
			//Default Background
			else if (transparencyMode == 3)
			{
				DrawEdge(backgroundDC, &rectPosition, EDGE_RAISED, BF_RECT | BF_SOFT | BF_MIDDLE);
			}
					
			//Normal Background Painting if no Animation
			if ( (!normalAnim && !hoverAnim) || frameCounter == 0 || (usePressed && pressedActive) )
			{
				if (pressedActive && usePressed)
				{
					background->apply(backgroundDC, 0, 0, width, height, transparencyMode, 2);
				}
				else if (hoverActive && useHover)
				{
					background->apply(backgroundDC, 0, 0, width, height, transparencyMode, 1);
				}
				else
				{
					background->apply(backgroundDC, 0, 0, width, height, transparencyMode, 0);
				}
			}
			//Special Background Painting if Animation (normal or hover)
			else
			{
				if (pressedActive && usePressed)
				{
					background->applyAnim(backgroundDC, 0, 0, width, height, transparencyMode, 1, frameCounter);
				}
				else if (hoverActive && useHover)
				{
					background->applyAnim(backgroundDC, 0, 0, width, height, transparencyMode, 1, frameCounter);
				}
				else
				{
					background->applyAnim(backgroundDC, 0, 0, width, height, transparencyMode, 0, frameCounter);
				}
			}

			if ( !ghostlabelList.empty() )
			{
				for(GhostLabelListIterator glit = ghostlabelList.begin(); glit != ghostlabelList.end(); glit++)
				{
					if ( *glit != NULL )
					{
						if ( (*glit)->visible )
						{
							if ( IsOS( OS_2KXP ) )
							{
								if ( (*glit)->pressed && ((pressedActive && usePressed && !(*glit)->standalone) || ((*glit)->isPressed && (*glit)->standalone)) )
									(*glit)->background->applyOverlay(backgroundDC, (*glit)->x, (*glit)->y, (*glit)->width, (*glit)->height, (*glit)->pressedalpha, 2);
								else if ( (*glit)->hover && ((hoverActive && useHover && !(*glit)->standalone) || ((*glit)->isHover && (*glit)->standalone)) )
									(*glit)->background->applyOverlay(backgroundDC, (*glit)->x, (*glit)->y, (*glit)->width, (*glit)->height, (*glit)->hoveralpha, 1);
								else if ( (*glit)->normal ) 
									(*glit)->background->applyOverlay(backgroundDC, (*glit)->x, (*glit)->y, (*glit)->width, (*glit)->height, (*glit)->alpha, 0);
							}
							else
							{
								if ( (*glit)->pressed && ((pressedActive && usePressed && !(*glit)->standalone) || ((*glit)->isPressed && (*glit)->standalone)) )
									(*glit)->background->apply(backgroundDC, (*glit)->x, (*glit)->y, (*glit)->width, (*glit)->height, transparencyMode, 2);
								else if ( (*glit)->hover && ((hoverActive && useHover && !(*glit)->standalone) || ((*glit)->isHover && (*glit)->standalone))  )
									(*glit)->background->apply(backgroundDC, (*glit)->x, (*glit)->y, (*glit)->width, (*glit)->height, transparencyMode, 1);
								else if ( (*glit)->normal ) 
									(*glit)->background->apply(backgroundDC, (*glit)->x, (*glit)->y, (*glit)->width, (*glit)->height, transparencyMode, 0);
							}
						}
					}
					else
					{
						break;
					}
				}
			}

		//2*********************************************************************
		}
		//2*********************************************************************

		//Create ScrollBitmap, if it doesn't exist!
		//**********************************************************************
		if ( !useSpecialRectPainting && scroll != 0 && scrollBitmap == 0 && scrollDC == 0 )
		{
			if ( scrollDC == 0 )
				scrollDC = CreateCompatibleDC(NULL);

			BITMAPINFO alpha;
			ZeroMemory( &alpha.bmiHeader, sizeof(BITMAPINFOHEADER) );
			alpha.bmiHeader.biWidth=width;      // Set size you need
			alpha.bmiHeader.biHeight=height;    // Set size you need
			alpha.bmiHeader.biPlanes=1;
			alpha.bmiHeader.biBitCount=32;      // Can be 8, 16, 32 bpp or other number
			alpha.bmiHeader.biSizeImage=0;
			alpha.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
			alpha.bmiHeader.biClrUsed= 0;
			alpha.bmiHeader.biClrImportant= 0;
			VOID *pvBits;

			scrollBitmap = CreateDIBSection( scrollDC,
												&alpha,
												DIB_RGB_COLORS,
												&pvBits,
												NULL,
												0 );

			oldscrollBMP = (HBITMAP) SelectObject(scrollDC, scrollBitmap);

			BitBlt(scrollDC, 0, 0, width, height, backgroundDC, 0, 0, SRCCOPY);
		}
		//**********************************************************************

		//Only needed with Text
		if ( !text.empty() && !useSpecialRectPainting )
		{
			if ( embossText != 0 )
			{
				HDC textDC = CreateCompatibleDC(NULL);
				HBITMAP textBitmap = CreateCompatibleBitmap(textDC, width, height);
				HBITMAP old0Bitmap = (HBITMAP) SelectObject(textDC, textBitmap);

				::SetBkColor(textDC, RGB(255, 0, 255));
				::ExtTextOut(textDC, 0, 0, ETO_OPAQUE, &rectPosition, NULL, 0, NULL);

				onTextPaint(textDC, width, height, true);

				textBitmap = (HBITMAP) SelectObject(textDC, old0Bitmap);

				HBITMAP workBGBitmap = (HBITMAP) SelectObject(backgroundDC, oldBMP);

				backgroundBitmap = Emboss(textBitmap, workBGBitmap, (embossText == 1 ? true : false), 0, 0, lightembossColor, darkembossColor);

				oldBMP = (HBITMAP) SelectObject(backgroundDC, backgroundBitmap);

				DeleteObject(workBGBitmap);
				DeleteObject(textBitmap);
				DeleteDC(textDC);
			}
			else if ( useTextFade )
			{
				TextFadeProcessing(backgroundDC, width, height);
			}
			else
			{
				onTextPaint(backgroundDC, width, height);
			}
		}

		if ( !ghostlabelList.empty() && !useSpecialRectPainting )
		{
			for(GhostLabelListIterator glit = ghostlabelList.begin(); glit != ghostlabelList.end(); glit++)
			{
				if ( *glit != NULL )
				{
					if ( (*glit)->visible )
					{
						if ( (*glit)->font != 0 )
						{
							int glleft = (*glit)->x+(*glit)->font->leftBorder;
							int glright = (*glit)->x+(*glit)->width-(*glit)->font->rightBorder;
							int gltop = (*glit)->y+(*glit)->font->topBorder;
							int glbottom = (*glit)->y+(*glit)->height-(*glit)->font->bottomBorder;

							//Create Region in which text drawing is allowed
							HRGN allowedArea = CreateRectRgn( glleft, gltop, glright, glbottom );
							SelectClipRgn(backgroundDC, allowedArea);
							DeleteObject(allowedArea);

							StringList tmplabelHREFClickRegions;
							if (pressedActive && (*glit)->font->usePressedText)
							{
								(*glit)->font->apply(backgroundDC,
														glleft,
														glright,
														gltop,
														glbottom,
														(*glit)->text,
														tmplabelHREFClickRegions,
														2,
														(*glit)->alphamap);
							}
							else if (hoverActive && (*glit)->font->useHoverText)
							{
								(*glit)->font->apply(backgroundDC,
														glleft,
														glright,
														gltop,
														glbottom,
														(*glit)->text,
														tmplabelHREFClickRegions,
														1,
														(*glit)->alphamap);
							}
							else
							{
								(*glit)->font->apply(backgroundDC,
														glleft,
														glright,
														gltop,
														glbottom,
														(*glit)->text,
														tmplabelHREFClickRegions,
														0,
														(*glit)->alphamap);
							}

							//Reset Region in which text drawing is allowed
							SelectClipRgn(backgroundDC, NULL);
						}
					}
				}
				else
				{
					break;
				}
			}
		}

		//TextEditBox BG
		//----------------------------------------------------------------------------
		if ( hWndTextBox && textboxVisible )
		{
			if ( textboxPaintingMode == -1 )
			{
				::SetBkColor(backgroundDC, textboxPlainColor);
				::ExtTextOut(backgroundDC, 0, 0, ETO_OPAQUE, &TextEditBoxSizeBorder, NULL, 0, NULL);
			}
			else if ( texteditboxbackground )
			{
				texteditboxbackground->apply(backgroundDC, TextEditBoxSizeBorder.left, TextEditBoxSizeBorder.top, (TextEditBoxSizeBorder.right-TextEditBoxSizeBorder.left), (TextEditBoxSizeBorder.bottom-TextEditBoxSizeBorder.top), false, 0);
			}
		}
		//----------------------------------------------------------------------------

		// blt the composed image to the display
		if ( useSpecialRectPainting )
			BitBlt(hDC, 0, 0, (specialRect->right-specialRect->left), (specialRect->bottom-specialRect->top), backgroundDC, specialRect->left, specialRect->top, SRCCOPY);
		else
			BitBlt(hDC, 0, 0, width, height, backgroundDC, 0, 0, SRCCOPY);

		//Create BackupBitmap, if it doesn't exist!
		//----------------------------------------------------------------------
		if ( !useSpecialRectPainting && backupBitmap == 0 && backupDC == 0 )
		{
			if ( backupDC == 0 )
				backupDC = CreateCompatibleDC(NULL);

			BITMAPINFO alpha;
			ZeroMemory( &alpha.bmiHeader, sizeof(BITMAPINFOHEADER) );
			alpha.bmiHeader.biWidth=width;      // Set size you need
			alpha.bmiHeader.biHeight=height;    // Set size you need
			alpha.bmiHeader.biPlanes=1;
			alpha.bmiHeader.biBitCount=32;      // Can be 8, 16, 32 bpp or other number
			alpha.bmiHeader.biSizeImage=0;
			alpha.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
			alpha.bmiHeader.biClrUsed= 0;
			alpha.bmiHeader.biClrImportant= 0;
			VOID *pvBits;

			backupBitmap = CreateDIBSection( backupDC,
												&alpha,
												DIB_RGB_COLORS,
												&pvBits,
												NULL,
												0 );

			oldbackupBMP = (HBITMAP) SelectObject(backupDC, backupBitmap);

			BitBlt(backupDC, 0, 0, width, height, backgroundDC, 0, 0, SRCCOPY);
		}
		//----------------------------------------------------------------------

	//1----------------------------------------------------------------
	}
	//1----------------------------------------------------------------

	if ( backgroundBitmap != 0 && backgroundDC != 0 )
	{
		backgroundBitmap = (HBITMAP) SelectObject(backgroundDC, oldBMP);
		DeleteObject(backgroundBitmap);
		DeleteDC(backgroundDC);
	}
}

void Label::onTextPaint(HDC desthdc, int width, int height, bool smoothoff)
{
	//--------------------------------------------------------------------------------
	// start rendering text
	//--------------------------------------------------------------------------------
	//Create Region in which text drawing is allowed
	HRGN allowedArea = CreateRectRgn( leftBorder, topBorder, width-rightBorder, height-bottomBorder );
	SelectClipRgn(desthdc, allowedArea);
	DeleteObject(allowedArea);
	
	int textWidth = oldscrolltextwidth;
	int textHeight = oldscrolltextheight;

	if ( ((scroll == 1 || scroll == 3) && textWidth > width-leftBorder-rightBorder) || ((scroll == 2 || scroll == 4) && textHeight > height-topBorder-bottomBorder) )
	{			
		if ( skipScrollRepaint )
		{
			skipScrollRepaint = false;
			scrollPosition = 0;
		}

		if ( !restartScrollRepaint )
			restartScrollRepaint = true;

		if (scroll == 1 || scroll == 3)
		{
			if(scrollPosition <= 0)
			{
				scrollPosition = textWidth + scrollPadLength;
					
				if (scrollLimit)
				{
					scrollLimit--;
					if (scrollLimit == 0)
						setScrolling(0);
				}
			}

			if(scrollPosition > textWidth + scrollPadLength)
				scrollPosition = 0;
		}
		else if (scroll == 2 || scroll == 4)
		{
			if(scrollPosition <= 0)
			{
				scrollPosition = textHeight + scrollPadLength;
					
				if (scrollLimit)
				{
					scrollLimit--;
					if (scrollLimit == 0)
						setScrolling(0);
				}
			}

			if(scrollPosition > textHeight + scrollPadLength )
				scrollPosition = 0;
		}
		
		if (scroll == 1 || scroll == 3)
		{
			if (pressedActive && font->usePressedText)
			{
				font->apply(desthdc,
							scrollPosition,
							scrollPosition - scrollPadLength,
							topBorder,
							height - bottomBorder,
							text,
							labelHREFClickRegions,
							2,
							smoothoff,
							0,
							-1,
							true);

				StringList tmplabelHREFClickRegions;
				tmplabelHREFClickRegions.clear();

				font->apply(desthdc,
							scrollPosition - scrollPadLength - textWidth,
							scrollPosition + textWidth,
							topBorder,
							height - bottomBorder,
							text,
							tmplabelHREFClickRegions,
							2,
							smoothoff,
							0,
							-1,
							true);

				labelHREFClickRegions.merge(tmplabelHREFClickRegions);
				tmplabelHREFClickRegions.clear();
			}
			else if (hoverActive && font->useHoverText)
			{
				font->apply(desthdc,
							scrollPosition,
							scrollPosition - scrollPadLength,
							topBorder,
							height - bottomBorder,
							text,
							labelHREFClickRegions,
							1,
							smoothoff,
							0,
							-1,
							true);

				StringList tmplabelHREFClickRegions;
				tmplabelHREFClickRegions.clear();

				font->apply(desthdc,
							scrollPosition - scrollPadLength - textWidth,
							scrollPosition + textWidth,
							topBorder,
							height - bottomBorder,
							text,
							tmplabelHREFClickRegions,
							1,
							smoothoff,
							0,
							-1,
							true);

				labelHREFClickRegions.merge(tmplabelHREFClickRegions);
				tmplabelHREFClickRegions.clear();
			}
			else
			{
				font->apply(desthdc,
							scrollPosition,
							scrollPosition - scrollPadLength,
							topBorder,
							height - bottomBorder,
							text,
							labelHREFClickRegions,
							0,
							smoothoff,
							0,
							-1,
							true);

				StringList tmplabelHREFClickRegions;
				tmplabelHREFClickRegions.clear();

				font->apply(desthdc,
							scrollPosition - scrollPadLength - textWidth,
							scrollPosition + textWidth,
							topBorder,
							height - bottomBorder,
							text,
							tmplabelHREFClickRegions,
							0,
							smoothoff,
							0,
							-1,
							true);

				labelHREFClickRegions.merge(tmplabelHREFClickRegions);
				tmplabelHREFClickRegions.clear();
			}
		}
		else if (scroll == 2 || scroll == 4)
		{
			if (pressedActive && font->usePressedText)
			{
				font->apply(desthdc,
						leftBorder,
						width - rightBorder,
						scrollPosition,
						scrollPosition - scrollPadLength,
						text,
						labelHREFClickRegions,
						2,
						smoothoff,
						-1,
						2,
						true);

				StringList tmplabelHREFClickRegions;
				tmplabelHREFClickRegions.clear();

				font->apply(desthdc,
						leftBorder,
						width - rightBorder,
						scrollPosition - scrollPadLength - textHeight,
						scrollPosition + textHeight,
						text,
						tmplabelHREFClickRegions,
						2,
						smoothoff,
						-1,
						2,
						true);

				labelHREFClickRegions.merge(tmplabelHREFClickRegions);
				tmplabelHREFClickRegions.clear();
			}
			else if (hoverActive && font->useHoverText)
			{
				font->apply(desthdc,
						leftBorder,
						width - rightBorder,
						scrollPosition,
						scrollPosition - scrollPadLength,
						text,
						labelHREFClickRegions,
						1,
						smoothoff,
						-1,
						2,
						true);

				StringList tmplabelHREFClickRegions;
				tmplabelHREFClickRegions.clear();

				font->apply(desthdc,
						leftBorder,
						width - rightBorder,
						scrollPosition - scrollPadLength - textHeight,
						scrollPosition + textHeight,
						text,
						tmplabelHREFClickRegions,
						1,
						smoothoff,
						-1,
						2,
						true);

				labelHREFClickRegions.merge(tmplabelHREFClickRegions);
				tmplabelHREFClickRegions.clear();
			}
			else
			{
				font->apply(desthdc,
						leftBorder,
						width - rightBorder,
						scrollPosition,
						scrollPosition - scrollPadLength,
						text,
						labelHREFClickRegions,
						0,
						smoothoff,
						-1,
						2,
						true);

				StringList tmplabelHREFClickRegions;
				tmplabelHREFClickRegions.clear();

				font->apply(desthdc,
						leftBorder,
						width - rightBorder,
						scrollPosition - scrollPadLength - textHeight,
						scrollPosition + textHeight,
						text,
						tmplabelHREFClickRegions,
						0,
						smoothoff,
						-1,
						2,
						true);

				labelHREFClickRegions.merge(tmplabelHREFClickRegions);
				tmplabelHREFClickRegions.clear();
			}
		}
	}
	else
	{
		if ( !skipScrollRepaint )
		{
			skipScrollRepaint = true;
		}

		if (pressedActive && font->usePressedText)
		{
			font->apply(desthdc,
						leftBorder,
						width-rightBorder,
						topBorder,
						height-bottomBorder,
						text,
						labelHREFClickRegions,
						2,
						smoothoff);
		}
		else if (hoverActive && font->useHoverText)
		{
			font->apply(desthdc,
						leftBorder,
						width-rightBorder,
						topBorder,
						height-bottomBorder,
						text,
						labelHREFClickRegions,
						1,
						smoothoff);
		}
		else 
		{
			font->apply(desthdc,
						leftBorder,
						width-rightBorder,
						topBorder,
						height-bottomBorder,
						text,
						labelHREFClickRegions,
						0,
						smoothoff);
		}
	}

	//Reset Region in which text drawing is allowed
	SelectClipRgn(desthdc, NULL);
	//--------------------------------------------------------------------------------
	// stop rendering text
	//--------------------------------------------------------------------------------
}

void Label::TextFadeProcessing(HDC workingDC, int width, int height)
{
	if ( width <= 0 || height <= 0 )
		return;

	HDC textFadeDC = CreateCompatibleDC(NULL);

	BITMAPINFO alpha;
	ZeroMemory( &alpha.bmiHeader, sizeof(BITMAPINFOHEADER) );
	alpha.bmiHeader.biWidth=width;      // Set size you need
	alpha.bmiHeader.biHeight=height;    // Set size you need
	alpha.bmiHeader.biPlanes=1;
	alpha.bmiHeader.biBitCount=32;      // Can be 8, 16, 32 bpp or other number
	alpha.bmiHeader.biSizeImage=0;
	alpha.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
	alpha.bmiHeader.biClrUsed= 0;
	alpha.bmiHeader.biClrImportant= 0;
	VOID *pvBits;

	HBITMAP textFadeBitmap = CreateDIBSection( textFadeDC,
												&alpha,
												DIB_RGB_COLORS,
												&pvBits,
												NULL,
												0 );

	HBITMAP oldBMP = (HBITMAP) SelectObject(textFadeDC, textFadeBitmap);

	BitBlt(textFadeDC, 0, 0, width, height, workingDC, 0, 0, SRCCOPY);

	BITMAP bm;
	GetObject(textFadeBitmap, sizeof(BITMAP), &bm);

	BYTE *pPixel = (LPBYTE)bm.bmBits;

	int iSize = bm.bmHeight * bm.bmWidth;

	for (int j = 0; j < iSize; ++j)
	{
		if ( alphaMap || fakealphaMap )
		{
			pPixel[0] = 0;
			pPixel[1] = 0;
			pPixel[2] = 0;
			pPixel[3] = 1;
		}
		else
		{
			pPixel[3] = 1;
		}

		pPixel+= 4;
	}

	onTextPaint(textFadeDC, width, height, (alphaMap || fakealphaMap) ? true : false);

	GetObject(textFadeBitmap, sizeof(BITMAP), &bm);
	pPixel = (LPBYTE)bm.bmBits;

	int tmpx = 0;
	int tmpy = 0;
	int xval = 255;
	int yval = 255;

	iSize = bm.bmHeight * bm.bmWidth;

	int tmpleftValue = leftBorder+textFadeLeft;
	int tmprightValue1 = width-rightBorder-textFadeRight;
	int tmprightValue2 = width-rightBorder; 
	int tmptopValue1 = height-topBorder-textFadeTop;
	int tmptopValue2 = height-topBorder;
	int tmpbottomValue = bottomBorder+textFadeBottom;

	for (j = 0; j < iSize; ++j)
	{
		//Convert Text to Full Opacity
		if (pPixel[3] == 0)
		{
			xval = 255;
			yval = 255;

			tmpx = (j+1)%width;
			tmpy = j/width;

			if ( tmpx == 0 )
				tmpx = width-1;
			if ( tmpy == 0 )
				tmpy = 1;

			if ( textFadeLeft != 0 && tmpx > leftBorder && tmpx <= tmpleftValue )
			{
				xval = (255/textFadeLeft)*(tmpx-leftBorder);
			}
			else if ( textFadeRight != 0 && tmpx > tmprightValue1 && tmpx <= tmprightValue2 )
			{
				xval = (255/textFadeRight)*(tmprightValue2-tmpx);
			}

			//Order is reverted Top<->Bottom
			if ( textFadeBottom != 0 && tmpy > bottomBorder && tmpy < tmpbottomValue )
			{
				yval = (255/textFadeBottom)*(tmpy-bottomBorder);
			}
			else if ( textFadeTop != 0 && tmpy >= tmptopValue1 && tmpy < tmptopValue2 )
			{
				yval = (255/textFadeTop)*(tmptopValue2-tmpy);
			}
			
			if ( xval == 255 )
				pPixel[3] = yval;
			else if ( yval == 255 )
				pPixel[3] = xval;
			else
				pPixel[3] = (xval*yval)/255;
				
			//PreMultiply
			pPixel[0] = pPixel[0] * pPixel[3] / 255;
			pPixel[1] = pPixel[1] * pPixel[3] / 255;
			pPixel[2] = pPixel[2] * pPixel[3] / 255;
		}
		else
		{
			pPixel[0] = 0;
			pPixel[1] = 0;
			pPixel[2] = 0;
			pPixel[3] = 0; 
		}
					
		pPixel+= 4;	
	}

	BLENDFUNCTION blendFunction = { AC_SRC_OVER, 0, 255, AC_SRC_ALPHA };

	AlphaBlend(workingDC, 0, 0, width, height, textFadeDC, 0, 0, width, height, blendFunction);

	textFadeBitmap = (HBITMAP) SelectObject(textFadeDC, oldBMP);
	DeleteObject(textFadeBitmap);
	DeleteDC(textFadeDC);
}

HBITMAP Label::Emboss( HBITMAP hBitmap, HBITMAP hbmBackGnd,
			   BOOL bRaised, int xDest, int yDest,
			   COLORREF clrHighlight, COLORREF clrShadow )
{
	const DWORD PSDPxax = 0x00B8074A;
	BITMAP   bmInfo;

	HBITMAP  hbmOld = NULL;
	HBITMAP	 hbmShadow = NULL;
	HBITMAP	 hbmHighlight = NULL;
	HBITMAP	 hbmResult = NULL;
	HBITMAP	 hbmOldMem = NULL;

	HBRUSH   hbrPat = NULL;

	HDC		 hColorDC = NULL;
	HDC		 hMonoDC = NULL;
	HDC		 hMemDC = NULL;

	if( !bRaised )
	{
		// Swap the highlight and shadow color
		COLORREF clrTemp = clrShadow;
		clrShadow = clrHighlight;
		clrHighlight = clrTemp;
	}

	// We create two monochrome bitmaps. One of them will contain the
	// highlighted edge and the other will contain the shadow. These
	// bitmaps are then used to paint the highlight and shadow on the
	// background image.

	// Create a compatible DCs
	hMemDC = CreateCompatibleDC( NULL );
	hMonoDC = CreateCompatibleDC( NULL );
	hColorDC = CreateCompatibleDC( NULL );

	if( hMemDC == NULL || hMonoDC == NULL || hColorDC == NULL )
	{
		if( hMemDC )
			DeleteDC( hMemDC );
		if( hMonoDC )
			DeleteDC( hMonoDC );
		if( hColorDC )
			DeleteDC( hColorDC );

		return NULL;
	}

	// Select the background image into memory DC so that we can draw it
	hbmOldMem = (HBITMAP) SelectObject( hMemDC, hbmBackGnd );

	// Get dimensions of the background image
	BITMAP bm;
	GetObject( hbmBackGnd, sizeof(bm), &bm );

	// Create the monochrome and compatible color bitmaps 
	GetObject( hBitmap, sizeof(BITMAP), (LPSTR) &bmInfo );
	hbmShadow =	CreateBitmap( bmInfo.bmWidth, bmInfo.bmHeight, 1, 1, NULL );
	hbmHighlight = CreateBitmap( bmInfo.bmWidth, bmInfo.bmHeight, 1, 1, NULL );

	BITMAPINFO alpha;
	ZeroMemory( &alpha.bmiHeader, sizeof(BITMAPINFOHEADER) );
	alpha.bmiHeader.biWidth=bmInfo.bmWidth;      // Set size you need
	alpha.bmiHeader.biHeight=bmInfo.bmHeight;    // Set size you need
	alpha.bmiHeader.biPlanes=1;
	alpha.bmiHeader.biBitCount=32;      // Can be 8, 16, 32 bpp or other number
	alpha.bmiHeader.biSizeImage=0;
	alpha.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
	alpha.bmiHeader.biClrUsed= 0;
	alpha.bmiHeader.biClrImportant= 0;
	VOID *pvBits;

	hbmResult = CreateDIBSection( hMemDC,
										&alpha,
										DIB_RGB_COLORS,
										&pvBits,
										NULL,
										0 );

	hbmOld = (HBITMAP) SelectObject( hColorDC, hBitmap );

	// Set background color of bitmap for mono conversion
	// We assume that the pixel in the top left corner has the background color
	SetBkColor( hColorDC, GetPixel( hColorDC, 0, 0 ) ) ;

	// Create the highlight bitmap.
	hbmHighlight = (HBITMAP) SelectObject( hMonoDC, hbmHighlight );
	PatBlt( hMonoDC, 0, 0, bmInfo.bmWidth, bmInfo.bmHeight, WHITENESS );
	BitBlt( hMonoDC, 0, 0, bmInfo.bmWidth-1, bmInfo.bmHeight-1, hColorDC, 1, 1, SRCCOPY );
	BitBlt( hMonoDC, 0, 0, bmInfo.bmWidth, bmInfo.bmHeight,	hColorDC, 0, 0, MERGEPAINT );
	hbmHighlight = (HBITMAP) SelectObject( hMonoDC, hbmHighlight );

	// create the shadow bitmap
	hbmShadow = (HBITMAP) SelectObject( hMonoDC, hbmShadow );
	PatBlt( hMonoDC, 0, 0, bmInfo.bmWidth, bmInfo.bmHeight, WHITENESS );
	BitBlt( hMonoDC, 1, 1, bmInfo.bmWidth-1, bmInfo.bmHeight-1, hColorDC, 0, 0, SRCCOPY );
	BitBlt( hMonoDC, 0, 0, bmInfo.bmWidth, bmInfo.bmHeight,	hColorDC, 0, 0, MERGEPAINT );
	hbmShadow = (HBITMAP) SelectObject( hMonoDC, hbmShadow );

	// Now let's start working on the final image
	SelectObject( hColorDC, hbmResult );
	
	// Draw the background image
	BitBlt(hColorDC, 0, 0, bm.bmWidth, bm.bmHeight, hMemDC, 0, 0,SRCCOPY);
	// Restore the old bitmap in the hMemDC
	SelectObject( hMemDC, hbmOldMem );

	// Set the background and foreground color for the raster operations
	SetBkColor( hColorDC, RGB(255,255,255) );
	SetTextColor( hColorDC, RGB(0,0,0) );

	// blt the highlight edge
	hbrPat = CreateSolidBrush( clrHighlight );
	SelectObject( hColorDC, hbrPat );
	hbmHighlight = (HBITMAP) SelectObject( hMonoDC, hbmHighlight );
	BitBlt( hColorDC, xDest, yDest, bmInfo.bmWidth, bmInfo.bmHeight, hMonoDC, 0, 0, PSDPxax );
	DeleteObject( hbrPat );
	hbmHighlight = (HBITMAP) SelectObject( hMonoDC, hbmHighlight );

	// blt the shadow edge
	hbrPat = CreateSolidBrush( clrShadow );
	SelectObject( hColorDC, hbrPat );
	hbmShadow = (HBITMAP) SelectObject( hMonoDC, hbmShadow );
	BitBlt( hColorDC, xDest, yDest, bmInfo.bmWidth, bmInfo.bmHeight, hMonoDC, 0, 0, PSDPxax );
	DeleteObject( hbrPat );
	hbmShadow = (HBITMAP) SelectObject( hMonoDC, hbmShadow );

	// select old bitmap into color DC 
	SelectObject( hColorDC, hbmOld );

	DeleteObject( hbmShadow );
	DeleteObject( hbmHighlight );

	DeleteDC(hMemDC);
	DeleteDC(hColorDC);
	DeleteDC(hMonoDC);

	return ( hbmResult );
}